# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####


bl_info = {
    "name": "Auto Bake",
    "description": "Automate your texture baking and saving process, set up baked versions automatically for materials and objects, and export the baked objects!",
    "author": "Tsybe",
    "version": (1, 5),
    "blender": (3, 3, 0),
    "location": "Properties > Render > Auto Bake",
    "category": "Baking",
    "doc_url": "https://trello.com/b/9lVjALHM",
    "tracker_url": "https://discord.gg/KduczDQ7JT",
    }

import bpy
import os
import platform
import re
import math
import mathutils

from bpy.props import (StringProperty, BoolProperty, IntProperty, FloatProperty, EnumProperty, PointerProperty, CollectionProperty)
from bpy.types import (Panel, Operator, PropertyGroup, UIList, Menu)

 
#------------------------------------------------------------------------------------
#   Variables

bake_status = "IDLE"
is_udim_bake = False
next_object_bake = False
finished_bake_count = 0
reconnect_nodes = []
delete_nodes = []
bake_results = {}
bake_order = []

bake_items = [
    ('', "Shader", "", -1),
    ('Base Color', "Base Color", "Bakes input sockets named 'Base Color', 'Color', and 'Tint' for shader nodes", 1),
    ('Metallic', "Metallic", "Bakes input sockets named 'Metallic' for shader nodes", 2),
    ('Roughness', "Roughness", "Bakes input sockets named 'Roughness' for shader nodes", 3),
    ('IOR', "IOR", "Bakes input sockets named 'IOR' for shader nodes", 4),
    ('Alpha', "Alpha", "Bakes input sockets named 'Alpha' and 'Transparency' for shader nodes", 5),
    ('Normal', "Normal", "Bakes input sockets named 'Normal' for shader nodes", 6),

    ('', "* Subsurface", "", -1),
    ('Subsurface Weight', "Subsurface Weight", "Bakes input sockets named 'Subsurface Weight' for shader nodes", 7),
    ('Subsurface Radius', "Subsurface Radius", "Bakes input sockets named 'Subsurface Radius' and 'Radius' for shader nodes", 8),
    ('Subsurface Scale', "Subsurface Scale", "Bakes input sockets named 'Subsurface Scale' for shader nodes", 9),
    ('Subsurface IOR', "Subsurface IOR", "Bakes input sockets named 'Subsurface IOR' for shader nodes", 10),
    ('Subsurface Anisotropy', "Subsurface Anisotropy", "Bakes input sockets named 'Subsurface Anisotropy' and 'Anisotropy' for shader nodes", 11),

    ('', "* Specular", "", -1),
    ('Specular IOR Level', "Specular IOR Level", "Bakes input sockets named 'Specular IOR Level' for shader nodes", 12),
    ('Specular Tint', "Specular Tint", "Bakes input sockets named 'Specular Tint' for shader nodes", 13),
    ('Anisotropic', "Anisotropic", "Bakes input sockets named 'Anisotropic' for shader nodes", 14),
    ('Anisotropic Rotation', 'Anisotropic Rotation', "Bakes input sockets named 'Anisotropic Rotation' for shader nodes", 15),
    ('Tangent', "Tangent", "Bakes input sockets named 'Tangent' for shader nodes", 16),

    ('', "* Coat", "", -1),
    ('Coat Weight', 'Coat Weight', "Bakes input sockets named 'Coat Weight' and 'Clear Coat' for shader nodes", 17),
    ('Coat Roughness', 'Coat Roughness', "Bakes input sockets named 'Coat Roughness' and 'Clear Coat Roughness' for shader nodes", 18),
    ('Coat IOR', 'Coat IOR', "Bakes input sockets named 'Coat IOR' for shader nodes", 19),
    ('Coat Tint', 'Coat Tint', "Bakes input sockets named 'Coat Tint' for shader nodes", 20),
    ('Coat Normal', 'Coat Normal', "Bakes input sockets named 'Coat Normal' and 'Clear Coat Normal' for shader nodes", 21),

    ('', "*", "", -1),
    ('Transmission Weight', "Transmission Weight", "Bakes input sockets named 'Transmission Weight' for shader nodes", 22),
    ('Sheen Weight', "Sheen Weight", "Bakes input sockets named 'Sheen Weight' for shader nodes", 23),
    ('Sheen Roughness', "Sheen Roughness", "Bakes input sockets named 'Sheen Roughness' for shader nodes", 24),
    ('Sheen Tint', 'Sheen Tint', "Bakes input sockets named 'Sheen Tint' for shader nodes", 25),
    ('Emission Color', "Emission Color", "Bakes input sockets named 'Emission Color' and 'Emissive Color' for shader nodes", 26),
    ('Emission Strength', "Emission Strength", "Bakes input sockets named 'Emission Strength' for shader nodes", 27),

    ('', "Miscellaneous", "", -1),
    ('Channel Packing', "Channel Packing", "Bakes up to 3 different BW bake types into a single RGB texture", 28),
    ('Color Attribute', "Color Attribute", "Bakes the set or available color attribute", 29),
    ('Ambient Occlusion', 'Ambient Occlusion', "Bakes ambient occlusion using the 'Ambient Occlusion' node", 30),
    ('Pointiness', 'Pointiness', "Bakes pointiness using the 'Geometry' node's 'Pointiness' socket, with a 'Brightness/Contrast' node", 31),
    ('Displacement ', "Displacement ", "Bakes socket values of the 'Displacement' input socket node of the 'Material Output'", 32),

    ('', "Multires", "", -1),
    ('Normals', "Normals", "Bakes normal data from the object's 'Multiresolution' modifier", 33),
    ('Displacement', "Displacement", "Bakes displacement data from the object's 'Multiresolution' modifier", 34),

    ('', "Standard", "", -1),
    ('Combined', 'Combined', "Blender's default 'Combined' bake pass. Bakes all materials, textures, and lighting except specularity. The passes that contribute to the combined pass can be toggled individually to form the final map", 35),
    ('Ambient Occlusion ', "Ambient Occlusion ", "Blender's default 'Ambient Occlusion' bake pass. Bakes ambient occlusion as specified in the World panels. Ignores all lights in the scene", 36),
    ('Normal ', 'Normal ', "Blender's default 'Normal' bake pass. Bakes the normal pass of the materials", 37),
    ('Roughness ', 'Roughness ', "Blender's default 'Roughness' bake pass. Bakes the roughness pass of the materials", 38),
    ('Glossy', 'Glossy', "Blender's default 'Glossy' bake pass. Bakes the glossiness pass of the materials", 39),
    ('Position', 'Position', "Blender's default 'Position' bake pass. Bakes the mesh's position data", 40),

    ('', "*", "", -1),
    ('Shadow', "Shadow", "Blender's default 'Shadow' bake pass. Bakes shadows and lighting", 41),
    ('Diffuse', "Diffuse", "Blender's default 'Diffuse' bake pass. Bakes the diffuse pass of the materials", 42),
    ('UV', "UV", "Blender's default 'UV' bake pass. Mapped UV coordinates, used to represent where on a mesh a texture gets mapped to", 43),
    ('Transmission', "Transmission", "Blender's default 'Transmission' bake pass. Bakes the transmission pass of the materials", 44),
    ('Environment', "Environment", "Blender's default 'Environment' bake pass. Bakes the environment (i.e. the world surface shader defined for the scene) onto the selected object(s) as seen by rays cast from the world origin", 45),
    ('Emit', "Emit", "Blender's default 'Emit' bake pass. Bakes Emission, or the Glow color of materials", 46),

    ] if bpy.app.version >= (4, 0, 0) else [
    
    ('', "Shader", "", -1),
    ('Base Color', "Base Color", "Bakes input sockets named 'Base Color', 'Color', and 'Tint' for shader nodes", 1),
    ('Subsurface', "Subsurface", "Bakes input sockets named 'Subsurface' for shader nodes", 7),
    ('Subsurface Radius', "Subsurface Radius", "Bakes input sockets named 'Subsurface Radius' and 'Radius' for shader nodes", 8),
    ('Subsurface Color', "Subsurface Color", "Bakes input sockets named 'Subsurface Color' for shader nodes", 51),
    ('Subsurface IOR', "Subsurface IOR", "Bakes input sockets named 'Subsurface IOR' for shader nodes", 10),
    ('Subsurface Anisotropy', "Subsurface Anisotropy", "Bakes input sockets named 'Subsurface Anisotropy' and 'Anisotropy' for shader nodes", 11),
    ('Metallic', "Metallic", "Bakes input sockets named 'Metallic' for shader nodes", 2),
    ('Specular', "Specular", "Bakes input sockets named 'Specular' for shader nodes", 12),
    ('Specular Tint', "Specular Tint", "Bakes input sockets named 'Specular Tint' for shader nodes", 13),

    ('', "*", "", -1),
    ('Roughness', "Roughness", "Bakes input sockets named 'Roughness' for shader nodes", 3),
    ('Anisotropic', "Anisotropic", "Bakes input sockets named 'Anisotropic' for shader nodes", 14),
    ('Anisotropic Rotation', "Anisotropic Rotation", "Bakes input sockets named 'Anisotropic Rotation' for shader nodes", 15),
    ('Sheen', "Sheen", "Bakes input sockets named 'Sheen' for shader nodes", 23),
    ('Sheen Tint', 'Sheen Tint', "Bakes input sockets named 'Sheen Tint' for shader nodes", 25),
    ('Clearcoat', 'Clearcoat', "Bakes input sockets named 'Clearcoat' and 'Clear Coat' for shader nodes", 17),
    ('Clearcoat Roughness', 'Clearcoat Roughness', "Bakes input sockets named 'Clearcoat Roughness' and 'Clear Coat Roughness' for shader nodes", 18),
    ('IOR', "IOR", "Bakes input sockets named 'IOR' for shader nodes", 4),
    ('Transmission', "Transmission", "Bakes input sockets named 'Transmission' for shader nodes", 22),

    ('', "*", "", -1),
    ('Transmission Roughness', "Transmission Roughness", "Bakes input sockets named 'Transmission Roughness' for shader nodes", 50),
    ('Emission', "Emission", "Bakes input sockets named 'Emission', 'Emissive Color', and 'Emission Color' for shader nodes", 26),
    ('Emission Strength', "Emission Strength", "Bakes input sockets named 'Emission Strength' for shader nodes", 27),
    ('Alpha', "Alpha", "Bakes input sockets named 'Alpha' and 'Transparency' for shader nodes", 5),
    ('Normal', "Normal", "Bakes input sockets named 'Normal' for shader nodes", 6),
    ('Clearcoat Normal', 'Clearcoat Normal', "Bakes input sockets named 'Clearcoat Normal' and 'Clear Coat Normal' for shader nodes", 21),
    ('Tangent', "Tangent", "Bakes input sockets named 'Tangent' for shader nodes", 16),

    ('', "Miscellaneous", "", -1),
    ('Channel Packing', "Channel Packing", "Bakes up to 3 different BW bake types into a single RGB texture", 28),
    ('Color Attribute', "Color Attribute", "Bakes the set or available color attribute", 29),
    ('Ambient Occlusion', 'Ambient Occlusion', "Bakes ambient occlusion using the 'Ambient Occlusion' node", 30),
    ('Pointiness', 'Pointiness', "Bakes pointiness using the 'Geometry' node's 'Pointiness' socket, with a 'Brightness/Contrast' node", 31),
    ('Displacement ', "Displacement ", "Bakes socket values of the 'Displacement' input socket node of the 'Material Output'", 32),

    ('', "Multires", "", -1),
    ('Normals', "Normals", "Bakes normal data from the object's 'Multiresolution' modifier", 33),
    ('Displacement', "Displacement", "Bakes displacement data from the object's 'Multiresolution' modifier", 34),

    ('', "Standard", "", -1),
    ('Combined', 'Combined', "Blender's default 'Combined' bake pass. Bakes all materials, textures, and lighting except specularity. The passes that contribute to the combined pass can be toggled individually to form the final map", 35),
    ('Ambient Occlusion ', "Ambient Occlusion ", "Blender's default 'Ambient Occlusion' bake pass. Bakes ambient occlusion as specified in the World panels. Ignores all lights in the scene", 36),
    ('Normal ', 'Normal ', "Blender's default 'Normal' bake pass. Bakes the normal pass of the materials", 37),
    ('Roughness ', 'Roughness ', "Blender's default 'Roughness' bake pass. Bakes the roughness pass of the materials", 38),
    ('Glossy', 'Glossy', "Blender's default 'Glossy' bake pass. Bakes the glossiness pass of the materials", 39),
    ('Position', 'Position', "Blender's default 'Position' bake pass. Bakes the mesh's position data", 40),
    ('Shadow', "Shadow", "Blender's default 'Shadow' bake pass. Bakes shadows and lighting", 41),
    ('Diffuse', "Diffuse", "Blender's default 'Diffuse' bake pass. Bakes the diffuse pass of the materials", 42),
    ('UV', "UV", "Blender's default 'UV' bake pass. Mapped UV coordinates, used to represent where on a mesh a texture gets mapped to", 43),

    ('', "*", "", -1),
    ('Transmission', "Transmission", "Blender's default 'Transmission' bake pass. Bakes the transmission pass of the materials", 44),
    ('Environment', "Environment", "Blender's default 'Environment' bake pass. Bakes the environment (i.e. the world surface shader defined for the scene) onto the selected object(s) as seen by rays cast from the world origin", 45),
    ('Emit', "Emit", "Blender's default 'Emit' bake pass. Bakes Emission, or the Glow color of materials", 46)
    ]
    
channel_packing_items = [
    ('', "Shader", "", -1),
    ('Metallic', "Metallic", "Bakes input sockets named 'Metallic' for shader nodes", 2),
    ('Roughness', "Roughness", "Bakes input sockets named 'Roughness' for shader nodes", 3),
    ('IOR', "IOR", "Bakes input sockets named 'IOR' for shader nodes", 4),
    ('Alpha', "Alpha", "Bakes input sockets named 'Alpha' and 'Transparency' for shader nodes", 5),
    
    ('', "Subsurface", "", -1),
    ('Subsurface Weight', "Subsurface Weight", "Bakes input sockets named 'Subsurface Weight' for shader nodes", 7),
    ('Subsurface Scale', "Subsurface Scale", "Bakes input sockets named 'Subsurface Scale' for shader nodes", 9),
    ('Subsurface IOR', "Subsurface IOR", "Bakes input sockets named 'Subsurface IOR' for shader nodes", 10),
    ('Subsurface Anisotropy', "Subsurface Anisotropy", "Bakes input sockets named 'Subsurface Anisotropy' and 'Anisotropy' for shader nodes", 11),
    
    ('', "Specular", "", -1),
    ('Specular IOR Level', "Specular IOR Level", "Bakes input sockets named 'Specular IOR Level' for shader nodes", 12),
    ('Anisotropic', "Anisotropic", "Bakes input sockets named 'Anisotropic' for shader nodes", 14),
    ('Anisotropic Rotation', "Anisotropic Rotation", "Bakes input sockets named 'Anisotropic Rotation' for shader nodes", 15),
    
    ('', "Coat", "", -1),
    ('Coat Weight', 'Coat Weight', "Bakes input sockets named 'Coat Weight' and 'Clear Coat' for shader nodes", 17),
    ('Coat Roughness', 'Coat Roughness', "Bakes input sockets named 'Coat Roughness' and 'Clear Coat Roughness' for shader nodes", 18),
    ('Coat IOR', 'Coat IOR', "Bakes input sockets named 'Coat IOR' for shader nodes", 19),
    
    ('', "Sheen", "", -1),
    ('Sheen Weight', "Sheen Weight", "Bakes input sockets named 'Sheen Weight' for shader nodes", 23),
    ('Sheen Roughness', "Sheen Roughness", "Bakes input sockets named 'Sheen Roughness' for shader nodes", 24),
    
    ('', "*", "", -1),
    ('Transmission Weight', "Transmission Weight", "Bakes input sockets named 'Transmission Weight' for shader nodes", 22),
    ('Emission Strength', "Emission Strength", "Bakes input sockets named 'Emission Strength' for shader nodes", 27),
    
    ('', "Miscellaneous", "", -1),
    ('Ambient Occlusion', 'Ambient Occlusion', "Bakes ambient occlusion using the 'Ambient Occlusion' node", 30),
    ('Pointiness', 'Pointiness', "Bakes pointiness using the 'Geometry' node's 'Pointiness' socket, with a 'Brightness/Contrast' node", 31),

    ('', "None", "", -1),
    ('None', "None", "Nothing will be baked on this channel", 0)
    
    ] if bpy.app.version >= (4, 0, 0) else [

    ('', "Principled BSDF", "", -1),
    ('Metallic', "Metallic", "Bakes input sockets named 'Metallic' for shader nodes", 2),
    ('Roughness', "Roughness", "Bakes input sockets named 'Roughness' for shader nodes", 3),
    ('Alpha', "Alpha", "Bakes input sockets named 'Alpha' and 'Transparency' for shader nodes", 5),
    ('IOR', "IOR", "Bakes input sockets named 'IOR' for shader nodes", 4),
    ('Specular', "Specular", "Bakes input sockets named 'Specular' for shader nodes", 12),
    ('Specular Tint', "Specular Tint", "Bakes input sockets named 'Specular Tint' for shader nodes", 13),
    
    ('', "", "", -1), 
    ('Subsurface', "Subsurface", "Bakes input sockets named 'Subsurface' for shader nodes", 7),
    ('Subsurface IOR', "Subsurface IOR", "Bakes input sockets named 'Subsurface IOR' for shader nodes", 10),
    ('Subsurface Anisotropy', "Subsurface Anisotropy", "Bakes input sockets named 'Subsurface Anisotropy' and 'Anisotropy' for shader nodes", 11),
    ('Anisotropic', "Anisotropic", "Bakes input sockets named 'Anisotropic' for shader nodes", 14),
    ('Anisotropic Rotation', "Anisotropic Rotation", "Bakes input sockets named 'Anisotropic Rotation' for shader nodes", 15),
    ('Emission Strength', "Emission Strength", "Bakes input sockets named 'Emission Strength' for shader nodes", 27),
    
    ('', "", "", -1),
    ('Sheen', "Sheen", "Bakes input sockets named 'Sheen' for shader nodes", 23),
    ('Sheen Tint', 'Sheen Tint', "Bakes input sockets named 'Sheen Tint' for shader nodes", 25),
    ('Clearcoat', 'Clearcoat', "Bakes input sockets named 'Clearcoat' and 'Clear Coat' for shader nodes", 17),
    ('Clearcoat Roughness', 'Clearcoat Roughness', "Bakes input sockets named 'Clearcoat Roughness' and 'Clear Coat Roughness' for shader nodes", 18),
    ('Transmission', "Transmission", "Bakes input sockets named 'Transmission' for shader nodes", 22),
    ('Transmission Roughness', "Transmission Roughness", "Bakes input sockets named 'Transmission Roughness' for shader nodes", 50),

    ('', "Miscellaneous", "", -1),
    ('Ambient Occlusion', 'Ambient Occlusion', "Bakes ambient occlusion using the 'Ambient Occlusion' node", 30),
    ('Pointiness', 'Pointiness', "Bakes pointiness using the 'Geometry' node's 'Pointiness' socket, with a 'Brightness/Contrast' node", 31),

    ('', "None", "", -1),
    ('None', "None", "Nothing will be baked on this channel", 0)
    ]

type_aliases = {
    'Base Color': ['Color', 'Tint'],
    'Alpha': ['Transparency'],
    'Subsurface Anisotropy': ['Anisotropy'],
    'Subsurface Radius': ['Radius'],
    'Coat Weight': ['Clear Coat'],
    'Clearcoat': ['Clear Coat'],
    'Coat Normal': ['Clear Coat Normal'],
    'Clearcoat Normal': ['Clear Coat Normal'],
    'Coat Roughness': ['Clear Coat Roughness'],
    'Clearcoat Roughness': ['Clear Coat Roughness'],
    'Emission': ['Emission Color', 'Emissive Color'],
    'Emission Color': ['Emissive Color', 'Color'],
    'Emission Strength': ['Strength'],
    }

color_space = [
    ('', "Color Space", ""),
    ("ACES2065-1", "ACES2065-1", ""),
    ("ACEScg", "ACEScg", ""),
    ("AgX Base Display P3", "AgX Base Display P3", ""),
    ("AgX Base Rec.1886", "AgX Base Rec.1886", ""),
    ("AgX Base Rec.2020", "AgX Base Rec.2020", ""),
    ("AgX Base sRGB", "AgX Base sRGB", ""),
    ("AgX Log", "AgX Log", ""),
    ("Display P3", "Display P3", ""),
    ("Filmic Log", "Filmic Log", ""),
    ("Filmic sRGB", "Filmic sRGB", ""),
    ('', "", ""),
    ("Linear CIE-XYZ D65", "Linear CIE-XYZ D65", ""),
    ("Linear CIE-XYZ E", "Linear CIE-XYZ E", ""),
    ("Linear DCI-P3 D65", "Linear DCI-P3 D65", ""),
    ("Linear FilmLight E-Gamut", "Linear FilmLight E-Gamut", ""),
    ("Linear Rec.2020", "Linear Rec.2020", ""),
    ("Linear Rec.709", "Linear Rec.709", ""),
    ("Non-Color", "Non-Color", ""),
    ("Rec.1886", "Rec.1886", ""),
    ("Rec.2020", "Rec.2020", ""),
    ("sRGB", "sRGB", "")
    
    ] if bpy.app.version >= (4, 0, 0) else [

    ('', "Input Color Space", ""),
    ("XYZ", "XYZ", ""),
    ("sRGB", "sRGB", ""),
    ("Raw", "Raw", ""),
    ("Non-Color", "Non-Color", ""),
    ("Linear ACEScg", "Linear ACEScg", ""),
    ("Linear ACES", "Linear ACES", ""),
    ("Linear", "Linear", ""),
    ("Filmic sRGB", "Filmic sRGB", ""),
    ("Filmic Log", "Filmic Log", "")
    ]
    
    
#------------------------------------------------------------------------------------
#   Properties

class AUTOBAKE_PG_Collection(PropertyGroup):
    def update_file_format(self, context):
        context.scene.render.image_settings.file_format = self.ab_fileformat
    
    def reserved_char_check(self, context):
        if platform.system() == 'Windows' and self.ab_prefix != (re.sub('[{}]'.format(re.escape('<>:/\|?*')), '', self.ab_prefix)):
            self.ab_prefix = re.sub('[{}]'.format(re.escape('<>:/\|?*')), '', self.ab_prefix)
            
        if platform.system() == 'Windows' and self.ab_bridge != (re.sub('[{}]'.format(re.escape('<>:"/\|?*')), '', self.ab_bridge)):
            self.ab_bridge = re.sub('[{}]'.format(re.escape('<>:"/\|?*')), '', self.ab_bridge)
            
        if platform.system() == 'Windows' and self.ab_suffix != (re.sub('[{}]'.format(re.escape('<>:"/\|?*')), '', self.ab_suffix)):
            self.ab_suffix = re.sub('[{}]'.format(re.escape('<>:"/\|?*')), '', self.ab_suffix)
        
    def update_item_name(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        list = scene.autobake_udimlist if self.ab_udim_bake else scene.autobake_bakelist
        if len(list) > 0:
            type_names = {}
            for item in abp.ab_baketype_name_all.split(', '):
                type_names[item.split(':')[0]] = item.split(':')[1]
                
            active_type = scene.autobake_udimlist[scene.autobake_udimlist_index].Type if self.ab_udim_bake else scene.autobake_bakelist[scene.autobake_bakelist_index].Type
            if active_type not in type_names:
                type_names[active_type] = active_type.strip()
            abp.ab_baketype_name = type_names[active_type]
            
    def update_custom_type_name(self, context):
        scene = context.scene
        abp = scene.autobake_properties
                
        if platform.system() == 'Windows' and self.ab_baketype_name != (re.sub('[{}]'.format(re.escape('<>:"/\|?*')), '', self.ab_baketype_name)):
            self.ab_baketype_name = re.sub('[{}]'.format(re.escape('<>:"/\|?*')), '', self.ab_baketype_name)

        type_names = {}
        for item in abp.ab_baketype_name_all.split(', '):
            type_names[item.split(':')[0]] = item.split(':')[1]

        active_type = scene.autobake_udimlist[scene.autobake_udimlist_index].Type if self.ab_udim_bake else scene.autobake_bakelist[scene.autobake_bakelist_index].Type

        if len(self.ab_baketype_name.strip()) == 0:
            type_names[active_type] = active_type.strip()
            abp.ab_baketype_name = active_type.strip()
        else:
            type_names[active_type] = abp.ab_baketype_name

        abp.ab_baketype_name_all = ", ".join([f"{key}:{value}" for key, value in type_names.items()])

    def update_object_selection(self, context):
        scene = context.scene
        if self.ab_source_object_method == 'LIST' and not scene.autobake_sourceobject:
            scene.autobake_sourceobject.add()
        

# # # # # # # # # # # # # #
# Main

    ab_texture_export : BoolProperty(name="Export", description="Automatically export the baked texture(s) to the given file path", default = False, options = set())
    
    ab_bake_list_item_count : IntProperty(name='List Length', default=0, options = set())
    ab_udim_list_item_count : IntProperty(name='UDIM List Length', default=0, options = set())
    ab_udimtype_list_item_count : IntProperty(name='TIle list Length', default=0, options = set())

    ab_load_method : EnumProperty(default='Material', description='', options = set(), items=[
        ("Material", "Material", "Load items from the selected material"),
        ("Object", "Object", "Load items from all the materials of the selected object")])
    ab_load_linked_material : PointerProperty(name='Source Material', type=bpy.types.Material, options = set())
    ab_load_linked_object : PointerProperty(name='Source Object', type=bpy.types.Object, options = set())
    ab_import_tiles : PointerProperty(name='Import Tiles', type=bpy.types.Image, options = set())

    ab_udim_bake : BoolProperty(name="UDIM Bake", description="Switch between UDIM and default single texture bake", default = False, update=update_item_name, options = set())
    ab_udim_label : BoolProperty(default=False, description='Visibility of the UDIM Tile labels', options = set())


# # # # # # # # # # # # # #
# Bake List

# Item Details

    ab_multires_level : IntProperty(name='Level Viewport', default=0, min=-1, soft_max=10, options=set(), description="Level to set the Multires modifier's 'Level Viewport' value. Set -1 to leave it as it is. NOTE: Make sure the 'Level Viewport' and the 'Render' level is not the same, as this would result in an empty bake, since multires data is baked based on the difference between the two subdivision levels")

    ab_ao_use_normal : BoolProperty(name="Use Normal", description="Use shader normal data when baking Ambient Occlusion", default = False, options = set())
    ab_ao_local_only : EnumProperty(name="Local Only", default='FALSE', description="Disable objects for the texture bake. Could result a faster bake when there are a lot of objects in the scene", options = set(), items=[
        ('SELECTED', 'Selected', "Disable all the objects except the active and selected objects when 'Selected to Active' is enabled. If it's not enabled, only the active object will be the exception"),
        ('COLLECTION', 'Collection', "Disable all except those in the Active Object's collection"),
        ('FALSE', 'False', "Use all the enabled objects")])

    ab_ao_sample : IntProperty(name='Sample', default=16, min=1, max=128, description='Number of rays to trace per shader evaluation', options = set())
    ab_ao_sample_use : BoolProperty(name='Use Sample', default=True, description="Use the node's sample rate instead of the set sampling from the settings. If false, the bake type will use the 'High' sampling settings instead of the 'Low'", options = set())
    ab_ao_inside : BoolProperty(name='Inside', default=False, description='Trace rays towards the inside of the object', options = set())
    ab_ao_only_local : BoolProperty(name='Only Local', default=True, description='Only consider the object itself when computing ambient occlusion', options = set())
    ab_ao_distance : FloatProperty(name='Distance', default=1, min=0, max=1000, step=10, precision=3, description="Limit the ray's distance. Zero for no limit", options = set())
    
    ab_pointiness_contrast : FloatProperty(name='Contrast', default=5, min=-100, max=100, precision=3, step=10, description='Changes the contrast between the darkest and lightest parts of the pointiness map', options = set())
    ab_pointiness_brightness : FloatProperty(name='Brightness', default=0, min=-100, max=100, precision=3, step=10, description='Changes the overall brightness', options = set())
    
    ab_channel_pack_r : EnumProperty(name="Channel Packing R", items = channel_packing_items, description = "Select the bake type to use on the red channel for the 'Channel Packing' bake", default = "Metallic", options = set())
    ab_channel_pack_g : EnumProperty(name="Channel Packing G", items = channel_packing_items, description = "Select the bake type to use on the green channel for the 'Channel Packing' bake", default = "Roughness", options = set())
    ab_channel_pack_b : EnumProperty(name="Channel Packing B", items = channel_packing_items, description = "Select the bake type to use on the blue channel for the 'Channel Packing' bake", default = "Alpha", options = set())

    ab_uv_target : StringProperty(name='UV Map', default='UVMap', description='Preferred UV map coordinate to bake, useful for objects with multiple UV maps. When unspecified or the specified uv map is not found, the active (render) UV map will be used', options = set())
    ab_attribute_target : StringProperty(name='Color Attribute', default='Attribute', description='Preferred color attribute channel to bake, useful for objects with multiple color attributes. When unspecified or the specified color attribute is not found, the active (render) color attribute will be used', options = set())
    ab_displacement_source : BoolProperty(name='Source Only', default=False, description="Bake the source of the displacement/vector displacement node. This will be either the 'Height' or the 'Color' input socket value", options = set())


# # # # # # # # # # # # # #
# Image

# Name

    ab_prefix : StringProperty(name= "Prefix", description = "Custom prefix before the texture (type) name. Leave it empty to use the target object's name. For none use double apostrophes ('') or double quotation marks (\"\"). The prefix cannot contain reserved characters as defined by the operating system. The prefix will be ignored and the objects' names will be used when multiple objects are baked at once", update=reserved_char_check, options = set())
    ab_bridge : StringProperty(name= "Bridge", description = "Custom text between the texture type and scale. Leave it empty for none. The bridge cannot contain reserved characters as defined by the operating system", default=" - ", update=reserved_char_check, options = set())
    ab_suffix : StringProperty(name= "Suffix", description = "Custom suffix after the texture (scale) name. Leave it empty for none. The suffix cannot contain reserved characters as defined by the operating system", default="px", update=reserved_char_check, options = set())

    ab_baketype_name : StringProperty(name='Type Name', description="Customize the name of this bake type. Clear to reset its value", update=update_custom_type_name, options = set())
    ab_baketype_name_all : StringProperty(name='Type Name (All)', description="", default="x:x", options = set())

    ab_name_structure : StringProperty(name='Name Structure', description='Fully custumize your texture names', default="{prefix}{type}{bridge}{size}{suffix}", options = set())
    ab_name_structure_udim : StringProperty(name='Name Structure UDIM', description='Fully custumize your UDIM texture names', default="{prefix}{type}{bridge}{size}x", options = set())
    ab_name_structure_udim_export : StringProperty(name='Name Structure UDIM Export', description='Fully custumize your UDIM export texture names', default="{prefix}{type}{bridge}{size}x.{udim}", options = set())

# Format

    ab_floatbuffer: BoolProperty(name="Use Float", description="Create image with 32-bit floating-point bit depth", default = False, options = set())
    
# Export

    ab_subfolders : BoolProperty(name="Subfolders", description="Separate exported textures into different subolders based on bake type", default = False, options = set())
    ab_filepath : StringProperty(name= "Filepath", description = "File path to use when exporting the textures", options = set(), default = os.path.join(os.environ['USERPROFILE'], 'Desktop') if platform.system() == 'Windows' else os.path.join(os.environ['HOME'], 'Desktop') if platform.system() == 'Linux' else "")
    ab_fileformat : EnumProperty(name="File Format", description="File format to use when saving the image", default="PNG", update=update_file_format, options = set(), items=[
        ('', "Image File Formats", "", ),
        ('BMP', "BMP", "", ),
        ('IRIS', "Iris", "", ),
        ('PNG', "PNG", "", ),
        ('JPEG', "JPEG", "", ),
        ('JPEG2000', "JPEG 2000", "", ),
        ('TARGA', "Targa", "", ),
        ('TARGA_RAW', "Targa Raw", "", ),
        ('', "", "", ),
        ('CINEON', "Cineon", "", ),
        ('DPX', "DPX", "", ),
        ('OPEN_EXR_MULTILAYER', "OpenEXR MultiLayer", "", ),
        ('OPEN_EXR', "OpenEXR", "", ),
        ('HDR', "Radiance HDR", "", ),
        ('TIFF', "TIFF", "", ),
        ('WEBP', "WebP", "", )])

# Color Management
    ab_custom_color_management : BoolProperty(name='Custom Color Management', description="Use custom color management settings when exporting textures. Recommended to keep this option off for raw image exports. When it's off the following color management settings will be used:\n\u2022 Display Device: sRGB\n\u2022 View: Raw\n\u2022 Look: None\n\u2022 Exposure: 0\n\u2022 Gamma: 1\n\u2022 Use Curves: False", options = set())

# # # # # # # # # # # # # #
# Settings

# Margin
    
    ab_adaptive_margin : BoolProperty(name='Adaptive Margin', default=False, description="Automatically adjust the margin size based on the baked texture's scale. The currently set margin size will be used for textures with a resolution of 1024px, and will be correspondingly adjusted for different texture sizes", options = set())

# Selected to Active

    ab_active_as_final : BoolProperty(name='Active as Final', default=False, description='Use the active object as the final object', options = set())

# Selection Helper

    ab_target_object : PointerProperty(name='Target Object', type=bpy.types.Object, description = 'Low Poly Object', options = set())
    ab_source_object_method : EnumProperty(name='Source Object Method', default='COLLECTION', description='Choose your source object selection method', update=update_object_selection, options = set(), items=[
        ('COLLECTION', 'Collection', 'The collection containing your source objects'),
        ('LIST', 'List', 'Set up a list with your source objects')])
    ab_source_collection : PointerProperty(name='Source Collection', type=bpy.types.Collection, description='Collection that contains your high poly objects', options = set())

# Sampling

    ab_sampling_use_render : BoolProperty(name='Use Render Settings', default=False, description='Use the same sampling settings as for rendering images', options = set())
    ab_auto_pick_sampling : BoolProperty(name='Auto Select', default=True, options = set(), description="Automatically choose which sampling settings to use, with this you can optimize to only bake textures with high sampling settings that needs to be baked with. High sampling settings will be picked for: \n\u2022 Combined \n\u2022 Ambient Occlusion (Standard) \n\u2022 Glossy \n\u2022 Diffuse \n\u2022 Transmission \n\u2022 Shadow \n\u2022 Environment")

    ab_sampling_low_adaptive : BoolProperty(name='Adaptive Sampling', default=False, description='Automatically reduce the number of samples per pixel based on estimated noise level', options = set())
    ab_sampling_low_noise_threshold : FloatProperty(name='Noise Threshold', default=0.01, precision=4, min=0, max=1, description='Noise level step to stop sampling at, lower values reduce noise at the cost of render time. Zero for automatic setting based on number of AA sample', options = set())
    ab_sampling_low_max : IntProperty(name='Max Samples', default=1, min=1, max=16777216, description='Number of samples to render for each pixel', options = set())
    ab_sampling_low_min : IntProperty(name='Min Samples', default=0, min=0, max=4096, description='Minimum AA samples for adaptive sampling, to discover noisy features before stopping sampling. Zero for automatic setting based on noise threshold', options = set())
    ab_sampling_low_time_limit : FloatProperty(name='Time Limit', default=0, min=0, subtype='TIME_ABSOLUTE', description='Limit the render time (exclude synchronization time). Zero disables the limit', options = set())
    ab_sampling_low_denoise : BoolProperty(name='Denoise', default=False, description='Denoise the rendered image', options = set())
    ab_sampling_low_denoiser : EnumProperty(name='Denoiser', default='OPENIMAGEDENOISE', description='Denoise the image with the selected denoiser. For denoising the image after rendering:', options = set(), items=[
        ('OPENIMAGEDENOISE', 'OpenImageDenoise', 'Use Intel OpenImageDenoise AI denoiser running on the CPU'),
        ('OPTIX', 'OptiX', 'Use the OptiX AI denoiser with GPU acceleration, only available on NVIDIA GPUs')])
    ab_sampling_low_passes : EnumProperty(name='Passes', default='RGB_ALBEDO_NORMAL', description='Passes used by the denoiser to distinguish noise from shader and geometry detail', options = set(), items=[
        ('RGB', 'None', "Don't use utility passes for denoising",),
        ('RGB_ALBEDO', 'Albedo', 'Use albedo pass for denoising',),
        ('RGB_ALBEDO_NORMAL', 'Albedo and Normal', 'Use albedo and normal passes for denoising',)])
    ab_sampling_low_prefilter : EnumProperty(name='Prefilter', default='ACCURATE', description='Prefilter noise guiding (albedo and normal) passes to improve denoising quality when using OpenImageDenoiser', options = set(), items=[
        ('NONE', 'None', 'Noe prefiltering, use when guiding passes are noise-free'),
        ('FAST', 'Fast', 'Denoise color and guiding passes together. Improves quality when guiding passes are noisy using least amount of extra processing time'),
        ('ACCURATE', 'Accurate', 'Prefilter noise guaiding passes before denoising color. Improves quality when guiding passes are noisy using least amount of extra processing time')])
    
    ab_sampling_high_adaptive : BoolProperty(name='Adaptive Sampling', default=True, description='Automatically reduce the number of samples per pixel based on estimated noise level', options = set())
    ab_sampling_high_noise_threshold : FloatProperty(name='Noise Threshold', default=0.01, precision=4, min=0, max=1, description='Noise level step to stop sampling at, lower values reduce noise at the cost of render time. Zero for automatic setting based on number of AA sample', options = set())
    ab_sampling_high_max : IntProperty(name='Max Samples', default=4096, min=1, max=16777216, description='Number of samples to render for each pixel', options = set())
    ab_sampling_high_min : IntProperty(name='Min Samples', default=0, min=0, max=4096, description='Minimum AA samples for adaptive sampling, to discover noisy features before stopping sampling. Zero for automatic setting based on noise threshold', options = set())
    ab_sampling_high_time_limit : FloatProperty(name='Time Limit', default=0, min=0, subtype='TIME_ABSOLUTE', description='Limit the render time (exclude synchronization time). Zero disables the limit', options = set())
    ab_sampling_high_denoise : BoolProperty(name='Denoise', default=True, description='Denoise the rendered image', options = set())
    ab_sampling_high_denoiser : EnumProperty(name='Denoiser', default='OPENIMAGEDENOISE', description='Denoise the image with the selected denoiser. For denoising the image after rendering:', options = set(), items=[
        ('OPENIMAGEDENOISE', 'OpenImageDenoise', 'Use Intel OpenImageDenoise AI denoiser running on the CPU'),
        ('OPTIX', 'OptiX', 'Use the OptiX AI denoiser with GPU acceleration, only available on NVIDIA GPUs')])
    ab_sampling_high_passes : EnumProperty(name='Passes', default='RGB_ALBEDO_NORMAL', description='Passes used by the denoiser to distinguish noise from shader and geometry detail', options = set(), items=[
        ('RGB', 'None', "Don't use utility passes for denoising",),
        ('RGB_ALBEDO', 'Albedo', 'Use albedo pass for denoising',),
        ('RGB_ALBEDO_NORMAL', 'Albedo and Normal', 'Use albedo and normal passes for denoising',)])
    ab_sampling_high_prefilter : EnumProperty(name='Prefilter', default='ACCURATE', description='Prefilter noise guiding (albedo and normal) passes to improve denoising quality when using OpenImageDenoiser', options = set(), items=[
        ('NONE', 'None', 'Noe prefiltering, use when guiding passes are noise-free'),
        ('FAST', 'Fast', 'Denoise color and guiding passes together. Improves quality when guiding passes are noisy using least amount of extra processing time'),
        ('ACCURATE', 'Accurate', 'Prefilter noise guaiding passes before denoising color. Improves quality when guiding passes are noisy using least amount of extra processing time')])

# # # # # # # # # # # # # #
# Addon 

# Textures

    ab_textures_fakeuser : BoolProperty(name="Fake User", description="Save the image data-block even if it has no user", default=False, options = set())
    ab_pack_texture : BoolProperty(name='Pack Images', default=False, description='Pack the baked textures into the .blend file', options = set())
    ab_scaled_antialiasing : EnumProperty(name='Anti-aliasing', default='FALSE', description='Anti-aliasing method to use', options = set(), items=[
        ('FALSE', 'Off', 'No scaling will be applied for anti-alising', 'ALIASED', 0),
        ('UPSCALED', 'Upscaled', 'Textures will be created and baked at a higher resolution (original texture size * specified value), and then scaled back to the original texture size after the bake', 'ANTIALIASED', 1),
        ('DOWNSCALED', 'Downscaled', "After the baking process, textures are scaled to the specified percentage and then rescaled back to the original size. This operation can be repeated the set amount of times", 'ANTIALIASED', 2),])
    ab_antialiasing_downscaled : FloatProperty(name='Downscaled', default=75, precision=2, soft_min=25, soft_max=95, min=1, max=99.99, subtype='PERCENTAGE', description="Set the downscale percetange for the 'Downscaled' anti-aliasing", options = set())
    ab_antialiasing_repeat : IntProperty(name='Downscale Iteration', min=1, max=100, default=1, description='Repeat the rescale operation the set amount of time', options = set())
    ab_antialiasing_upscaled : FloatProperty(name='Upscaled', default=150, precision=2, min=100.1, soft_min=105, soft_max=800, max=10000, options = set(), description="Set the upscale percentage for the 'Upscaled' anti-aliasing. This value will be used to bake the texture at a higher resolution, resulting in longer bakes. 200% will result double the bake size and approximately double the bake time. After the bake the texture will be scaled back to its original size", subtype='PERCENTAGE')
    ab_color_space_float: EnumProperty(name='Float Image Color Space', default='Non-Color', items=color_space, options = set(), description='Set the default color space for the float image textures.\n'
                                                                                                                        "Affected Items: Metallic, Roughness, IOR, Alpha, Subsurface Weight, Subsurface Scale, Subsurface IOR, Subsurface Anisotropy, Specular IOR Level, Anisotropic, Anisotropic Rotation, Transmission Weight, Coat Weight, Coat Roughness, Coat IOR, Sheen Weight, Sheen Roughness, Emission Strength, Displacement (Multires), Roughness (Standard), Glossy, Shadow, Ambient Occlusion, Subsurface, Specular, Sheen, Clearcoat, Clearcoat Roughness, Transmission Roughness, Channel Packing")
    ab_color_space_color: EnumProperty(name='Color Image Color Space', default='sRGB', items=color_space, options = set(), description='Set the default color space for the color image textures.\n'
                                                                                                                        "Affected Items: Base Color, Specular Tint, Coat Tint, Sheen Tint, Emission Color, Color Attribute, Combined, Diffuse, Transmission, Environment, Emit, Subsurface Color, Emission")
    ab_color_space_vector: EnumProperty(name='Vector Image Color Space', default='Non-Color', items=color_space, options = set(), description='Set the default color space for the vector image textures.\n'
                                                                                                                        "Affected Items: Normal, Subsurface Radius, Tangent, Coat Normal, Normals, Normal (Standard), Position, UV, Displacement (Miscellaneous), Clearcoat Normal")
    ab_shared_textures : BoolProperty(name='Shared Textures', default=False, options = set(), description="Use this option when you want to bake each selected object onto the same textures. When shared texture is used with the final material option, only one final material will be created (and applied), and it will be named as the set prefix if set, or as the first object. Anti-aliasing will only be applied with the last baking object, likewise the textures exported. Please note that UV islands are not respected, so make sure you use a lower margin size to avoid texture bleeding. If this option is used with anti-aliasing, it will always be applied to the last object in the queue")

# Materials

    ab_final_material : BoolProperty(name='Final Material', description="Create a final material where the baked textures will be placed and kept regardless of what the 'Image Textures' property set to", default=False, options = set())
    ab_final_shader : EnumProperty(name='Final Shader', default='ShaderNodeBsdfPrincipled', description='Set what shader to use for the final material', options = set(), items=[
        ('', 'Shaders', ''),
        ('ShaderNodeBsdfPrincipled', 'Principled BSDF', ''),
        ('ShaderNodeBsdfDiffuse', 'Diffuse BSDF', ''),
        ('ShaderNodeEmission', 'Emission', ''),
        ('ShaderNodeBsdfTranslucent', 'Translucent BSDF', ''),
        ('ShaderNodeBsdfTransparent', 'Transparent BSDF', ''),
        ('ShaderNodeBsdfRefraction', 'Refraction BSDF', ''),
        ('ShaderNodeBsdfGlass', 'Glass BSDF', ''),
        ('ShaderNodeBsdfSheen', 'Sheen BSDF', ''),
        ('', '', ''),
        ('ShaderNodeBsdfToon', 'Toon BSDF', ''),
        ('ShaderNodeBsdfAnisotropic', 'Glossy BSDF', ''),
        ('ShaderNodeBsdfHair', 'Hair BSDF', ''),
        ('ShaderNodeBsdfHairPrincipled', 'Principled Hair BSDF', ''),
        ('ShaderNodeVolumeAbsorption', 'Volume Absorption', ''),
        ('ShaderNodeVolumeScatter', 'Volume Scatter', ''),
        ('ShaderNodeVolumePrincipled', 'Principled Volume', ''),
        ('ShaderNodeSubsurfaceScattering', 'Subsurface Scattering', '')
            ] if bpy.app.version >= (4, 0, 0) else [
        ('', 'Shaders', ''),
        ('ShaderNodeBsdfPrincipled', 'Principled BSDF', ''),
        ('ShaderNodeBsdfDiffuse', 'Diffuse BSDF', ''),
        ('ShaderNodeEmission', 'Emission', ''),
        ('ShaderNodeBsdfTranslucent', 'Translucent BSDF', ''),
        ('ShaderNodeBsdfTransparent', 'Transparent BSDF', ''),
        ('ShaderNodeBsdfRefraction', 'Refraction BSDF', ''),
        ('ShaderNodeBsdfGlass', 'Glass BSDF', ''),
        ('ShaderNodeBsdfVelvet', 'Velvet BSDF', ''),
        ('', '', ''),
        ('ShaderNodeBsdfToon', 'Toon BSDF', ''),
        ('ShaderNodeBsdfAnisotropic', 'Anisotropic BSDF', ''),
        ('ShaderNodeBsdfHair', 'Hair BSDF', ''),
        ('ShaderNodeBsdfHairPrincipled', 'Principled Hair BSDF', ''),
        ('ShaderNodeVolumeAbsorption', 'Volume Absorption', ''),
        ('ShaderNodeVolumeScatter', 'Volume Scatter', ''),
        ('ShaderNodeVolumePrincipled', 'Principled Volume', ''),
        ('ShaderNodeSubsurfaceScattering', 'Subsurface Scattering', '')
            ] if bpy.app.version >= (3, 6, 0) else [
        ('', 'Shaders', ''),
        ('ShaderNodeBsdfPrincipled', 'Principled BSDF', ''),
        ('ShaderNodeBsdfDiffuse', 'Diffuse BSDF', ''),
        ('ShaderNodeEmission', 'Emission', ''),
        ('ShaderNodeBsdfTranslucent', 'Translucent BSDF', ''),
        ('ShaderNodeBsdfTransparent', 'Transparent BSDF', ''),
        ('ShaderNodeBsdfRefraction', 'Refraction BSDF', ''),
        ('', '', ''),
        ('ShaderNodeBsdfGlass', 'Glass BSDF', ''),
        ('ShaderNodeBsdfAnisotropic', 'Glossy BSDF', ''),
        ('ShaderNodeVolumeAbsorption', 'Volume Absorption', ''),
        ('ShaderNodeVolumeScatter', 'Volume Scatter', ''),
        ('ShaderNodeVolumePrincipled', 'Principled Volume', ''),
        ('ShaderNodeSubsurfaceScattering', 'Subsurface Scattering', '')])
    
    ab_apply_textures : EnumProperty(name="Apply Textures", description="Select which baked textures should be connected (from each bake type) to the final material's shader. *Only works for bake types from the Shader and its sub-categories", default='Highest', options = set(), items=[
        ('False', 'False', "Don't connect the textures", 'X', 0),
        ('First', 'First', "Connect the first baked texture from the type", 'EVENT_A', 1),
        ('Last', 'Last', "Connect the last baked texture from the type", 'EVENT_Z', 2),
        ('Highest', 'Highest', "Connect the texture with the highest resolution from the type", 'SORT_DESC', 3),
        ('Lowest', 'Lowest', "Connect the texture with the lowest resolution from the type", 'SORT_ASC', 4)])
    ab_remove_imagetextures : BoolProperty(name="Keep Texture Nodes", description="Remove the 'Image Texture' nodes from the materials used in the bake process", default = False, options = set())
    ab_node_tiling : EnumProperty(name="Node Tiling", description="Choose how the 'Image Texture' nodes will be placed", default='GRID', options = set(), items=[
        ('OFF', 'Off', "All the 'Image Texture' nodes will be placed at the same location", 'X', 0),
        ('GRID', 'Grid', "The 'Image Texture' nodes will be placed in a grid-tile system", 'SNAP_VERTEX', 1),
        ('SINGLEROW', 'Single Row', "The 'Image Texture' nodes will be places into a single vertical row", 'DOT', 2),
        ('ROWBYTYPE', 'Row By Type', "Place every 'Image Texture' node into different rows by texture type", 'SNAP_EDGE', 3),
        ('ROWBYTYPECOMBINED', 'Row By Type +', "Place every 'Image Texture' node into different rows by texture type; single types are put into a combined row", 'SNAP_EDGE', 4)])
    ab_node_label : EnumProperty(name="Use Node Labels", default="Off", description="Add labels to the image texture nodes. This will not effect the texture's name", options = set(), items=[
        ('Off', 'Off', 'No label will be applied'),
        ('Name', 'Name', 'Use the custom type name as labels'),
        ('Type', 'Type', 'Use the type name as labels')])

# Objects
    
    ab_final_object : BoolProperty(name='Final Object', default=False, description='Duplicate the baked object and assign the final material to it (final material can only be assigned to the object if it is set to be created)', options = set())
    ab_final_collection : StringProperty(name='Collection', default='', description="Create a new collection with the set name for the final object(s). If left empty, no new collection will be created, and each final object will be located in the same collection as its original object", options = set())
    ab_final_collection_color : EnumProperty(name='Collection Color', default='NONE', description='Final object collection color', options = set(), items=[
        ('NONE', 'Collection', '', 'OUTLINER_COLLECTION', 0),
        ('COLOR_01', 'Collection', '', 'COLLECTION_COLOR_01', 1),
        ('COLOR_02', 'Collection', '', 'COLLECTION_COLOR_02', 2),
        ('COLOR_03', 'Collection', '', 'COLLECTION_COLOR_03', 3),
        ('COLOR_04', 'Collection', '', 'COLLECTION_COLOR_04', 4),
        ('COLOR_05', 'Collection', '', 'COLLECTION_COLOR_05', 5),
        ('COLOR_06', 'Collection', '', 'COLLECTION_COLOR_06', 6),
        ('COLOR_07', 'Collection', '', 'COLLECTION_COLOR_07', 7),
        ('COLOR_08', 'Collection', '', 'COLLECTION_COLOR_08', 8)])
    ab_object_location : EnumProperty(name='Object Location', default='Copy', items=[('Copy', 'Copy', 'Copy the location of the baked object', 'COPYDOWN', 0), ('Clear', 'Clear', 'Object will be placed to the world origin', 'LOOP_BACK', 1)], description='Location of the final object', options = set())
    ab_offset_direction : EnumProperty(options = set(), name='Offset Direction', default='Y', items=[("X", "X", ""), ("Y", "Y", ""), ("Z", "Z", "")], description='Direction to offset the final object')
    ab_object_offset : FloatProperty(options = set(), name='Offset', default=0.0, subtype='DISTANCE', description='Distance to offset the final object')
    ab_object_keep_name : EnumProperty(name='Keep Namme', description="Set which object should have the original name and which should have it with the differentiator. If the object is set to be exported, it will always be without the differentiator, regardless of this setting", default='Final', options = set(), items=[
        ('Final', 'Final', "The final object will take the name of the original object, and the differentiator will be added to the original object's name"),
        ('Original', 'Original', "Keep the name for the original object, and add the differentiator to the final object")])
    ab_object_differentiator : StringProperty(name='Differentiator', default='_', description="Differentiator for the selected object. This text will be added to the end of the object's name, this will help Auto Bake to avoid name conflicts for objects and materials", options = set())
    
# Export Object

    ab_export_object : BoolProperty(name='Export Object', default=False, description="Export the final object to the texture's location. (NOTE: OBJ and GLTF object export method uses viewport object selection)", options = set())
    ab_export_always : EnumProperty(name='Export When', default='Success', description='', options = set(), items=[
        ('Always', 'Always', 'Always export the object even if it has failed bake items'),
        ('Success', 'On-Success', "Only export when none of the object's bakes have failed (canceled items are not counted as failed items)")])
    ab_export_object_remove : BoolProperty(name='Remove Object', default=False, description='Remove the final object after export', options = set())
    ab_export_clear_location : BoolProperty(name='Clear Location', default=True, description='Export object with its location set to the world origin. After export the location will be set back to the original', options = set())
    ab_export_clear_rotation : BoolProperty(name='Clear Rotation', default=True, description='Export object with cleared rotation. After export the rotation will be set back to the original', options = set())
    ab_export_clear_scale : BoolProperty(name='Clear Scale', default=True, description='Export object with cleared scaling. After export the scale will be set back to the original', options = set())
    ab_export_restore_transform : BoolProperty(name='Restore Transform', default=True, description='Restore object transform after the export', options = set())
    
# Export Settings
    
    ab_export_object_as : EnumProperty(name="Export As", default="OBJ", description="Export objects with the selected file format. (When exporting, the exported objects will be selected in the 3D viewport, except during FBX export)", options = set(), items=[
        ('FBX', 'FBX (.fbx)', 'Write a FBX file'),
        ('GLTF', 'glTF 2.0 (.glb/.gltf)', 'Export the object as glTF 2.0 file'),
        ('OBJ', 'Wavefront (.obj)', 'Save the object to a Wavefront OBJ file.')])
    ab_export_pathmode : EnumProperty(name='Path Mode', default='AUTO', description='Method used to reference paths', options = set(), items=[
        ('AUTO', 'Auto', 'Use relative paths with subdirectories only'),
        ('ABSOLUTE', 'Absolute', 'Always write absolute paths'),
        ('RELATIVE', 'Relative', 'Always write relative paths (where possible)'),
        ('MATCH', 'Match', 'Match absolute/relative setting with input path'),
        ('STRIP', 'Strip Path', 'Filename only'),
        ('COPY', 'Copy', 'Copy the file to the destionation path (or subdirectory)')])
    ab_export_embedtextures : BoolProperty(name='Embed Textures', default=False, description='Embed textures in FBX binary file (only for "Copy" path mode)', options = set())
    ab_export_scale : FloatProperty(default=1, min=0.001, max=1000.0, description='Scale all data', name='', options = set())
    ab_export_applyunit : BoolProperty(name='Apply Unit', default=True, description='Take into account current Blender units settings (if unset, raw Blender Units values are used as-is)', options = set())
    ab_export_usespacetransform : BoolProperty(name='Use Space Transform', default=True, description='Apply global space transform to the object rotations. When disabled only the axis space is written to the file and all object transform are left as-is', options = set())
    ab_export_smoothing : EnumProperty(name='Smoothing', default='OFF', description="Export smoothing information (prefer 'Normals Only' option if your target importer understand split normals)", options = set(), items=[
        ('OFF', 'Normals Only', 'Export only normals instead of writing edge or face smoothing data'),
        ('FACE', 'Face', 'Write face smoothing'),
        ('EDGE', 'Edge', 'Write edge smoothing')])
    ab_export_subdivisionsurface : BoolProperty(name='Export Subdivision Surface', default=False, description="Export the last Catmull-Rom subdivision modifier as FBX subdivision (does not apply the modifier even if 'Apply Modifiers' is enabled)", options = set())
    ab_export_applymodifiers : BoolProperty(name='Apply Modifiers', default=True, description='Apply modifiers to mesh objects (except Armature ones) - WARNING: prevents exporting shape keys', options = set())
    ab_export_looseedges : BoolProperty(name='Loose Edges', default=False, description='Export loose edges (as two-vertices polygons)', options = set())
    ab_export_triangulatefaces : BoolProperty(name='Triangulate Faces', default=False, description='Convert all faces to triangles', options = set())
    ab_export_tangentspace : BoolProperty(name='Tangent Space', default=False, description='Add binormal and tangent vectors, together with normal they form the tangent space (will only work correctly with tris/quads only meshes)', options = set())
    ab_export_vertexcolors : EnumProperty(name='Vertex Colors', default='SRGB', description="Export vertex color attribute", options = set(), items=[
        ('NONE', 'None', 'Do not export colot attributes'),
        ('SRGB', 'sRGB', 'Export colors in sRGB color space'),
        ('LINEAR', 'Linear', 'Export colors in linear color space')])
    ab_export_prioritizeactivecolor : BoolProperty(name='Prioritize Active Color', default=False, description='Make sure active color will be exported first. Could be important since some other software can discard other color attributes besides the first one', options = set())
    ab_export_customprops : BoolProperty(name='Custom Properties', default=False, description='Export custom properties', options = set())
    ab_export_evalmode : EnumProperty(name='Properties', default='DAG_EVAL_VIEWPORT', description="Determines properties like object visibility, modifiers etc., where they differ for Render and Viewport", options = set(), items=[
        ('DAG_EVAL_RENDER', 'Render', 'Export objects as they appear in render'),
        ('DAG_EVAL_VIEWPORT', 'Viewport', 'Export objects as they appear in the viewport')])
    ab_export_triangulatedmesh : BoolProperty(name='Triangulated Mesh', default=False, description='All ngons with four or more vertices will be triangulated. Meshes in the scene will not be affected. Behaves like Triangulate Modifier with ngon-method: "Beauty", squad-method: "Shortest Diagonal", min vertices: 4', options = set())
    ab_export_exportcolors : BoolProperty(name='Vertex Colors', default=False, description='Export per-vertex colors', options = set())
    ab_export_vertexgroups : BoolProperty(name='Vertex Groups', default=False, description='Export the name of the vertex group of a face. It is approximated by choosing the vertex group with the most members among the vertices of a face', options = set())
    ab_export_smoothgroups : BoolProperty(name='Smooth Groups', default=False, description='Every smooth shaded-face is assigned group "1" and every flat-shaded face "off"', options = set())
    ab_export_groupbitflag : BoolProperty(name='Smooth Group Bitflags', default=False, description='Generate Bitflags for Smooth Groups', options = set())
    ab_export_pbrextension : BoolProperty(name='PBR Extensions', default=False, description='Export MTL library using PBR extensions (roughness, metallic, sheen, coat, anisotropy, transmission)', options = set())

# FBX Export

    ab_fbx_forward : EnumProperty(name='Forward', default='-Z', description='', options = set(), items=[
        ('X', 'X', ''),
        ('Y', 'Y', ''),
        ('Z', 'Z', ''),
        ('-X', '-X', ''),
        ('-Y', '-Y', ''),
        ('-Z', '-Z', '')])
    ab_fbx_up : EnumProperty(name='Up', default='Y', description='', options = set(), items=[
        ('X', 'X', ''),
        ('Y', 'Y', ''),
        ('Z', 'Z', ''),
        ('-X', '-X', ''),
        ('-Y', '-Y', ''),
        ('-Z', '-Z', '')])
    ab_fbx_apply_scaling : EnumProperty(name='', default='FBX_SCALE_NONE', description='How to apply custom and units scalings in generated FBX file (Blender uses FBX scale to detect units on import, but many other applications do not handle the same way)', options = set(), items=[
        ('FBX_SCALE_NONE', 'All Local', 'Apply custom scaling and units scaling to each object transormation, FBX scale remains at 1.0'),
        ('FBX_SCALE_UNITS', 'FBX Units Scale', 'Apply custom scaling to each object transformation, and units scaling to FBX scale'),
        ('FBX_SCALE_CUSTOM', 'FBX Custom Scale', 'Apply custom scaling to FBX scale, and units scaling to each object transformation'),
        ('FBX_SCALE_ALL', 'FBX All', 'Apply custom scaling and units scaling to FBX scale')])

# OBJ Export

    ab_obj_forward : EnumProperty(name='Forward Axis', default='NEGATIVE_Z', description='', options = set(), items=[
        ('X', 'X', ''),
        ('Y', 'Y', ''),
        ('Z', 'Z', ''),
        ('NEGATIVE_X', '-X', ''),
        ('NEGATIVE_Y', '-Y', ''),
        ('NEGATIVE_Z', '-Z', '')])
    ab_obj_up : EnumProperty(name='Up Axis', default='Y', description='', options = set(), items=[
        ('X', 'X', ''),
        ('Y', 'Y', ''),
        ('Z', 'Z', ''),
        ('NEGATIVE_X', '-X', ''),
        ('NEGATIVE_Y', '-Y', ''),
        ('NEGATIVE_Z', '-Z', '')])

# GLTF Export

    ab_gltf_format : EnumProperty(options = set(), name='Format', default='GLB', description='Output format', items=[
        ('GLB', 'glTF Binary (.glb)', 'Exports single file, with all data packed in binary form. Most efficient and protable, but more difficult to edit later'),
        ('GLTF_SEPARATE', 'glTF Separate (.gltf + .bin + textures)', 'Exports multiple files, with separate JSON, binary and texture data. Easiest to edit later')])
    ab_gltf_lighting : EnumProperty(options = set(), name='Lighting Mode', default='SPEC', description='Optional backwards compatibility for non-standard render engines. Applies to light', items=[
        ('SPEC', 'Standard', 'Physically-based glTF lighting units (cd, lx, nt)'),
        ('COMPAT', 'Unitless', 'Non-physical, unitless lighting. Useful when exposure controls are not available'),
        ('RAW', 'Raw (Deprecated)', 'Blender lighting strengths with no conversion')])
    ab_gltf_shape_keys : BoolProperty(options = set(), name='Shape Keys', default=False, description='Export shape keys (morph targets)')
    ab_gltf_shape_keys_normals : BoolProperty(options = set(), name='Shape Key Normals', default=True, description='Export vertex normals with shape keys (morph targets)')
    ab_gltf_shape_keys_tangents : BoolProperty(options = set(), name='Shape Key Tangents', default=False, description='Export vertex tangents with shape keys (morph targets)')
    ab_gltf_use_sparse : BoolProperty(options = set(), name='Use Sparse Accessor', default=True, description='Try using Sparse Accessor if it saves space')
    ab_gltf_omitting_sparse : BoolProperty(options = set(), name='Omitting Sparse Accessor', default=False, description='Omitting Sparse Accessor if data is empty')
    ab_gltf_compression : BoolProperty(options = set(), name='Compression', default=False, description='Compress mesh using Draco')
    ab_gltf_compression_level : IntProperty(options = set(), name='Compression Level', default=6, min=0, max=10,  description='')
    ab_gltf_compression_position : IntProperty(options = set(), name='Quantize Position', default=14, min=0, max=30,  description='Quantization bits for position values (0 = no quantization)')
    ab_gltf_compression_normal : IntProperty(options = set(), name='Normal', default=10, min=0, max=30,  description='Quantization bits for normal values (0 = no quantization)')
    ab_gltf_compression_texcoord : IntProperty(options = set(), name='Tex Coord', default=12, min=0, max=30,  description='Quantization bits for texture coordinate values (0 = no quantization)')
    ab_gltf_compression_color : IntProperty(options = set(), name='Color', default=10, min=0, max=30,  description='Quantization bits for color values (0 = no quantization)')
    ab_gltf_compression_generic : IntProperty(options = set(), name='Generic', default=12, min=0, max=30,  description='Quantization bits for generic values like weights or joints (0 = no quantization)')
    ab_gltf_yup : BoolProperty(options = set(), name='+Y Up', default=True, description='Export using glTF convention, +Y Up')
    ab_gltf_tangents : BoolProperty(options = set(), name='Tangents', default=False, description='Export vertex tangents with meshes')
    ab_gltf_vertex_colors : BoolProperty(options = set(), name='Vertex Colors', default=True, description='Export vertex colors with meshes')
    ab_gltf_attributes : BoolProperty(options = set(), name='Attributes', default=False, description='Export Attributes (when starting with underscore)')
    ab_gltf_loose_edges : BoolProperty(options = set(), name='Loose Edges', default=False, description='Export loose edges as lines, using the material from first material slot')
    ab_gltf_loose_points : BoolProperty(options = set(), name='Loose Points', default=False, description='Expaort loose points as glTF points, using the material from the first material slot')
    ab_gltf_keep_original : BoolProperty(options = set(), name='Same Location', default=False, description='Keep original textures files if possible. WARNING: if you use more than one texture, where pbr standard requires only one, only one texture will be used, This can lead to unexpected results')
    ab_gltf_texture_folder : StringProperty(options = set(), name='Image Folder', default='', description='Folder to place textures files in. Relative to the gltf. file')
    ab_gltf_copyright : StringProperty(options = set(), name='Copyright', default='', description='Legal rights and conditions for the model')
    ab_gltf_images : EnumProperty(options = set(), name='Images', default='AUTO', description='Output format for images', items=[
        ('AUTO', 'Automatic', ''),
        ('JPEG', 'JPEG Format (.jpg)', 'Save images as JPEGs'),
        ('WEBP', 'WebP Format', 'Save images as WebPs as main image (no fallback)'),
        ('NONE', 'None', "Don't export images")])
    ab_gltf_image_quality : IntProperty(options = set(), name='Image Quality', default=75, description='Quality of image export', min=0, max=100)
    ab_gltf_create_webp : BoolProperty(options = set(), name='Create WebP', default=False, description='Create WebP textures for every for every texture. For already WebP textures nothing happens')
    ab_gltf_webp_fallback : BoolProperty(options = set(), name='WebP fallback', default=False, description='For all WebP textures, create a PNG fallback texture')
    ab_gltf_original_spelucar : BoolProperty(options = set(), name='Original Specular', default=False, description='Export original glTF PBR Specular, instead of Blender Principled Shader Specular')

# Bake Items

    ab_new_item_method : EnumProperty(options = set(), name='New Item Method', default='Simple', description='Choose how the new items are added to the list', items=[
        ("Simple", "Single ", "Add a single new item to the list", 'ADD', 0),
        ("Advanced", "Multiple", "Use a panel where the new item(s) can be set. This panel can also be opened with 'CTRL + Click' on the 'Add New' button", 'ADD', 1)])
    ab_dynamic_scale: EnumProperty(options = set(), name="Dynamic Scale", description="Dynamically set the texture size for the new item based on the previous ones.", default='ENABLED', items=[
        ('FALSE', 'Disabled', "Scale by the active item", 'CHECKBOX_DEHLT', 0),
        ('ENABLED', 'Enabled', ("Scale by the previous two items.\n"
                                "\u2022 The size is doubled when the previous two are also doubled and the type is the same\n"
                                "\u2022 The size is divided when the previous two are also divided and the type is the same"), 'CHECKBOX_HLT', 1),
        ('ENABLED+', 'Enabled+', ("Scale by the previous two or three items.\n"
                                "\u2022 The size is doubled when the previous two are also doubled and the type is the same\n"
                                "\u2022 The size is divided when the previous two are also divided and the type is the same\n"
                                "\u2022 The size is doubled when the previous two items have the same size but different types and the third item's size is double of the second\n"
                                "\u2022 The size is divided when the previous two items have the same size but different types and the third item's size is half of the second"), 'CHECKBOX_HLT', 2)])

    ab_udim_item_tilescale_default : BoolProperty(options = set(), name="Tile Scale Default", description="Disable the default tile & scale item editor", default = True)
    ab_udim_item_default : BoolProperty(options = set(), name="UDIM Default", description="Disable the default type & multiplier editor", default = True)
    ab_udim_item_tile : BoolProperty(options = set(), name="UDIM Tile", description="Enable in-line tile edit", default = False)
    ab_udim_item_scale : BoolProperty(options = set(), name="UDIM Scale", description="Enable in-line tile scale edit", default = False)
    ab_udim_item_type : BoolProperty(options = set(), name="UDIM Type", description="Enable in-line type edit", default = True)
    ab_udim_item_multiplier : BoolProperty(options = set(), name="UDIM Multiplier", description="Enable in-line scale-multiplier edit", default = True)

    ab_item_details : BoolProperty(options = set(), name="Item Details", description="Enable/Disable the visibility of the item specific details", default=True)
    ab_item_type : BoolProperty(options = set(), name="Item Type", description="Enable in-line type edit", default = False)
    ab_item_scale : BoolProperty(options = set(), name="Item Scale", description="Enable in-line scale edit", default = False)
    ab_item_default : BoolProperty(options = set(), name="Item Default", description="Disable default item editor", default = True)

# Bake Queue

    ab_move_finished_bake : BoolProperty(options = set(), name="Finished Bake", description="Move the finished bake item to the bottom of the bake queue list", default = False)
    ab_move_active_bake : BoolProperty(options = set(), name="Active Bake", description="Move the active bake item to the top of the bake queue list", default = True)
    ab_auto_next : EnumProperty(options = set(), name='Auto Next Object', default='On-Success', description='Clear images from the export list', items=[
        ("On-Success", "On-Success", "Continue the bakes with the next object but only if no bakes failed", 'CHECKMARK', 0),
        ("On-Click", "On-Click", "Never continue with the next object, always wait for manual input", 'RESTRICT_SELECT_OFF', 1),
        ("Always", "Always", "Always continue with the next object, even ignoring failed bakes", 'PLAY', 2)])
    ab_auto_confirm : EnumProperty(options = set(), name='Auto Confirm Results', default='On-Click', description='Auto confirm bake results', items=[
        ("On-Success", "On-Success", "Auto-confirm bake results only if no bakes failed", 'CHECKMARK', 0),
        ("On-Click", "On-Click", "Never auto-confirm bake results, always wait for manual input", 'RESTRICT_SELECT_OFF', 1),
        ("Always", "Always", "Always auto-confirm bake results, even ignoring failed bakes", 'PLAY', 2)])

# Miscellaneous

    ab_confirm_start_bake : BoolProperty(options = set(), name='', default=True, description='Confirm window for bake start')
    ab_confirm_cancel_bake : BoolProperty(options = set(), name='', default=True, description='Confirm window for bake cancel')
    ab_confirm_bake_results : BoolProperty(options = set(), name='', default=True, description='Confirm window for bake results')
    ab_confirm_next_object : BoolProperty(options = set(), name='', default=True, description='Confirm window for next object bake')
    ab_confirm_queue_item_gate : BoolProperty(options = set(), name='', default=True, description='Confirm window for disabling and enabling items in the bake queue list')
    
    ab_start_popup_settings : BoolProperty(options = set(), name="Settings", default=True, description="Disable all the settings with this option")
    ab_start_popup_final_object : BoolProperty(options = set(), name="Final Object", default=True, description="")
    ab_start_popup_object_offset : BoolProperty(options = set(), name="Final Object Offset", default=True, description="")
    ab_start_popup_final_material : BoolProperty(options = set(), name="Final Material", default=True, description="")
    ab_start_popup_final_shader : BoolProperty(options = set(), name="Final Shader", default=True, description="")
    ab_start_popup_texture_apply : BoolProperty(options = set(), name="Texture Apply", default=True, description="")
    ab_start_popup_export_textures : BoolProperty(options = set(), name="Export Textures", default=True, description="")
    ab_start_popup_export_objects : BoolProperty(options = set(), name="Export Objects", default=True, description="")
    ab_start_popup_selected_to_active : BoolProperty(options = set(), name="Selected to Active", default=True, description="")
    ab_start_popup_keep_textures : BoolProperty(options = set(), name="Keep Textures", default=True, description="")
    
    ab_export_list : EnumProperty(options = set(), name='Clear Exports', default='Ask', description='Clear images from the export list', items=[
        ("Ask", "Ask", "Ask before bake starts", 'QUESTION', 0),
        ("Clear", "Clear", "Always remove the images from the list", 'TRASH', 1),
        ("Keep", "Keep", "Keep the images in the export list", 'GROUP', 2)])
        
    ab_list_padding : EnumProperty(options=set(), name='List Layout', default='Moderate', description='Choose the padding for the lists and their buttons', items=[
        ('Compact', 'Compact', 'No padding', 'SEQ_STRIP_DUPLICATE', 0),
        ('Moderate', 'Moderate', 'Moderate padding', 'SEQ_STRIP_DUPLICATE', 1),
        ('Spacious', 'Spacious', 'Maximum padding', 'SEQ_STRIP_DUPLICATE', 2)])
        
    ab_alert_texts : BoolProperty(options = set(), name='Show Alerts', default=True, description='Show text alerts below the alerted UI elements')
    ab_nodealert : BoolProperty(options = set(), name="Node Alert", description="Create a 'frame node' in the shader editor as a reminder not to delete, move, or modify nodes until the bake is completed", default = True)
    ab_bake_error_msg : BoolProperty(options = set(), name='Show Bake Failures', default=False, description='Show the error message in the list below the failed bake item')
    
    ab_report_requests : BoolProperty(options = set(), name="Start Requests", default=True, description="Report the collection of bake requests. \n\u2022 Auto Bake: Collecting objects to bake... \n\u2022 Auto Bake: ...2 objects are added to the object queue list.\n\u2022 Auto Bake: Collecting objects to bake... \n\u2022 Auto Bake: ...3 bake requests are added to the queue")
    ab_report_bake_start : BoolProperty(options = set(), name="Bake Start", default=True, description="Report when a texture's baking process is started. \n\u2022 Auto Bake: Texture 'Base Color - 512' is started baking")
    ab_report_bake_end : BoolProperty(options = set(), name="Bake End", default=True, description="Report when a texture's baking process is ended. \n\u2022 Auto Bake: Texture 'Base Color - 512' is finished baking")
    ab_report_object_start : BoolProperty(options = set(), name="Object Start", default=True, description="Report when an object starts its baking process. \n\u2022 Auto Bake: Object 'Cube' has been started baking")
    ab_report_object_end : BoolProperty(options = set(), name="Object End", default=True, description="Report when an object has completed all its baking items. \n\u2022 Auto Bake: Object 'Cube' is finished. Baked & Exported: 3; Canceled: 0; Failed: 0")
    ab_report_texture_export : BoolProperty(options = set(), name="Texture Export", default=True, description="Report when a texture is exported. \n\u2022 Auto Bake: 'Base Color - 512' successfully exported to: C:\Cube")
    ab_report_object_export : BoolProperty(options = set(), name="Object Export", default=True, description="Report when an object is exported. \n\u2022 Auto Bake: Object 'Cube' has been exported")
    ab_report_bake_summary : BoolProperty(options = set(), name="Bake Summary", default=True, description="Report a summary of bakes for the object. \n\u2022 Auto Bake: Bakes are finished for all (2) objects: Baked & Exported: 6; Canceled: 0; Failed: 0")
    ab_report_object_summary : BoolProperty(options = set(), name="Object Summary", default=True, description="Report a summary of the baked objects. \n\u2022 Auto Bake: Objects are finished baking. Baked & Exported: 2; Canceled: 0; Mixed: 0; Failed: 0")
    ab_report_bake_error : BoolProperty(options = set(), name="Bake Errors", default=True, description="Report when there is an issue with the bake. \n\u2022 Auto Bake: Target object 'Cube' has no active UV layer")

    ab_subfolder_use_prefix : BoolProperty(options = set(), default=False, description="Use prefix with bake type in subfolder names, otherwise only the bake type will be used", name='Use Prefix')
    
#------------------------------------------------------------------------------------
#   List Properties

class AUTOBAKE_PG_BakeQueue(PropertyGroup):
    Type : StringProperty(options = set(), name="", default="Unknown")
    Size : IntProperty(options = set(), name="", default=0)
    Multiplier : FloatProperty(options = set(), name='', default=1)
    Status : StringProperty(options = set(), name="", default="Pending")
    Cancel : BoolProperty(options = set(), name='', default=False)
    Enabled : BoolProperty(options = set(), name='', default=True)
    Icon : StringProperty(options = set(), name="", default="PREVIEW_RANGE")
    Error : StringProperty(options = set(), name='', default='')
    
    Image : PointerProperty(options = set(), type=bpy.types.Image)
    Prefix : StringProperty(options = set(), default='')
    Type_Name : StringProperty(options = set(), default='')
    

class AUTOBAKE_PG_Bake(PropertyGroup):
    def update_duplicates(self, context):
        for item in context.scene.autobake_bakelist:
            item.IsDuplicate = sum(1 for item2 in context.scene.autobake_bakelist if item.Type == item2.Type and item.Size == item2.Size) > 1
            item.name = f"{item.Type} {item.Size}" + (' Enabled' if item.Gate else ' Disabled') + (' Duplicate' if item.IsDuplicate else '')
                
    def update_gate(self, context):
        self.name = self.Type + " " + str(self.Size) + (' Enabled' if self.Gate else ' Disabled') + (' Duplicate' if self.IsDuplicate else '')
        context.scene.autobake_properties.ab_bake_list_item_count = sum(1 for item in context.scene.autobake_bakelist if item.Gate)
        
    def update_type(self, context):
        self.update_duplicates(context)
        
        scene = context.scene
        abp = scene.autobake_properties
        
        type_names = {}
        for item in abp.ab_baketype_name_all.split(', '):
            type_names[item.split(':')[0]] = item.split(':')[1]
            
        active_type = scene.autobake_bakelist[scene.autobake_bakelist_index].Type
        if active_type not in type_names:
            type_names[active_type] = active_type.strip()
            
        abp.ab_baketype_name = type_names[active_type]


    Type: EnumProperty(options = set(), name="", items=bake_items, default="Base Color", description="Texture Type", update=update_type)
    Size: IntProperty(options = set(), min=1, max=65536, default=512, name="", description="Texture scale", update=update_duplicates)
    Gate: BoolProperty(options = set(), name="", default=True, description="Enable item to Bake", update=update_gate)
    IsDuplicate: BoolProperty(options = set(), name="", default=False)


class AUTOBAKE_PG_UDIMType(PropertyGroup):
    def update_duplicates(self, context):
        for item in context.scene.autobake_udimlist:
            item.IsDuplicate = sum(1 for item2 in context.scene.autobake_udimlist if (item.Type == item2.Type and item.Size == item2.Size)) > 1
            item.name = f"{item.Type} {item.Size:.2f}" + (' Enabled' if item.Gate else ' Disabled') + (' Duplicated' if item.IsDuplicate else '')
            
    def update_gate(self, context):
        self.update_duplicates(context)
        context.scene.autobake_properties.ab_udimtype_list_item_count = sum(1 for item in context.scene.autobake_udimlist if item.Gate)

    def update_type(self, context):
        self.update_duplicates(context)
        
        scene = context.scene
        abp = scene.autobake_properties

        type_names = {}
        for item in abp.ab_baketype_name_all.split(', '):
            type_names[item.split(':')[0]] = item.split(':')[1]
            
        active_type = scene.autobake_udimlist[scene.autobake_udimlist_index].Type
        if active_type not in type_names:
            type_names[active_type] = active_type.strip()
            
        abp.ab_baketype_name = type_names[active_type]

        
    Type : EnumProperty(options = set(), name="", description="Bake Type", items=bake_items, default='Base Color', update=update_type)
    Size : FloatProperty(options = set(), name="", default=1, min=.01, soft_min=.1, max=100, soft_max=10, precision=2, step=10, description='UDIM Tile Size Multiplier', update=update_duplicates)
    Gate: BoolProperty(options = set(), name="", default=True, description="Enable item to Bake", update=update_gate)
    IsDuplicate: BoolProperty(options = set(), name="", default=False)


class AUTOBAKE_PG_UDIMTile(PropertyGroup):
    def update_name(self, context):
        self.name = str(self.UDIM)+' '+str(self.Size)+(' Enabled' if self.Gate else ' Disabled')+(' Duplicated' if self.IsDuplicate else '')
        
    def update_gate(self, context):
        self.update_name(context)
        context.scene.autobake_properties.ab_udim_list_item_count = sum(1 for item in context.scene.autobake_udimtilelist if item.Gate)
        
    def update_duplicates(self, context):
        self.update_name(context)
        scene = context.scene
        
        for item in scene.autobake_udimtilelist:
            item.IsDuplicate = sum(1 for item2 in scene.autobake_udimtilelist if item.UDIM == item2.UDIM) > 1
            item.name = str(item.UDIM)+' '+str(item.Size)+(' Enabled' if item.Gate else ' Disabled')+(' Duplicated' if item.IsDuplicate else '')
    
    def update_label(self, context):
        if self.Label == "":
            self.Label = "#Unknown"
            
    UDIM : IntProperty(options = set(), name="", default=1001, min=1001, max=2000, description="UDIM Tile Number", update=update_duplicates)
    Label : StringProperty(options = set(), name="UDIM Tile Label", default='#Unknown', update=update_label)
    Size : IntProperty(options = set(), name="", default=512, min=1, max=65536, description="UDIM Tile Size", update=update_name)
    Gate: BoolProperty(options = set(), name="", default=True, description="Enable UDIM Tile", update=update_gate)
    IsDuplicate: BoolProperty(options = set(), name="", default=False)
    
    
class AUTOBAKE_PG_SourceObjects(PropertyGroup):
    def object_update(self, context):
        scene = context.scene
        list = scene.autobake_sourceobject
        
        for index in range(len(list)-1, -1, -1):
            if list[index].Object == None or list[index].Object.type not in ['MESH', 'CURVE', 'SURFACE', 'META', 'FONT']:
                list.remove(index)
            elif sum(1 for item in list if (item.Object == list[index].Object)) > 1:
                list.remove(index)
        list.add()
   
    Object: PointerProperty(options = set(), type=bpy.types.Object, update=object_update)


class AUTOBAKE_PG_ImageExport(PropertyGroup):
    
    def update_name(self, context):
        if self.Image and self.Name != self.Image.name:
            if self.Name not in bpy.data.images:
                if self.Name == '':
                    self.Image.name = 'Auto Bake Image'
                else:
                    self.Image.name = self.Name
            
            self.Name = self.Image.name
            self.name = self.Name
            
    Name : StringProperty(options = set(), name="Export Name", update=update_name)
    Gate : BoolProperty(options = set(), default=True, description="Disable/Enable export for this image. If it's disabled at the export it will be removed from the list without exporting it", name='Toggle Export')
    Image : PointerProperty(options = set(), type=bpy.types.Image)
    Prefix : StringProperty(options = set(), )
    Type : StringProperty(options = set(), )
    Label : StringProperty(options = set(), )
        

class AUTOBAKE_PG_ObjectQueue(PropertyGroup):
    Object : PointerProperty(options = set(), name="Object", type=bpy.types.Object)
    Gate: BoolProperty(options = set(), name="", default=True, description="Enable object to bake")
    Status : StringProperty(options = set(), name="", default="Pending")
    Enabled : BoolProperty(options = set(), name='', default=True)
    Icon : StringProperty(options = set(), name="", default="PREVIEW_RANGE")
    Error : StringProperty(options = set(), name='', default='')
    Cancel : BoolProperty(options = set(), name='', default=False)


#------------------------------------------------------------------------------------
#   UI Lists

class AUTOBAKE_UL_Bake(UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        scene = context.scene
        abp = scene.autobake_properties
        layout.alert = item.IsDuplicate

        if abp.ab_item_type:
            row = layout.row()
            row.alignment = "LEFT"
            row.scale_x=.71
            row.label(text='', icon="TEXTURE_DATA")
            row_input=row.row()
            row_input.scale_x=1.3
            row_input.prop(item, "Type", emboss=False)     
        else:
            row = layout.row()
            row.alignment = "EXPAND"
            row.label(text=str(item.Type), icon="TEXTURE_DATA")
            
        if abp.ab_item_scale:
            row = layout.row()
            row.alignment = "RIGHT"
            row_size = row.row()
            row_size.scale_x = .75
            row_size.prop(item, "Size", text=' ', emboss=False, slider=True)
        else:
            row = layout.row()
            row.alignment = "RIGHT"
            row.label(text=str(item.Size)+' ')
            
        row = layout.row()
        row.scale_x=.9
        row.prop(item, "Gate")

        
class AUTOBAKE_UL_BakeQueue(UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        scene = context.scene
        layout.enabled = item.Enabled
        
        layout.alert = item.Status == 'Failed'
        col = layout.column(align=True)
        row_main = col.row()
        
        row = row_main.row()
        row.operator("autobake.toggle_queued", text="", icon= 'CHECKBOX_DEHLT' if item.Cancel else 'CHECKBOX_HLT', emboss=False).index = index
        row.label(text= f"{item.Type}  -  " + (f"{item.Multiplier:.2f}" if is_udim_bake else f"{item.Size}"))

        row_status = row_main.row()
        row_status.alignment = "RIGHT"
        row_status.label(text=item.Status)
        row_status.label(text="", icon=item.Icon)

        if scene.autobake_properties.ab_bake_error_msg and item.Status == 'Failed': 
            row_alert = col.row()
            row_alert.alert = True
            row_alert.label(text=f"{item.Error}")
    
    
class AUTOBAKE_UL_UDIMType(UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        scene = context.scene
        abp = scene.autobake_properties

        layout.alert = item.IsDuplicate
        row = layout.row()
        row.alignment = "LEFT"
        row.scale_x=.71
        row.label(text='', icon="TEXTURE_DATA")
        split = layout.split(factor=.7, align=True)
        row = split.row(align=True)
        row.alignment = "LEFT"
        row.scale_x=.975
        if abp.ab_udim_item_type:
            row.prop(item, "Type", emboss=False)     
        else:
            row.label(text='  '+str(item.Type))
        row = split.row(align=True)
        row.alignment = "RIGHT"
        if abp.ab_udim_item_multiplier:
            row.prop(item, "Size", text=' ', emboss=False, slider=True)
        else:
            row.label(text=f"{item.Size:.2f}"+' ')
        row = layout.row()
        row.scale_x=.9
        row.prop(item, "Gate")
        
        
class AUTOBAKE_UL_UDIMTile(UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        scene = context.scene
        abp = scene.autobake_properties
        
        layout.alert = item.IsDuplicate

        row = layout.row()
        row.alignment = "LEFT"
        row.scale_x=.71
        row.label(text='', icon="TEXTURE_DATA")
        
        split = layout.split(factor=.65 if abp.ab_udim_label else .55, align=True)

        row = split.row(align=True)
        row.alignment = 'LEFT'
        
        if abp.ab_udim_label:
            row.prop(item, 'Label', text='', emboss=False)
            
        if abp.ab_udim_item_tile:
            row.prop(item, "UDIM", slider=True, emboss=False)
        else:
            if not abp.ab_udim_label:
                row.label(text='      ')
            row.label(text=f'{item.UDIM}')
            
        row = split.row(align=True)
        row.alignment = "RIGHT"
        row.alert = False
        if abp.ab_udim_item_scale:
            row.prop(item, "Size", text=' ', slider=True, emboss=False)
        else:
            row.label(text=str(item.Size)+' ')
            
        row = layout.row(align=True)
        row.scale_x=.9
        row.prop(item, "Gate")

        
class AUTOBAKE_UL_SourceObjects(UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        scene = context.scene
        abp = scene.autobake_properties
        
        layout.prop(item, 'Object', text='')


class AUTOBAKE_UL_ImageExport(UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        scene = context.scene
        abp = scene.autobake_properties
        
        layout.enabled = item.Image.is_dirty
        
        split = layout.split(factor=.333, align=True)
        split.prop(item, 'Prefix', text='', icon='FILE_FOLDER', emboss=False)
        
        row = split.row(align=True)
        row.prop(item, 'Name', text='', icon='TEXTURE', emboss=False)
        
        row2 = row.row()
        if layout.active:
            row2.prop(item, 'Gate', text='')
        else:
            row2.label(text='', icon='ERROR')
        

class AUTOBAKE_UL_ObjectQueue(UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        scene = context.scene
        layout.enabled = item.Enabled
        
        layout.alert = item.Status in ['Failed', 'Mixed']
        col = layout.column(align=True)
        row_main = col.row()
        
        row = row_main.row()
        row.operator("autobake.toggle_queued_object", text="", icon= 'CHECKBOX_DEHLT' if item.Cancel else 'CHECKBOX_HLT', emboss=False).index = index
        row.label(text= f"{item.Object.name}", icon='OBJECT_DATAMODE')

        row_status = row_main.row()
        row_status.alignment = "RIGHT"
        row_status.label(text=item.Status)
        row_status.label(text="", icon=item.Icon)

        if scene.autobake_properties.ab_bake_error_msg and item.Status in ['Failed', 'Mixed']: 
            row_alert = col.row()
            row_alert.alert = True
            row_alert.label(text=f"{item.Error}")


#------------------------------------------------------------------------------------
#   Functions (Global)
            
def obj_bake_ready(self, context, target_obj, selected_obj):
    scene = context.scene
    bake = scene.render.bake
    abp = scene.autobake_properties
    
# Is Mesh
    if target_obj.type != "MESH":
        if abp.ab_report_bake_error:
            self.report({'ERROR'}, f"Auto Bake: Target object '{target_obj.name}' is a {target_obj.type} not a mesh.")
        return 'Object is not a mesh!'
    
# Has UV
    if not target_obj.data.uv_layers:
        if abp.ab_report_bake_error:
            self.report({'ERROR'}, f"Auto Bake: Target object '{target_obj.name}' has no active UV layer.")
        return 'Object has no UV Map!'
    
# Is Visible
    for obj in selected_obj:
        if obj.hide_render or obj.hide_viewport:
            if abp.ab_report_bake_error:
                self.report({'ERROR'}, f"Auto Bake: Object '{obj.name}' must be visible in the viewport and must be enabled for renders.")
            return 'Object is disabled in the viewport!'

# Filepath
    if abp.ab_texture_export and bool(os.path.exists(abp.ab_filepath)) == False:
        if abp.ab_report_bake_error:
            self.report({'ERROR'}, f"Auto Bake: Filepath '{abp.ab_filepath}' is invalid.")
        return 'Filepath is invalid!'
        
# Material
    if not any(slot.material is not None and slot.material.use_nodes for slot in target_obj.material_slots):
        if bake.use_selected_to_active:
            new_mat = bpy.data.materials.new(name=target_obj.name)
            if new_mat.name != target_obj.name:
                switch_name = str(new_mat.name)
                new_mat.name = ''
                bpy.data.materials[target_obj.name].name = switch_name
                new_mat.name = target_obj.name
            target_obj.data.materials.append(new_mat)
            new_mat.use_nodes = True
            if abp.ab_report_bake_error:
                self.report({'INFO'}, f"Auto Bake: Target object '{target_obj.name}' had no bake compatible material, '{new_mat.name}' was created and added to '{target_obj.name}'")
        else:
            if abp.ab_report_bake_error:
                self.report({'ERROR'}, f"Auto Bake: Target object '{target_obj.name}' must have a material to bake from with 'Use Nodes' enabled.")
            return 'Object must have a material!'
            
# Selected to Active
    if bake.use_selected_to_active:
        
    # Has Cage
        if bake.use_cage and bake.cage_object is None:
            if abp.ab_report_bake_error:
                self.report({'ERROR'}, "Auto Bake: Must have a set 'Cage Object' if 'Cage' is 'True'.")
            return "Must have set a 'Cage Object' if 'Cage' is 'True'"
    
    return True
                  
            
def export_texture(self, context, img, export_name, type, label, prefix):
    scene = context.scene
    abp = scene.autobake_properties
    image_settings = scene.render.image_settings

    if prefix == '':
        prefix = "Auto Bake Export"

# Folder
    foldername = re.sub('[{}]'.format(re.escape('<>:"/\|?*')), '', prefix)
    foldername = foldername.strip()
    mainpath = os.path.join(abp.ab_filepath, foldername)
    
    if not os.path.isdir(mainpath):
        os.makedirs(mainpath)
    
    if not abp.ab_subfolders:
        ab_filepath = mainpath
    else:
        if abp.ab_subfolder_use_prefix:
            subpath = os.path.join(mainpath, (foldername+'.'+type))
        else:
            subpath = os.path.join(mainpath, (type))
            
        if os.path.isdir(subpath):
            ab_filepath = subpath
        else:
            os.makedirs(subpath)
            ab_filepath = subpath

# Color Mode
    image_settings.color_mode = 'BW' if label in ['Float', 'AO', 'Ambient Occlusion', 'Pointiness'] else 'RGB'
        
# Color Override
    if not abp.ab_custom_color_management:
        stored_color_management = {
            'Color Management': scene.render.image_settings.color_management,
            'Display Device': scene.render.image_settings.display_settings.display_device,
            'View': scene.render.image_settings.view_settings.view_transform,
            'Look': scene.render.image_settings.view_settings.look,
            'Exposure': scene.render.image_settings.view_settings.exposure,
            'Gamma': scene.render.image_settings.view_settings.gamma,
            'Use Curves': scene.render.image_settings.view_settings.use_curve_mapping}
            
        scene.render.image_settings.color_management = 'OVERRIDE'
        scene.render.image_settings.display_settings.display_device = 'sRGB'
        scene.render.image_settings.view_settings.view_transform = 'Raw'
        scene.render.image_settings.view_settings.look = 'None'
        scene.render.image_settings.view_settings.exposure = 0
        scene.render.image_settings.view_settings.gamma = 1
        scene.render.image_settings.view_settings.use_curve_mapping = False

# File Format
    image_settings.file_format = abp.ab_fileformat
    
    file_extensions = {'BMP': 'bmp',
                       'IRIS': 'rgb',
                       'PNG': 'png',
                       'JPEG': 'jpg',
                       'JPEG2000': 'jp2' if image_settings.jpeg2k_codec == "JP2" else 'j2c',
                       'TARGA': 'tga',
                       'TARGA_RAW': 'tga',
                       'CINEON': 'cin',
                       'DPX': 'dpx',
                       'OPEN_EXR_MULTILAYER': 'exr',
                       'OPEN_EXR': 'exr',
                       'HDR': 'hdr',
                       'TIFF': 'tif',
                       'WEBP': 'webp'}
    file_extension = file_extensions[abp.ab_fileformat]

# Export
    img.filepath_raw = os.path.join(ab_filepath, f"{export_name}.{file_extension}")
    img.save_render(filepath = img.filepath_raw)

# Info
    if abp.ab_report_texture_export:
        self.report({'INFO'}, f"Auto Bake: '{img.name}' successfully exported to: {ab_filepath}")
        
# Restore Color Override
    if not abp.ab_custom_color_management:
        scene.render.image_settings.color_management = stored_color_management['Color Management']
        scene.render.image_settings.display_settings.display_device = stored_color_management['Display Device']
        scene.render.image_settings.view_settings.view_transform = stored_color_management['View']
        scene.render.image_settings.view_settings.look = stored_color_management['Look']
        scene.render.image_settings.view_settings.exposure = stored_color_management['Exposure']
        scene.render.image_settings.view_settings.gamma = stored_color_management['Gamma']
        scene.render.image_settings.view_settings.use_curve_mapping = stored_color_management['Use Curves']

    
def remove_handlers_timers():
    if bpy.app.timers.is_registered(restore_nodes):
        bpy.app.timers.unregister(restore_nodes)
        
    for handler in bpy.app.handlers.object_bake_cancel:
        if handler.__name__ == "ab_bake_cancel":
            bpy.app.handlers.object_bake_cancel.remove(handler)
            break
    for handler in bpy.app.handlers.object_bake_complete:
        if handler.__name__ == "ab_bake_complete":
            bpy.app.handlers.object_bake_complete.remove(handler)
            break
        

def restore_nodes():
    global delete_nodes
    global reconnect_nodes
    
    for item in reversed(reconnect_nodes):
        item[0].links.new(item[1], item[2])
        reconnect_nodes.remove(item)

    for item in reversed(delete_nodes):
        if (node := item[0].nodes.get(item[1].name, None)) is not None:
            item[0].nodes.remove(item[1])
        delete_nodes.remove(item)
        

def name_by_index(self, context):
    scene = context.scene
    abp = scene.autobake_properties
    
    list = scene.autobake_udimlist if abp.ab_udim_bake else scene.autobake_bakelist
    
    if len(list):
        index = scene.autobake_udimlist_index if abp.ab_udim_bake else scene.autobake_bakelist_index
    
        type_names = {}
        for item in abp.ab_baketype_name_all.split(', '):
            type_names[item.split(':')[0]] = item.split(':')[1]
            
        active_type = list[index].Type
        
        if active_type not in type_names:
            type_names[active_type] = active_type.strip()
        abp.ab_baketype_name = type_names[active_type]
        
        
def bake_result(self, context):
    scene = context.scene
    abp = scene.autobake_properties
    
    if len(scene.autobake_objectqueuelist) > 0:
        obj = scene.autobake_objectqueuelist[scene.autobake_objectqueuelist_index].Object
        
        if obj.name in context.scene.objects:
            for selected in context.selected_objects:
                selected.select_set(False)

            obj.select_set(True)
            context.view_layer.objects.active = obj
    
        if bake_status == 'IDLE' and len(scene.autobake_queuelist) > 0 and obj.name in bake_results:
            obj_bakes = bake_results[obj.name]
            stored_values = {key: value for dictionary in obj_bakes for key, value in dictionary.items()}
            
            for item in scene.autobake_queuelist:
                queue_item = f"{item.Type} " + (f"{item.Multiplier:.2f}" if is_udim_bake else f"{item.Size}")
                item.Status = stored_values[queue_item]['Status']
                item.Icon = stored_values[queue_item]['Icon']
                item.Error = stored_values[queue_item]['Error']
                item.Cancel = stored_values[queue_item]['Cancel']


#------------------------------------------------------------------------------------
#   Operators

class AUTOBAKE_OT_LoadLinked(Operator):
    bl_idname = "autobake.bakelist_load_linked"
    bl_label = "Load by Linked"
    bl_description = "Add new items to the list based on the connected shader sockets"
    bl_options = {'INTERNAL', 'UNDO'}

    reset_source : BoolProperty(options = set(), default=True, description='Remove the selected source (material or object), and auto set a new one the next time')
    texture_size : IntProperty(options = set(), default=512, min=1, max=65536, name='Texture Scale', description='The texture scale for the new items')
    type_size : FloatProperty(options = set(), default=1, min=0.1, max=10.0, precision=2, step=1, name='Type Scale', description='The tile-scale multiplier')
    place_item : EnumProperty(options = set(), default="After", name="", description="Select how to place the new items in the list", items=[
                    ("Before", "Before", "New items will be placed before the currently selected item"),
                    ("After", "After", "New items will be placed after the currently selected item"),
                    ("First", "First", "New items will be placed at the top of the list"),
                    ("Last", "Last", "New items will be placed at the bottom of the list")])
    
    
    def invoke(self, context, event):
        scene = context.scene
        abp = scene.autobake_properties
        
        if self.reset_source:
            abp.ab_load_linked_material = None
            if context.active_object:
                for slot in context.active_object.material_slots:
                    if slot.material is not None and slot.material.use_nodes:
                        abp.ab_load_linked_material = slot.material
                        break
            abp.ab_load_linked_object = None
            if context.active_object:
                abp.ab_load_linked_object = context.active_object
                    
        return context.window_manager.invoke_props_dialog(self, width=250)

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
        
        box = layout.box()
        col = box.column_flow(columns=1, align=True)
        
        split = col.split(factor=.4)
        split.label(text='Source Method')
        
        col2 = split.column()
        col2.row().prop(abp, 'ab_load_method', text=' ', expand=True)
        
        if abp.ab_load_method == 'Material':
            col2.prop(abp, 'ab_load_linked_material', text='')
        else:
            col2.prop(abp, 'ab_load_linked_object', text='')
        
        col2.prop(self, 'reset_source', text='Reset')
        
        split = col.split(factor=.4)
        if abp.ab_udim_bake:
            split.label(text='Type Scale')
            split.prop(self, 'type_size', text='')
        else:
            split.label(text='Texture Scale')
            split.prop(self, 'texture_size', text='')
            
        split = col.split(factor=.4)
        split.label(text="Placement")
        split.prop(self, "place_item")

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties

    # Collecting Default Node Trees
        node_trees = []
        
        if abp.ab_load_method == 'Material':
            if abp.ab_load_linked_material is None:
                self.report({'ERROR'}, "Auto Bake: Must select a source material!")
                return {'FINISHED'}
            else:
                if abp.ab_load_linked_material.use_nodes:
                    if abp.ab_load_linked_material.node_tree not in node_trees:
                        node_trees.append(abp.ab_load_linked_material.node_tree)
        
        elif abp.ab_load_method == 'Object':
            if abp.ab_load_linked_object is None:
                self.report({'ERROR'}, "Auto Bake: Must select a source object!")
                return {'FINISHED'}
            else:
                for slot in abp.ab_load_linked_object.material_slots:
                    if slot.material.use_nodes:
                        
                        if slot.material.node_tree not in node_trees:
                            node_trees.append(slot.material.node_tree)
                    
    # Branching
        shaders = []
        
        for tree in node_trees:
            branches = []
            for node in tree.nodes:
                
            # Looking for: Branch Start
                if node.type in ['OUTPUT_MATERIAL', 'GROUP_OUTPUT'] and node.is_active_output:
                    if node.inputs[0].is_linked:
                        if node.inputs[0].links[0].from_node.type in ['BSDF_PRINCIPLED', 'EMISSION', 'BSDF_TRANSLUCENT', 'BSDF_TRANSPARENT', 'BSDF_REFRACTION', 'BSDF_DIFFUSE', 'BSDF_GLASS', 'BSDF_SHEEN', 'BSDF_HAIR', 'BSDF_TOON', 'BSDF_GLOSSY', 'VOLUME_ABSORPTION', 'VOLUME_SCATTER', 'SUBSURFACE_SCATTERING', 'PRINCIPLED_VOLUME', 'BSDF_HAIR_PRINCIPLED', 'BSDF_VELVET', 'BSDF_ANISOTROPIC', 'EEVEE_SPECULAR']:
                            shaders.append(node.inputs[0].links[0].from_node)
                            
                        elif node.inputs[0].links[0].from_node.type == 'GROUP':
                            if node.inputs[0].links[0].from_node.node_tree not in node_trees:
                                node_trees.append(node.inputs[0].links[0].from_node.node_tree)
                            
                            if node.inputs[0].links[0].from_node not in branches:
                                branches.append(node.inputs[0].links[0].from_node)
                        else:
                            if node.inputs[0].links[0].from_node not in branches:
                                branches.append(node.inputs[0].links[0].from_node)
                    break
                
            # Looking for: New Branches & New Trees & Shader nodes
            for node in branches:
                for input in node.inputs:
                    if input.is_linked:
                        if input.links[0].from_node.type in ['BSDF_PRINCIPLED', 'EMISSION', 'BSDF_TRANSLUCENT', 'BSDF_TRANSPARENT', 'BSDF_REFRACTION', 'BSDF_DIFFUSE', 'BSDF_GLASS', 'BSDF_SHEEN', 'BSDF_HAIR', 'BSDF_TOON', 'BSDF_GLOSSY', 'VOLUME_ABSORPTION', 'VOLUME_SCATTER', 'SUBSURFACE_SCATTERING', 'PRINCIPLED_VOLUME', 'BSDF_HAIR_PRINCIPLED', 'BSDF_VELVET', 'BSDF_ANISOTROPIC', 'EEVEE_SPECULAR']:
                            shaders.append(input.links[0].from_node)
                            
                        elif input.links[0].from_node.type == 'GROUP':
                            if input.links[0].from_node.node_tree not in node_trees:
                                node_trees.append(input.links[0].from_node.node_tree)
                            if input.links[0].from_node not in branches:
                                branches.append(input.links[0].from_node)
                        else:
                            if input.links[0].from_node not in branches:
                                branches.append(input.links[0].from_node)
                           
   # Saved as Linked
        linked_sockets = []

        for shader in shaders:
            for input in shader.inputs:
                if input.is_linked:
                    input_aliases = [input.name,]
                    
                    socket_name = [key for key, aliases in type_aliases.items() if input.name in aliases]
                    
                    if socket_name != []:
                        for key in socket_name:
                            input_aliases.extend(type_aliases[key])
                            input_aliases.append(key)

                    for type in bake_items:
                        if type[0] in input_aliases and type[0] not in linked_sockets:
                            linked_sockets.append(type[0])
                    
        list = scene.autobake_udimlist if abp.ab_udim_bake else scene.autobake_bakelist
        item = None
        
        for socket in linked_sockets:
            item = list.add()
            item.Type = socket
            item.Size = self.type_size if abp.ab_udim_bake else self.texture_size
        
            if self.place_item == 'First':
                list.move(len(list)-1, 0)
                
            elif self.place_item == 'Before':
                if abp.ab_udim_bake:
                    list.move(len(list)-1, scene.autobake_udimlist_index)
                else:
                    list.move(len(list)-1, scene.autobake_bakelist_index)
                
            elif self.place_item == 'After':
                if abp.ab_udim_bake:
                    list.move(len(list)-1, scene.autobake_udimlist_index+1)
                else:
                    list.move(len(list)-1, scene.autobake_bakelist_index+1)
                
    # List Index
        if self.place_item == 'First':
            if abp.ab_udim_bake:
                scene.autobake_udimlist_index = min(len(linked_sockets)-1, len(list)-1)
            else:
                scene.autobake_bakelist_index = min(len(linked_sockets)-1, len(list)-1)

        elif self.place_item == 'Last':
            if abp.ab_udim_bake:
                scene.autobake_udimlist_index = len(list)-1
            else:
                scene.autobake_bakelist_index = len(list)-1
            
        elif self.place_item == 'Before':
            if abp.ab_udim_bake:
                scene.autobake_udimlist_index = max(scene.autobake_udimlist_index - 1 + len(linked_sockets), 0)
            else:
                scene.autobake_bakelist_index = max(scene.autobake_bakelist_index - 1 + len(linked_sockets), 0)
            
        elif self.place_item == 'After':
            if abp.ab_udim_bake:
                scene.autobake_udimlist_index = min(scene.autobake_udimlist_index + len(linked_sockets), len(list)-1)
            else:
                scene.autobake_bakelist_index = min(scene.autobake_bakelist_index + len(linked_sockets), len(list)-1)

    # List Count & Dupe Update
        if item is not None:
            item.Gate = False
            item.Gate = True                 
   
        return {'FINISHED'}


class AUTOBAKE_OT_Add(Operator):
    bl_idname = "autobake.bakelist_add"
    bl_label = "Add New"
    bl_description = "Add a new item to the list. Use 'CTRL + Click' to add more items"
    bl_options = {'INTERNAL', 'UNDO'}
        
    iteration : IntProperty(options = set(), name='', default=1, min=1, max=10, description='Choose how many items add to the list')
    scale_method : EnumProperty(options = set(), default='Multiply', name='Scale Method', description='Choose how the new items scale should change', items=[
        ('Copy', 'Copy', 'All the new items will share the same value'),
        ('Multiply', 'Multiply', "Each item's size will be multiplied"),
        ('Divide', 'Divide', "Each item's size will be divided")])
    multiply_value : FloatProperty(options = set(), name='Multiply Value', default=2, precision=2, step=1, min=1, max=10)
    divide_value : FloatProperty(options = set(), name='Divide Value', default=2, precision=2, step=1, min=1, max=10)
    use_ctrl : BoolProperty(options = set(), default=False)

    def invoke(self, context, event):
        self.use_ctrl = event.ctrl
        abp = context.scene.autobake_properties

        if (abp.ab_new_item_method == 'Advanced' and not self.use_ctrl) or (abp.ab_new_item_method == 'Simple' and self.use_ctrl):
            
            return context.window_manager.invoke_props_dialog(self, width=185)
        else:
            return self.execute(context)

    def draw(self, context):
        layout = self.layout   
        
        box = layout.box()
        col = box.column_flow(columns=1)
        
        row = col.row(align=True)
        row.label(text='New Items')
        row.prop(self, "iteration", text="")
        
        row = col.row(align=True)
        row.label(text='Scale')
        
        col = row.column(align=True)
        col.prop(self, "scale_method", text="")
        
        if self.scale_method == 'Multiply':
            col.prop(self, "multiply_value", text="")
        if self.scale_method == 'Divide':
            col.prop(self, "divide_value", text="")

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        list = scene.autobake_udimlist if abp.ab_udim_bake else scene.autobake_bakelist
        
    # Add Multiple
        if (abp.ab_new_item_method == 'Advanced' and not self.use_ctrl) or (abp.ab_new_item_method == 'Simple' and self.use_ctrl):
            self.use_ctrl = False
            
            for iteration in range(self.iteration):
                index = scene.autobake_udimlist_index if abp.ab_udim_bake else scene.autobake_bakelist_index
                
                item = list.add()

                if len(list) == 1:
                    if abp.ab_udim_bake:
                        scene.autobake_udimlist_index = 0
                    else:
                        scene.autobake_bakelist_index = 0
                else:
                    item.Type = list[index].Type
                    item.Size = list[index].Size

                    list.move(len(list)-1, index + 1)
                    
                    if abp.ab_udim_bake:
                        scene.autobake_udimlist_index = index + 1
                        abp.ab_udimtype_list_item_count = sum(1 for item in list if item.Gate)
                    else:
                        scene.autobake_bakelist_index = index + 1
                        abp.ab_bake_list_item_count = sum(1 for item in list if item.Gate)
                        
                    if self.scale_method == 'Multiply':
                        item.Size = int(list[index].Size * self.multiply_value)
                    elif self.scale_method == 'Divide':
                        item.Size = int(list[index].Size / self.divide_value)
    # Add One
        else:         
            index = scene.autobake_udimlist_index if abp.ab_udim_bake else scene.autobake_bakelist_index   
            item = list.add()

            if len(list) == 1:
                if abp.ab_udim_bake:
                    scene.autobake_udimlist_index = 0
                else:
                    scene.autobake_bakelist_index = 0
            else:
                item.Type = list[index].Type
                item.Size = list[index].Size

                list.move(len(list)-1, index + 1)
                
                if not abp.ab_udim_bake:
                    if scene.autobake_properties.ab_dynamic_scale != 'FALSE':
                        if len(list) > 2 and index > 0 and (list[index].Type == list[index-1].Type) and (list[index-1].Size*2 == list[index].Size):
                            item.Size = item.Size*2
                        elif len(list) > 2 and index > 0 and (list[index].Type == list[index-1].Type) and (list[index].Size*2 == list[index-1].Size):
                            item.Size = int(item.Size/2)
                            
                        if scene.autobake_properties.ab_dynamic_scale == 'ENABLED+':
                            if len(list) > 3 and index > 1 and (list[index].Size == list[index-1].Size) and (list[index-1].Size == list[index-2].Size*2) and (list[index].Type != list[index-1].Type) and (list[index-1].Type == list[index-2].Type):
                                item.Size = int(item.Size/2)
                                
                            elif len(list) > 3 and index > 1 and (list[index].Size == list[index-1].Size) and (list[index-1].Size == list[index-2].Size/2) and (list[index].Type != list[index-1].Type) and (list[index-1].Type == list[index-2].Type):
                                item.Size = item.Size*2
                
    # Item Count
        if abp.ab_udim_bake:
            scene.autobake_udimlist_index = min(index + 1, len(list)-1)
            abp.ab_udimtype_list_item_count = sum(1 for item in list if item.Gate)
        else:
            scene.autobake_bakelist_index = min(index + 1, len(list)-1)
            abp.ab_bake_list_item_count = sum(1 for item in list if item.Gate)
                                
    # Duplicate
        for item in list:
            item.IsDuplicate = sum(1 for item2 in list if (item.Type == item2.Type and item.Size == item2.Size)) > 1
            
        index = scene.autobake_udimlist_index if abp.ab_udim_bake else scene.autobake_bakelist_index
        if index > 0 and list[index].IsDuplicate and list[index].Type == list[index-1].Type and list[index].Size == list[index-1].Size and sum(1 for item in list if (item.Type == list[index].Type and item.Size == list[index].Size))<3:
            list[index].IsDuplicate = False
            list[index-1].IsDuplicate = False

    # Name
        for item in list:
            if abp.ab_udim_bake:
                item.name = f"{item.Type} {item.Size:.2f}" + (' Enabled' if item.Gate else ' Disabled') + (' Duplicated' if item.IsDuplicate else '')
            else:
                item.name = f"{item.Type} {item.Size}" + (' Enabled' if item.Gate else ' Disabled') + (' Duplicate' if item.IsDuplicate else '')
                
        return {'FINISHED'}


class AUTOBAKE_OT_Remove(Operator):
    bl_idname = "autobake.bakelist_remove"
    bl_label = "Remove Item"
    bl_description = "Remove the selected item from the list. Use 'CTRL + Click' to remove more items"
    bl_options = {'INTERNAL', 'UNDO'}

    use_ctrl : BoolProperty(options = set(), default=False)
    remove_method : EnumProperty(options = set(), name='Remove Method', default='All', description='Choose what items to remove from the list', items=[
        ("All", "All", "All items will be removed from the list"),
        ("Duplicated", "Duplicated", "Remove all the duplicated items from the list"),
        ("Disabled", "Disabled", "Remove all the disabled items from the list")])

    def invoke(self, context, event):
        self.use_ctrl = event.ctrl
        
        if self.use_ctrl:
            return context.window_manager.invoke_props_dialog(self, width=175)
        else:
            return self.execute(context)

    def draw(self, context):
        layout = self.layout   
        box = layout.box()
        split = box.split(factor=.4)
        
        split.label(text='Remove:')
        split.prop(self, 'remove_method', text='')
        
    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if self.use_ctrl:
            if self.remove_method == 'All':
                bpy.ops.autobake.remove_all()
            elif self.remove_method == 'Disabled':
                bpy.ops.autobake.remove_disabled()
            elif self.remove_method == 'Duplicated':
                bpy.ops.autobake.remove_duplicates()
            
        else:
            list = scene.autobake_udimlist if abp.ab_udim_bake else scene.autobake_bakelist
            index = scene.autobake_udimlist_index if abp.ab_udim_bake else scene.autobake_bakelist_index
            list.remove(index)

            if abp.ab_udim_bake:
                scene.autobake_udimlist_index = max(scene.autobake_udimlist_index - 1, 0)
                abp.ab_udimtype_list_item_count = sum(1 for item in list if item.Gate)
            else:
                scene.autobake_bakelist_index = max(scene.autobake_bakelist_index - 1, 0)
                abp.ab_bake_list_item_count = sum(1 for item in list if item.Gate)
                
            for item in list:
                item.IsDuplicate = sum(1 for item2 in list if (item.Type == item2.Type and item.Size == item2.Size)) > 1

                if abp.ab_udim_bake:
                    item.name = f"{item.Type} {item.Size:.2f}" + (' Enabled' if item.Gate else ' Disabled') + (' Duplicated' if item.IsDuplicate else '')
                else:
                    item.name = f"{item.Type} {item.Size}" + (' Enabled' if item.Gate else ' Disabled') + (' Duplicate' if item.IsDuplicate else '')
        return {'FINISHED'}

    
class AUTOBAKE_OT_Up(Operator):
    bl_idname = "autobake.bakelist_move_up"
    bl_label = "Move Up"
    bl_description = "Move item up in the list. Use 'CTRL + Click' to move it to the top"
    bl_options = {'INTERNAL', 'UNDO'}

    use_ctrl : BoolProperty(options = set(), default=False)

    def invoke(self, context, event):
        self.use_ctrl = event.ctrl
        return self.execute(context)

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        list = scene.autobake_udimlist if abp.ab_udim_bake else scene.autobake_bakelist
        index = scene.autobake_udimlist_index if abp.ab_udim_bake else scene.autobake_bakelist_index

        if self.use_ctrl:
            bpy.ops.autobake.move_to_top()
        else:
            if index > 0:
                list.move(index, index - 1)
                if abp.ab_udim_bake:
                    scene.autobake_udimlist_index -= 1
                else:
                    scene.autobake_bakelist_index -= 1
        return {'FINISHED'}


class AUTOBAKE_OT_Down(Operator):
    bl_idname = "autobake.bakelist_move_down"
    bl_label = "Move Down"
    bl_description = "Move item down in the list. Use 'CTRL + Click' to move it to the bottom"
    bl_options = {'INTERNAL', 'UNDO'}

    use_ctrl : BoolProperty(options = set(), default=False)

    def invoke(self, context, event):
        self.use_ctrl = event.ctrl
        return self.execute(context)

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        list = scene.autobake_udimlist if abp.ab_udim_bake else scene.autobake_bakelist
        index = scene.autobake_udimlist_index if abp.ab_udim_bake else scene.autobake_bakelist_index
        
        if self.use_ctrl:
            bpy.ops.autobake.move_to_bottom()
        else:
            if index < len(list) - 1:
                list.move(index, index + 1)
                if abp.ab_udim_bake:
                    scene.autobake_udimlist_index += 1
                else:
                    scene.autobake_bakelist_index += 1
        return {'FINISHED'}


class AUTOBAKE_OT_MoveTop(Operator):
    bl_idname = "autobake.move_to_top"
    bl_label = "Move To Top"
    bl_description = "Move the active item to the top of the list. Call this function with 'CTRL + Click' on the 'Move Up' button"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if not abp.ab_udim_bake:
            scene.autobake_bakelist.move(scene.autobake_bakelist_index, 0)
            scene.autobake_bakelist_index = 0
        else:
            scene.autobake_udimlist.move(scene.autobake_udimlist_index, 0)
            scene.autobake_udimlist_index = 0
        return {'FINISHED'} 
    
    
class AUTOBAKE_OT_MoveBottom(Operator):
    bl_idname = "autobake.move_to_bottom"
    bl_label = "Move To Top"
    bl_description = "Move the active item to the top of the list. Call this function with 'CTRL + Click' on the 'Move Down' button"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if not abp.ab_udim_bake:
            scene.autobake_bakelist.move(scene.autobake_bakelist_index, len(scene.autobake_bakelist)-1)
            scene.autobake_bakelist_index = len(scene.autobake_bakelist)-1
        else:
            scene.autobake_udimlist.move(scene.autobake_udimlist_index, len(scene.autobake_udimlist)-1)
            scene.autobake_udimlist_index = len(scene.autobake_udimlist)-1
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_RemoveDuplicates(Operator):
    bl_idname = "autobake.remove_duplicates"
    bl_label = "Remove Duplicated"
    bl_description = "Remove all the duplicated items from the list. Call this function with 'CTRL + Click' on the 'Remove Item' button"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if not abp.ab_udim_bake:
            for index in range(len(scene.autobake_bakelist)-1, -1, -1):
                if scene.autobake_bakelist[index].IsDuplicate:
                    scene.autobake_bakelist.remove(index)
                    if scene.autobake_bakelist_index > len(scene.autobake_bakelist)-1:
                        scene.autobake_bakelist_index = len(scene.autobake_bakelist)-1
                    for item in scene.autobake_bakelist:
                        item.IsDuplicate = sum(1 for Item in context.scene.autobake_bakelist if Item.Type == item.Type and Item.Size == item.Size) > 1
            scene.autobake_properties.ab_bake_list_item_count = sum(1 for item in scene.autobake_bakelist if item.Gate)
        else:
            for index in range(len(scene.autobake_udimlist)-1, -1, -1):
                if scene.autobake_udimlist[index].IsDuplicate:
                    scene.autobake_udimlist.remove(index)
                    if scene.autobake_udimlist_index > len(scene.autobake_udimlist)-1:
                        scene.autobake_udimlist_index = len(scene.autobake_udimlist)-1
                    for item in scene.autobake_udimlist:
                        item.IsDuplicate = sum(1 for Item in scene.autobake_udimlist if (item.Type == Item.Type and item.Size == Item.Size)) > 1
            abp.ab_bake_list_item_count = sum(1 for item in scene.autobake_udimlist if item.Gate)
        return {'FINISHED'}


class AUTOBAKE_OT_RemoveDisabled(Operator):
    bl_idname = "autobake.remove_disabled"
    bl_label = "Remove Disabled"
    bl_description = "Remove all the disabled items from the list. Call this function with 'CTRL + Click' on the 'Remove Item' button"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if not abp.ab_udim_bake:
            for index in range(len(scene.autobake_bakelist)-1, -1, -1):
                if scene.autobake_bakelist[index].Gate == False:
                    scene.autobake_bakelist.remove(index)
            if scene.autobake_bakelist_index > len(scene.autobake_bakelist)-1:
                scene.autobake_bakelist_index = len(scene.autobake_bakelist)-1
            for item in scene.autobake_bakelist:
                item.IsDuplicate = sum(1 for Item in context.scene.autobake_bakelist if Item.Type == item.Type and Item.Size == item.Size) > 1
        else:
            for index in range(len(scene.autobake_udimlist)-1, -1, -1):
                if scene.autobake_udimlist[index].Gate == False:
                    scene.autobake_udimlist.remove(index)
            if scene.autobake_udimlist_index > len(scene.autobake_udimlist)-1:
                scene.autobake_udimlist_index = len(scene.autobake_udimlist)-1
            for item in scene.autobake_udimlist:
                item.IsDuplicate = sum(1 for Item in scene.autobake_udimlist if (item.Type == Item.Type and item.Size == Item.Size)) > 1
        return {'FINISHED'}
        
        
class AUTOBAKE_OT_RemoveAll(Operator):
    bl_idname = "autobake.remove_all"
    bl_label = "Remove All"
    bl_description = "Remove all the items from the list. Call this function with 'CTRL + Click' on the 'Remove Item' button"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if not abp.ab_udim_bake:
            scene.autobake_bakelist.clear()
            abp.ab_bake_list_item_count = 0
        else:
            scene.autobake_udimlist.clear()
            abp.ab_bake_list_item_count = 0
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_EnableAll(Operator):
    bl_idname = "autobake.enable_all"
    bl_label = "Enable All"
    bl_description = "Enable all the items in the list"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if not abp.ab_udim_bake:
            for item in scene.autobake_bakelist:
                item.Gate = True
        else:
            for item in scene.autobake_udimlist:
                item.Gate = True
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_DisableAll(Operator):
    bl_idname = "autobake.disable_all"
    bl_label = "Enable All"
    bl_description = "Disable all the items in the list"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if not abp.ab_udim_bake:
            for item in scene.autobake_bakelist:
                item.Gate = False
        else:
            for item in scene.autobake_udimlist:
                item.Gate = False
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_InvertAll(Operator):
    bl_idname = "autobake.invert_all"
    bl_label = "Invert All"
    bl_description = "Invert all the items in the list"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if not abp.ab_udim_bake:
            for item in scene.autobake_bakelist:
                item.Gate = not bool(item.Gate)
        else:
            for item in scene.autobake_udimlist:
                item.Gate = not bool(item.Gate)
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_Sort(Operator):
    bl_idname = "autobake.reorder"
    bl_label = "Sort Items"
    bl_description = "Reorder all the items by type and scale"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if not abp.ab_udim_bake:
            autobake_bakelist = scene.autobake_bakelist
            types = []
            active_item = ((autobake_bakelist[scene.autobake_bakelist_index].Type, autobake_bakelist[scene.autobake_bakelist_index].Size, autobake_bakelist[scene.autobake_bakelist_index].Gate, autobake_bakelist[scene.autobake_bakelist_index].IsDuplicate))
            
            for index in range(len(autobake_bakelist)):
                item = autobake_bakelist[index]
                
                if item.Type not in types:
                    types.append(item.Type)
                elif item.Type in types:
                    adjust_index = 0
                    if autobake_bakelist[index].Type != autobake_bakelist[index-1].Type:
                        while autobake_bakelist[index-adjust_index].Type != autobake_bakelist[index-adjust_index-1].Type:
                            autobake_bakelist.move(index-adjust_index, index-adjust_index-1)
                            adjust_index = adjust_index+1
                            
            for index in range(len(autobake_bakelist)):
                item = autobake_bakelist[index]
                adjust_index = 0
                
                if autobake_bakelist[index].Type == autobake_bakelist[index-1].Type and autobake_bakelist[index].Size < autobake_bakelist[index-1].Size:
                    while autobake_bakelist[index-adjust_index].Type == autobake_bakelist[index-adjust_index-1].Type and autobake_bakelist[index-adjust_index].Size < autobake_bakelist[index-adjust_index-1].Size:
                        autobake_bakelist.move(index-adjust_index, index-adjust_index-1)
                        adjust_index = adjust_index+1
            
            for index in range(len(autobake_bakelist)):
                item = autobake_bakelist[index]
                if item.Type == active_item[0] and item.Size == active_item[1] and item.Gate == active_item[2] and item.IsDuplicate == active_item[3]:
                    scene.autobake_bakelist_index = index
        else:
            autobake_udimlist = scene.autobake_udimlist
            types = []
            active_item = ((autobake_udimlist[scene.autobake_udimlist_index].Type, autobake_udimlist[scene.autobake_udimlist_index].Size, autobake_udimlist[scene.autobake_udimlist_index].Gate, autobake_udimlist[scene.autobake_udimlist_index].IsDuplicate))
            
            for index in range(len(autobake_udimlist)):
                item = autobake_udimlist[index]
                
                if item.Type not in types:
                    types.append(item.Type)
                elif item.Type in types:
                    adjust_index = 0
                    if autobake_udimlist[index].Type != autobake_udimlist[index-1].Type:
                        while autobake_udimlist[index-adjust_index].Type != autobake_udimlist[index-adjust_index-1].Type:
                            autobake_udimlist.move(index-adjust_index, index-adjust_index-1)
                            adjust_index = adjust_index+1
                            
            for index in range(len(autobake_udimlist)):
                item = autobake_udimlist[index]
                adjust_index = 0
                
                if autobake_udimlist[index].Type == autobake_udimlist[index-1].Type and autobake_udimlist[index].Size < autobake_udimlist[index-1].Size:
                    while autobake_udimlist[index-adjust_index].Type == autobake_udimlist[index-adjust_index-1].Type and autobake_udimlist[index-adjust_index].Size < autobake_udimlist[index-adjust_index-1].Size:
                        autobake_udimlist.move(index-adjust_index, index-adjust_index-1)
                        adjust_index = adjust_index+1
            
            for index in range(len(autobake_udimlist)):
                item = autobake_udimlist[index]
                if item.Type == active_item[0] and item.Size == active_item[1] and item.Gate == active_item[2] and item.IsDuplicate == active_item[3]:
                    scene.autobake_udimlist_index = index
        return {'FINISHED'}


class AUTOBAKE_OT_ScaleUp(Operator):
    bl_idname = "autobake.scale_up"
    bl_label = "Scale Up"
    bl_description = "Multiplies the texture size by 2"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if abp.ab_udim_bake:
            item = scene.autobake_udimtilelist[scene.autobake_udimtilelist_index]
            item.Size = int(item.Size * 2)
        else:
            item = scene.autobake_bakelist[scene.autobake_bakelist_index]
            item.Size = int(item.Size * 2)
        return {'FINISHED'}


class AUTOBAKE_OT_ScaleDown(Operator):
    bl_idname = "autobake.scale_down"
    bl_label = "Scale Down"
    bl_description = "Divides the texture size by 2"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if abp.ab_udim_bake:
            item = scene.autobake_udimtilelist[scene.autobake_udimtilelist_index]
            item.Size = int(item.Size / 2)
        else:
            item = scene.autobake_bakelist[scene.autobake_bakelist_index]
            item.Size = int(item.Size / 2)
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_Add_UDIM(Operator):
    bl_idname = "autobake.udimlist_add"
    bl_label = "New Item"
    bl_description = "Add a new item to the list"
    bl_options = {'INTERNAL', 'UNDO'}

    use_ctrl : BoolProperty(options = set(), default=False)
    new_items : IntProperty(options = set(), default=1, name='Item Count', description='Set how many items will be added to the list', min=1, max=100)
    start_tile : IntProperty(options = set(), default=1001, min=1001, max=2000, description='The new items will start with this tile')

    def invoke(self, context, event):
        self.use_ctrl = event.ctrl
        scene = context.scene
        
        if event.ctrl or scene.autobake_properties.ab_new_item_method == 'Advanced':
            list = scene.autobake_udimtilelist
            if len(list) > 0:
                self.start_tile = list[scene.autobake_udimtilelist_index].UDIM + 1
            
            return context.window_manager.invoke_props_dialog(self, width=175)
        else:
            return self.execute(context)

    def draw(self, context):
        layout = self.layout   
        box = layout.box()
        col = box.column_flow(columns=1)
        
        split = col.split(factor=.4)
        split.label(text='New Items')
        split.prop(self, 'new_items', text='')
        
        split = col.split(factor=.4)
        split.label(text='Start Tile')
        split.prop(self, 'start_tile', text='')
        
        split = col.split(factor=.4)
        split.active = False
        split.label(text='Last Tile')
        row = split.row()
        row.alignment = 'CENTER'
        row.label(text=str(self.start_tile+self.new_items-1))
        
    def execute(self, context):
        scene = context.scene
        list = scene.autobake_udimtilelist
        abp = scene.autobake_properties
        
        if self.use_ctrl or abp.ab_new_item_method == 'Advanced':
            for index in range(self.new_items):
                item = list.add()
                item.UDIM = self.start_tile + index
                if len(list)>1:
                    scene.autobake_udimtilelist_index += 1
                else:
                    scene.autobake_udimtilelist_index = 0
        else:
            item = list.add()
            if len(list)>1:
                scene.autobake_udimtilelist_index += 1
                item.UDIM = list[scene.autobake_udimtilelist_index-1].UDIM+1
                item.Size = list[scene.autobake_udimtilelist_index-1].Size
            else:
                scene.autobake_udimtilelist_index = 0
            
        scene.autobake_properties.ab_udim_list_item_count = sum(1 for item in context.scene.autobake_udimtilelist if item.Gate)

        for item in list:
            item.IsDuplicate = sum(1 for item2 in list if item.UDIM == item2.UDIM) > 1
            item.name = f"{item.UDIM} {item.Size} " + ('Enabled' if item.Gate else 'Disabled') + (' Duplicated' if item.IsDuplicate else '')
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_Remove_UDIM(Operator):
    bl_idname = "autobake.udimlist_remove"
    bl_label = "Remove Item"
    bl_description = "Add a new item to the list"
    bl_options = {'INTERNAL', 'UNDO'}


    use_ctrl : BoolProperty(options = set(), default=False)
    remove_method : EnumProperty(options = set(), name='Remove Method', default='All', description='Choose what items to remove from the list', items=[
        ("All", "All", "All items will be removed from the list"),
        ("Duplicated", "Duplicated", "Remove all the duplicated items from the list"),
        ("Disabled", "Disabled", "Remove all the disabled items from the list")])

    def invoke(self, context, event):
        self.use_ctrl = event.ctrl
        
        if self.use_ctrl:
            return context.window_manager.invoke_props_dialog(self, width=175)
        else:
            return self.execute(context)

    def draw(self, context):
        layout = self.layout   
        box = layout.box()
        split = box.split(factor=.4)
        
        split.label(text='Remove:')
        split.prop(self, 'remove_method', text='')

    def execute(self, context):
        scene = context.scene
        list = scene.autobake_udimtilelist
        index = scene.autobake_udimtilelist_index
        
        if self.use_ctrl:
            if self.remove_method == 'All':
                bpy.ops.autobake.udimlist_remove()
            elif self.remove_method == 'Disabled':
                bpy.ops.autobake.udimlist_remove_disabled()
            elif self.remove_method == 'Duplicated':
                bpy.ops.autobake.udimlist_remove_duplicates()
        else:
            if index >=1:
                scene.autobake_udimtilelist_index-=1
            else:
                scene.autobake_udimtilelist_index=0
            list.remove(index)
            scene.autobake_properties.ab_udim_list_item_count = sum(1 for item in context.scene.autobake_udimtilelist if item.Gate)
            for item in list:
                item.IsDuplicate = sum(1 for item2 in list if item.UDIM == item2.UDIM) > 1
                item.name = f"{item.UDIM} {item.Size} " + ('Enabled' if item.Gate else 'Disabled') + (' Duplicated' if item.IsDuplicate else '')
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_Up_UDIM(Operator):
    bl_idname = "autobake.udimlist_up"
    bl_label = "Move Up"
    bl_description = "Move item up in the list"
    bl_options = {'INTERNAL', 'UNDO'}

    use_ctrl : BoolProperty(options = set(), default=False)

    def invoke(self, context, event):
        self.use_ctrl = event.ctrl
        return self.execute(context)

    def execute(self, context):
        scene = context.scene
        index = scene.autobake_udimtilelist_index
        
        if self.use_ctrl:
            bpy.ops.autobake.move_to_top_udim()
        else:
            if index >= 1:
                scene.autobake_udimtilelist.move(index, index - 1)
                scene.autobake_udimtilelist_index -= 1
     
        return {'FINISHED'}


class AUTOBAKE_OT_Down_UDIM(Operator):
    bl_idname = "autobake.udimlist_down"
    bl_label = "Move Down"
    bl_description = "Move item down in the list"
    bl_options = {'INTERNAL', 'UNDO'}

    use_ctrl : BoolProperty(options = set(), default=False)

    def invoke(self, context, event):
        self.use_ctrl = event.ctrl
        return self.execute(context)

    def execute(self, context):
        scene = context.scene
        index = scene.autobake_udimtilelist_index
        
        if self.use_ctrl:
            bpy.ops.autobake.move_to_bottom_udim()
        else:
            if index < len(scene.autobake_udimtilelist) - 1:
                scene.autobake_udimtilelist.move(index, index + 1)
                scene.autobake_udimtilelist_index += 1
        return {'FINISHED'}


class AUTOBAKE_OT_MoveTop_UDIM(Operator):
    bl_idname = "autobake.move_to_top_udim"
    bl_label = "Move To Top"
    bl_description = "Move the active item to the top of the list. Call this function with 'CTRL + Click' on the 'Move Up' button"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        scene.autobake_bakelist.move(scene.autobake_udimtilelist_index, 0)
        scene.autobake_udimtilelist_index = 0
        return {'FINISHED'} 
    
    
class AUTOBAKE_OT_MoveBottom_UDIM(Operator):
    bl_idname = "autobake.move_to_bottom_udim"
    bl_label = "Move To Bottom"
    bl_description = "Move the active item to the bottom of the list. Call this function with 'CTRL + Click' on the 'Move Down' button"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        scene.autobake_udimtilelist.move(scene.autobake_udimtilelist_index, len(scene.autobake_udimtilelist)-1)
        scene.autobake_udimtilelist_index = len(scene.autobake_udimtilelist)-1
        return {'FINISHED'}


class AUTOBAKE_OT_RemoveAll_UDIM(Operator):
    bl_idname = "autobake.udimlist_remove_all"
    bl_label = "Remove All"
    bl_description = "Remove all the items from the list"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        scene.autobake_udimtilelist.clear()
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_RemoveDuplicates_UDIM(Operator):
    bl_idname = "autobake.udimlist_remove_duplicates"
    bl_label = "Remove Duplicated"
    bl_description = "Remove all the duplicated items from the list"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        list = scene.autobake_udimtilelist
        
        for index in range(len(list)-1, -1, -1):
            if list[index].IsDuplicate:
                list.remove(index)
                if scene.autobake_udimtilelist_index > len(list)-1:
                    scene.autobake_udimtilelist_index = len(list)-1
                for item in list:
                    item.IsDuplicate = sum(1 for Item in list if (item.UDIM == Item.UDIM and item.UDIM == Item.UDIM)) > 1
        
        for item in list:
            item.IsDuplicate = sum(1 for item2 in list if item.UDIM == item2.UDIM) > 1

        return {'FINISHED'}
    

class AUTOBAKE_OT_RemoveDisabled_UDIM(Operator):
    bl_idname = "autobake.udimlist_remove_disabled"
    bl_label = "Remove Disabled"
    bl_description = "Remove all the disabled items from the list"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        list = scene.autobake_udimtilelist

        for index in range(len(list)-1, -1, -1):
            if list[index].Gate == False:
                list.remove(index)
        if scene.autobake_udimtilelist_index > len(list)-1:
            scene.autobake_udimtilelist_index = len(list)-1
        for item in list:
            item.IsDuplicate = sum(1 for item2 in list if item.UDIM == item2.UDIM) > 1
        return {'FINISHED'}


class AUTOBAKE_OT_EnableAll_UDIM(Operator):
    bl_idname = "autobake.udimlist_enable_all"
    bl_label = "Enable All"
    bl_description = "Enable all the items in the list"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        for item in scene.autobake_udimtilelist:
            item.Gate = True
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_DisableAll_UDIM(Operator):
    bl_idname = "autobake.udimlist_disable_all"
    bl_label = "Disable All"
    bl_description = "Disable all the items in the list"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties

        for item in scene.autobake_udimtilelist:
            item.Gate = False
            
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_InvertAll_UDIM(Operator):
    bl_idname = "autobake.udimlist_invert_all"
    bl_label = "Invert All"
    bl_description = "Invert all the items in the list"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        for item in scene.autobake_udimtilelist:
            item.Gate = not bool(item.Gate)
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_Sort_UDIM(Operator):
    bl_idname = "autobake.udimlist_sort"
    bl_label = "Sort Items"
    bl_description = "Reorder all the items by type and scale"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        autobake_udimtilelist = scene.autobake_udimtilelist

        for index in range(len(autobake_udimtilelist)):
            item = autobake_udimtilelist[index]
            adjust_index = 0
            if autobake_udimtilelist[index].UDIM < autobake_udimtilelist[index-1].UDIM:
                while autobake_udimtilelist[index-adjust_index].UDIM < autobake_udimtilelist[index-adjust_index-1].UDIM:
                    autobake_udimtilelist.move(index-adjust_index, index-adjust_index-1)
                    adjust_index = adjust_index+1
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_ImportTiles_UDIM(Operator):
    bl_idname = "autobake.udimlist_import"
    bl_label = "Source Image"
    bl_description = "Import the existing UDIM tiles from the selected image"
    bl_options = {'INTERNAL', 'UNDO'}
    
    clear_list : BoolProperty(options = set(), default=True, description="Clear the current items in the 'UDIM Tiles' list")
    
    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self, width=250)
    
    def draw(self, context):
        layout = self.layout
        layout.prop(context.scene.autobake_properties, "ab_import_tiles", text="")
        layout.prop(self, 'clear_list', text='Clear List')
        

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        image = abp.ab_import_tiles
        list = scene.autobake_udimtilelist
        
        if image.source == 'TILED':
            if self.clear_list:
                list.clear()
            for tile in image.tiles:
                item = list.add()
                item.UDIM = tile.number
                item.Size = tile.size[0]
            abp.ab_import_tiles=None
            abp.ab_udim_list_item_count = sum(1 for item in context.scene.autobake_udimtilelist if item.Gate)
        else:
            self.report({'ERROR'}, "Source Image needs to be a UDIM image")

        return {'FINISHED'}
    
    
class AUTOBAKE_OT_NameStructure(Operator):
    bl_idname = "autobake.edit_name_structure"
    bl_label = "Name Structure Editor"
    bl_options = {'INTERNAL', 'UNDO'}
    bl_description = "Edit your textures' name structures"
   
    name_structure: StringProperty(options = set(), name="Name Structure", default="", description='Structure to use when creating and exporting non-udim images. Leave if empty for reset')
    name_structure_udim: StringProperty(options = set(), name="UDIM Name Structure", default="", description='Structure to use when creating UDIM images. Leave if empty for reset')
    name_structure_udim_export: StringProperty(options = set(), name="UDIM Name Structure: Export", default='', description="Structure to use when exporting UDIM images. Must contain the UDIM Number placeholder Leave if empty for reset")
    variable_values = {}
    variable_values_udim = {}
    udim_texture = BoolProperty(options = set(), default=False)
    
    def invoke(self, context, event):
        scene = context.scene
        abp = scene.autobake_properties
        
        self.name_structure = abp.ab_name_structure
        self.name_structure_udim = abp.ab_name_structure_udim
        self.name_structure_udim_export = abp.ab_name_structure_udim_export
        
        self.variable_values = {"prefix": abp.ab_prefix if abp.ab_prefix != '' else context.active_object.name if context.active_object is not None else 'Cube',
                                "bridge": abp.ab_bridge,
                                "suffix": abp.ab_suffix,
                                "type": 'Metallic',
                                "size": '512',
                                "udim": '1021',
                                "uvtile": 'u1_v3'}
                                
        self.variable_values_udim = {"prefix": abp.ab_prefix if abp.ab_prefix != '' else context.active_object.name if context.active_object is not None else 'Cube',
                                     "bridge": abp.ab_bridge,
                                     "suffix": abp.ab_suffix,
                                     "type": 'Metallic',
                                     "size": '1.00',
                                     "udim": '1021',
                                     "uvtile": 'u1_v3'}
                                
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties

    # Name Structure
        row = layout.row()
        row.alignment = 'CENTER'
        row.label(text='• Name Structure •')
        layout.prop(self, 'name_structure', text='')
        box = layout.box()
        if self.name_structure == '':
            box.label(text='Structure will be reset!')
        else:
            try:
                result_string = self.name_structure.lower().format(**self.variable_values)
                if any(element in self.name_structure.lower() for element in ['{udim}', '{uvtile}']):
                    box.alert = True
                    box.label(text="Structure can't contain 'UDIM' and 'UVTILE' elements!")
                else:
                    box.label(text=f"Output: {result_string}")
                    box.label(text=f"Export: {result_string}.{abp.ab_fileformat}")
            except KeyError as e:
                box.alert = True
                box.label(text=f"Input error! Invalid Key: '{{{e.args[0]}}}'.")
        layout.separator()

    # Name Structure UDIM
        row=layout.row()
        row.alignment = 'CENTER'
        row.label(text='• UDIM Name Structure •')
        layout.prop(self, 'name_structure_udim', text='')
        box = layout.box()
        if self.name_structure_udim == '':
            box.label(text='Structure will be reset!')
        else:
            try:
                result_string_udim = self.name_structure_udim.lower().format(**self.variable_values_udim)
                if any(element in self.name_structure_udim.lower() for element in ['{udim}', '{uvtile}']):
                    box.alert = True
                    box.label(text="Structure can't contain 'UDIM' and 'UVTILE' elements!")
                else:
                    box.label(text=f"Output: {result_string_udim}")
            except KeyError as e:
                box.alert = True
                box.label(text=f"Input error! Invalid Key: '{{{e.args[0]}}}'.")
        layout.separator()
            
    # Name Structure UDIM Export
        row=layout.row()
        row.alignment = 'CENTER'
        row.label(text='• UDIM Export Name Structure •')
        layout.prop(self, 'name_structure_udim_export', text='')
        box = layout.box()
        if self.name_structure_udim_export == '':
            box.label(text='Structure will be reset!')
        else:
            try:
                result_string_udim_export = self.name_structure_udim_export.lower().format(**self.variable_values_udim)
                if '{udim}' not in self.name_structure_udim_export.lower() and '{uvtile}' not in self.name_structure_udim_export.lower():
                    box.alert = True
                    box.label(text="Missing 'UDIM' or 'UVTILE' element!")
                else:
                    box.label(text=f"Export: {result_string_udim_export}.{abp.ab_fileformat}")
            except KeyError as e:
                result_string_udim = f"Error: Invalid key: '{{{e.args[0]}}}'."      
        layout.separator()
        
    # Elements
        row=layout.row()
        row.alignment = 'CENTER'
        row.label(text='• Structure Elements •')
        box = layout.box()
        column = box.grid_flow(align=True, columns=2)
        column.label(text="Prefix: {prefix}", icon='DOT')
        column.label(text="Type: {type}", icon='DOT')
        column.label(text="Bridge: {bridge}", icon='DOT')
        column.label(text="Size: {size}", icon='DOT')
        column.label(text="Suffix: {suffix}", icon='DOT')
        column.label(text="UDIM: {udim}", icon='DOT')
        column.label(text="UVTILE: {uvtile}", icon='DOT')
      
    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
    
    # Single
        if self.name_structure == '':
            abp.ab_name_structure = '{prefix}{type}{bridge}{size}{suffix}'
            self.report({'INFO'}, f"Auto Bake: 'Name Structure' got reset!")
        else:
            try:
                result_string = self.name_structure.lower().format(**self.variable_values)
                if any(element in self.name_structure.lower() for element in ['{udim}', '{uvtile}']):
                    self.report({'ERROR'}, "Auto Bake: 'Name Structure' can't contain 'UDIM' and 'UVTILE' elements!")
                else:
                    abp.ab_name_structure = self.name_structure
            except KeyError as e:
                self.report({'ERROR'}, "Auto Bake: 'Name Structure' Input error! Invalid Key: '{{{e.args[0]}}}'.")
        
    # UDIM
        if self.name_structure_udim == '':
            abp.ab_name_structure_udim = '{prefix}{type}{bridge}{size}x'
            self.report({'INFO'}, f"Auto Bake: 'UDIM Name Structure' got reset!")
        else:
            try:
                result_string_udim = self.name_structure_udim.lower().format(**self.variable_values_udim)
                if any(element in self.name_structure_udim.lower() for element in ['{udim}', '{uvtile}']):
                    self.report({'ERROR'}, "Auto Bake: 'UDIM Name Structure' can't contain 'UDIM' and 'UVTILE' elements!")
                else:
                    abp.ab_name_structure_udim = self.name_structure_udim
            except KeyError as e:
                self.report({'ERROR'}, f"Auto Bake: Failed to edit 'UDIM Name Structure'. Invalid key: '{e.args[0]}'")
        
    # UDIM Export
        if self.name_structure_udim_export == '':
            abp.ab_name_structure_udim_export = '{prefix}{type}{bridge}{size}x.{udim}'
            self.report({'INFO'}, f"Auto Bake: 'UDIM Export Name Structure' got reset!")
        else:
            try:
                result_string_udim_export = self.name_structure_udim_export.lower().format(**self.variable_values_udim)
                if '{udim}' not in self.name_structure_udim_export.lower() and '{uvtile}' not in self.name_structure_udim_export.lower():
                    self.report({'ERROR'}, "Auto Bake: 'UDIM Export Name Structure' must contain 'UDIM' or 'UVTILE' element")
                else:
                    abp.ab_name_structure_udim_export = self.name_structure_udim_export
            except KeyError as e:
                self.report({'ERROR'}, f"Auto Bake: Failed to edit 'UDIM Export Name Structure'. Invalid key: '{e.args[0]}'")  
                
        return {'FINISHED'}


class AUTOBAKE_OT_SelectFromList(Operator):
    bl_idname = "autobake.select_from_list"
    bl_label = "Select Objects"
    bl_description = "Select the items in the 3D viewport from the list"
    bl_options = {'INTERNAL', 'UNDO'}

    clear : BoolProperty(options = set(), name='Deselect Selected', description='Deselect all the currently selected objects in the 3D viewport before selecting the objects from the list', default=True)

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self, width=200)
    
    def draw(self, context):
        layout = self.layout
        layout.prop(self, 'clear')

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
    
        if self.clear:
            for obj in reversed(bpy.context.selected_objects):
                obj.select_set(False)
                
        if abp.ab_source_object_method == 'LIST':
            for item in scene.autobake_sourceobject:
                if item.Object != None:
                    item.Object.select_set(True)
        else:
            for obj in abp.ab_source_collection.objects:
                if obj.type != 'LIGHT':
                    obj.select_set(True)

        abp.ab_target_object.select_set(True)
        context.view_layer.objects.active = abp.ab_target_object

        return {'FINISHED'}


class AUTOBAKE_OT_LoadFromSelected(Operator):
    bl_idname = "autobake.load_selected_objects"
    bl_label = "Load from Selected"
    bl_description = "Load the currently selected items from the 3D viewport to the list"
    bl_options = {'INTERNAL', 'UNDO'}

    clear : BoolProperty(options = set(), name='Clear List', description='Remove all the list items before loading them from the 3D viewport')
    active : BoolProperty(options = set(), name='Active as Target', description='Load the active object from the selection as the new target object')

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self, width=300)
    
    def draw(self, context):
        layout = self.layout
        
        row = layout.row(align=True)
        row.prop(self, 'clear', toggle=True, icon='X', text='Clear Objects')
        row.prop(self, 'active', toggle=True, icon='OBJECT_DATAMODE')

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        if self.clear:
            scene.autobake_sourceobject.clear()
            
        for obj in bpy.context.selected_objects:
            if not any(item.Object == obj for item in scene.autobake_sourceobject):
                item = scene.autobake_sourceobject.add()
                item.Object = obj
                
        if self.active and context.active_object != None and context.active_object.select_get():
            abp.ab_target_object = context.active_object
    
            for index, item in enumerate(scene.autobake_sourceobject):
                if item.Object == abp.ab_target_object:
                    scene.autobake_sourceobject.remove(index)
                    break
                
        if len(scene.autobake_sourceobject) == 0:
            scene.autobake_sourceobject.add()
            
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_ToggleQueueItem(Operator):
    bl_idname = "autobake.toggle_queued"
    bl_label = "Enable / Disable"
    bl_description = "Enable or Disable the bake item in the queue"
    bl_options = {'INTERNAL', 'UNDO'}
    
    index : IntProperty(options = set(), )
    
    def invoke(self, context, event):
        if event.ctrl or not context.scene.autobake_properties.ab_confirm_queue_item_gate:
            return self.execute(context)
        else:
            return context.window_manager.invoke_confirm(self, event)
    
    def execute(self, context):
        scene = context.scene
        item = scene.autobake_queuelist[self.index]
        if item.Status == 'Pending':
            item.Status = 'Canceled'
            item.Cancel = True
            item.Icon = 'CANCEL'
        elif item.Status == 'Canceled':
            item.Status = 'Pending'
            item.Cancel = False
            item.Icon = 'PREVIEW_RANGE'
            
        item.name = f"{item.Type} {item.Size} {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')
        
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_ToggleObjectQueueItem(Operator):
    bl_idname = "autobake.toggle_queued_object"
    bl_label = "Enable / Disable"
    bl_description = "Enable or Disable the bake item in the queue"
    bl_options = {'INTERNAL', 'UNDO'}
    
    index : IntProperty(options = set(), )
    
    def invoke(self, context, event):
        if event.ctrl or not context.scene.autobake_properties.ab_confirm_queue_item_gate:
            return self.execute(context)
        else:
            return context.window_manager.invoke_confirm(self, event)
    
    def execute(self, context):
        scene = context.scene
        item = scene.autobake_objectqueuelist[self.index]
        if item.Status == 'Pending':
            item.Status = 'Canceled'
            item.Cancel = True
            item.Icon = 'CANCEL'
        elif item.Status == 'Canceled':
            item.Status = 'Pending'
            item.Cancel = False
            item.Icon = 'PREVIEW_RANGE'
            
        item.name = f"{item.Object.name} {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')
        return {'FINISHED'}


class AUTOBAKE_OT_ExportTextures(Operator):
    bl_idname = "autobake.export_textures"
    bl_label = "Export List"
    bl_description = "Export the unsaved textures from the previous bakes"
    bl_options = {'INTERNAL', 'UNDO'}

    def toggle_all(self, context):
        scene = context.scene
        list = scene.autobake_imageexport
        
        for item in list:
            if item.Image.is_dirty:
                item.Gate = self.toggle_all

    toggle_all : BoolProperty(options = set(), default=True, update=toggle_all)

    def invoke(self, context, event):
        scene = context.scene
        list = scene.autobake_imageexport

        for index in reversed(range(len(list))):
            if list[index].Image is None:
                list.remove(index)
    
        for item in list:
            if item.Image.is_dirty:
                self.toggle_all = item.Gate
                break
             
        for item in list:
            if not item.Image.is_dirty:
                item.Gate = False
            
        return context.window_manager.invoke_props_dialog(self, width=375)
    

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        
        
        box = layout.box()
        
        split = box.split(factor=.3225, align=True)
        
        split.label(text='Folder', icon='FILE_FOLDER')
        
        row = split.row()
        
        row.label(text='Texture', icon='TEXTURE')
        row2 = row.row()
        row2.alignment = 'RIGHT'
        row2.label(text='Export')
        row.prop(self, 'toggle_all', text='')

        layout.template_list("AUTOBAKE_UL_ImageExport", "Image_Export_List", scene, "autobake_imageexport", scene, "autobake_imageexport_index", rows=2)

        if any(item.Image and item.Image.is_dirty == False for item in scene.autobake_imageexport):
            layout.label(icon='ERROR', text='Unsaved image data is lost or already saved.')

    def execute(self, context):
        scene = context.scene
        
        for item in scene.autobake_imageexport:
            if item.Gate:
                export_texture(self, context, item.Image, item.Name, item.Type, item.Label, item.Prefix)
        scene.autobake_imageexport.clear()
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_FolderExplorer(Operator):
    bl_idname = "autobake.file_explorer"
    bl_label = "Select Folder"
    bl_options = {'INTERNAL', 'UNDO'}
    
    directory: StringProperty(options = set(), )

    def execute(self, context):
        bpy.context.scene.autobake_properties.ab_filepath = os.path.dirname(self.directory)
        return {'FINISHED'}
        
    def invoke(self, context, event):
        if bool(os.path.exists(bpy.context.scene.autobake_properties.ab_filepath)):
            self.directory = bpy.context.scene.autobake_properties.ab_filepath
        context.window_manager.fileselect_add(self)
        
        return {'RUNNING_MODAL'} 


class AUTOBAKE_OT_BakeConfirm(Operator):
    bl_idname = "autobake.confirm_bakes"
    bl_label = "Confirm & Clear"
    bl_description = "Confirming the bakes will hide the panel above, and remove all its items"
    bl_options = {'INTERNAL', 'UNDO'}

    def invoke(self, context, event):
        if event.ctrl or not context.scene.autobake_properties.ab_confirm_bake_results:
            return self.execute(context)
        else:
            return context.window_manager.invoke_confirm(self, event)

    def execute(self, context):
        scene = context.scene
        scene.autobake_queuelist.clear()
        scene.autobake_objectqueuelist.clear()
        
        return {'FINISHED'}
    
    
class AUTOBAKE_OT_NextObject(Operator):
    bl_idname = "autobake.next_object"
    bl_label = "Next Object"
    bl_description = "Continue the bake with the next object in the queue"
    bl_options = {'INTERNAL', 'UNDO'}

    def invoke(self, context, event):
        if event.ctrl or not context.scene.autobake_properties.ab_confirm_next_object:
            return self.execute(context)
        else:
            return context.window_manager.invoke_confirm(self, event)

    def execute(self, context):
        scene = context.scene
        
    # Restore Bake Items
        for item in scene.autobake_queuelist:
            item.Enabled = True
            
            if item.Status in ['Baked', 'Exported', 'Failed']:
                item.Status = 'Pending'
                item.Icon = 'PREVIEW_RANGE'
                item.name = f"{item.Type} " + (f"{item.Multiplier:.2f}" if scene.autobake_properties.ab_udim_bake else f"{item.Size}") + f" {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')

    # Restore Bake Queue Order
        global bake_order
        for index, item in enumerate(bake_order):
            for index_, item_ in enumerate(scene.autobake_queuelist):
                if (f"{item_.Type} " + (f"{item_.Multiplier:.2f}" if is_udim_bake else f"{item_.Size}")) == item:
                    scene.autobake_queuelist.move(index_, index)
                    break


        global next_object_bake
        next_object_bake = True
        
        global bake_status
        if bake_status == 'PAUSED':
            bake_status = 'INPROGRESS'
        
        return {'FINISHED'}


class AUTOBAKE_OT_CancelBake(Operator):
    bl_idname = "autobake.cancel_bake"
    bl_label = "Cancel"
    bl_description = "Cancel all the pending items in the queue"
    bl_options = {'INTERNAL'}

    def invoke(self, context, event):
        if event.ctrl or not context.scene.autobake_properties.ab_confirm_cancel_bake:
            return self.execute(context)
        else:
            return context.window_manager.invoke_confirm(self, event)

    def execute(self, context):
        scene = context.scene
        
        for item in scene.autobake_queuelist:
            if item.Status == 'Pending':
                item.Enabled = False
                item.Status = 'Canceled'
                item.name = f"{item.Type} {item.Size} {item.Status} " + ('Enabled' if item.Enabled else 'Disabled')
                item.Icon = 'CANCEL'
                
        global bake_results   
         
        for item in scene.autobake_objectqueuelist:
            if item.Status == 'Pending':
                item.Enabled = False
                item.Status = 'Canceled'
                item.name = f"{item.Object.name} {item.Status} " + ('Enabled' if item.Enabled else 'Disabled')
                item.Icon = 'CANCEL'
        
                results = []
                for item_ in scene.autobake_queuelist:
                    results.append({f"{item_.Type} " + (f"{item_.Multiplier:.2f}" if is_udim_bake else f"{item_.Size}"): {'Status': 'Canceled', 'Icon': 'CANCEL', 'Error': '' ,'Cancel': True}})
                bake_results[item.Object.name] = results
        
        global next_object_bake
        next_object_bake = True
                
        global bake_status
        bake_status = 'INPROGRESS'
        
        return {'FINISHED'}                            


class AUTOBAKE_OT_PauseBake(Operator):
    bl_idname = "autobake.pause_bake"
    bl_label = "Pause"
    bl_description = "Waits for the current bake, then pauses the process and keeps the queue"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        global bake_status
        bake_status = 'PAUSED'
        return {'FINISHED'}


class AUTOBAKE_OT_ResumeBake(Operator):
    bl_idname = "autobake.resume_bake"
    bl_label = "Resume"
    bl_description = "Continues the queue"
    bl_options = {'INTERNAL', 'UNDO'}

    def execute(self, context):
        scene = context.scene
        
        global next_object_bake
        global bake_status
        bake_status = 'INPROGRESS'
        
        if not next_object_bake:
            for item in scene.autobake_queuelist:
                item.Enabled = True
                if item.Status in ['Baked', 'Exported', 'Failed']:
                    item.Status = 'Pending'
                    item.Icon = 'PREVIEW_RANGE'
                    item.name = f"{item.Type} " + (f"{item.Multiplier:.2f}" if scene.autobake_properties.ab_udim_bake else f"{item.Size}") + f" {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')

            next_object_bake = True
        
        return {'FINISHED'}


class AUTOBAKE_OT_BakeStart(Operator):
    bl_idname = "autobake.start_bake"
    bl_label = "Start Bake"
    bl_description = "Generates a baking queue based on the 'Bake List' and initiates the texture baking process. Once a bake is completed, the texture will be automatically exported if the export checkbox is selected. Otherwise, the texture will be added to a list for future export using the 'Export' button below"
    bl_options = {'INTERNAL', 'UNDO'}

    clear_list : BoolProperty(options = set(), default=False, description='Remove all the unsaved textures from the export list')

    bake_type_info = {
                    "Base Color": ("EMIT", "Color"),
                      "Metallic": ("EMIT", "Float"),
                     "Roughness": ("EMIT", "Float"),
                           "IOR": ("EMIT", "Float"),
                         "Alpha": ("EMIT", "Float"),
                        "Normal": ("NORMAL", "Vector"),
                       "Tangent": ("NORMAL", "Vector"),
                       
                   "Anisotropic": ("EMIT", "Float"),
          "Anisotropic Rotation": ("EMIT", "Float"),
    
                         "Sheen": ("EMIT", "Float"),
                    "Sheen Tint": ("EMIT", "Color" if bpy.app.version >= (4, 0, 0) else "Float"),
                  "Sheen Weight": ("EMIT", "Float"),
               "Sheen Roughness": ("EMIT", "Float"),
                         
                      "Coat IOR": ("EMIT", "Float"),
                     "Coat Tint": ("EMIT", "Color"),
                   "Coat Weight": ("EMIT", "Float"),
                   "Coat Normal": ("NORMAL", "Vector"),
                "Coat Roughness": ("EMIT", "Float"),
                     "Clearcoat": ("EMIT", "Float"),
           "Clearcoat Roughness": ("EMIT", "Float"),
              "Clearcoat Normal": ("NORMAL", "Vector"),
                      
             "Subsurface Radius": ("NORMAL", "Vector"),
                "Subsurface IOR": ("EMIT", "Float"),
         "Subsurface Anisotropy": ("EMIT", "Float"),
                    "Subsurface": ("EMIT", "Float"),
              "Subsurface Scale": ("EMIT", "Float"),
              "Subsurface Color": ("EMIT", "Color"),
             "Subsurface Weight": ("EMIT", "Float"),
              
                  "Transmission": ("EMIT", "Float"),
           "Transmission Weight": ("EMIT", "Float"),
        "Transmission Roughness": ("EMIT", "Float"),
                      
                      "Specular": ("EMIT", "Float"),
                 "Specular Tint": ("EMIT", "Color" if bpy.app.version >= (4, 0, 0) else "Float"),
            "Specular IOR Level": ("EMIT", "Float"),
            
             "Emission Strength": ("EMIT", "Float"),
                      "Emission": ("EMIT", "Color"),
                "Emission Color": ("EMIT", "Color"),

               "Channel Packing": ("EMIT", "Channel Packing"),
               "Color Attribute": ("EMIT", "Color Attribute"),
             "Ambient Occlusion": ("EMIT", "Ambient Occlusion"),
                    "Pointiness": ("EMIT", "Pointiness"),
                 "Displacement ": ("EMIT", "Displacement"),

                       "Normals": ("NORMALS", "Multires"),
                  "Displacement": ("DISPLACEMENT", "Multires"),

                      "Combined": ("COMBINED", "Combined"),
            "Ambient Occlusion ": ("AO", "AO"),
                       "Normal ": ("NORMAL", "Standard"),
                    "Roughness ": ("ROUGHNESS", "Standard"),
                        "Glossy": ("GLOSSY", "Glossy"),
                      "Position": ("POSITION", "Standard"),
                        "Shadow": ("SHADOW", "Standard"),
                       "Diffuse": ("DIFFUSE", "Diffuse"),
                            "UV": ("UV", "UV"),
                  "Transmission": ("TRANSMISSION", "Transmission"),
                   "Environment": ("ENVIRONMENT", "Standard"),
                          "Emit": ("EMIT", "Standard")}
                      
    gridscalex = 0
    type_name = ""
    label = ()
    img = ()
    phase_bake_locked = False
    phase_export_locked = True
    phase_finished_locked = True
    bake_canceled = False
    item_index = 0
    render_disabled = {}
    node_tiling = []
    udim_tiles = []
    name_structure = {}
    image_scale_name = ''
    antialiasing_method = ''
    scale_reset = {}
    prefix = ''
    img_start_loc = {}
    source_objects = []
    selected_to_active = False
    bake_order = []
    final_material = None
    sampling = []
    margin_original = 0

    def ab_bake_complete_multires(self):
        if not self.img.is_dirty:
            return 0.5
        else:
            self.ab_bake_complete(None, None)
       
    def ab_bake_complete(self, obj, empty):
        context = bpy.context
        scene = context.scene
        abp = scene.autobake_properties
        
        if not self.bake_canceled:
            if abp.ab_report_bake_end:
                self.report({'INFO'}, f"Auto Bake: Texture '{self.img.name}' is finished baking.")
            
        remove_handlers_timers()
        restore_nodes()
        
        self.phase_export_locked = False

    def ab_bake_cancel(self, object, empty):
        self.bake_canceled = True
        context = bpy.context
        scene = context.scene
        abp = scene.autobake_properties
        
        if abp.ab_report_bake_end:
            self.report({'ERROR'}, f"Auto Bake: '{self.img.name}' was forced to stop baking...")
            
        self.update_queue_item(context, 'Baking', False, 'Failed', 'ERROR', fail_msg = 'Bake was forced to stop!')
        
        remove_handlers_timers()
        restore_nodes()
        
        self.phase_finished_locked = False
        
    def update_queue_item(self, context, edit_status, set_enabled, set_status, set_icon, fail_msg=''):
        scene = context.scene
        for item in scene.autobake_queuelist:
            if item.Status == edit_status:
                item.Enabled = set_enabled
                item.Status = set_status
                item.Error = fail_msg
                item.Icon = set_icon
                item.name = f"{item.Type} " + (f"{item.Multiplier:.2f}" if scene.autobake_properties.ab_udim_bake else f"{item.Size}") + f" {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')
           
    def bake_item_check_fail(self, context, error):
        scene = context.scene
        abp = scene.autobake_properties
        
        self.phase_bake_locked = True
        self.phase_finished_locked = False
        self.update_queue_item(context, 'Baking', False, 'Failed', 'ERROR', fail_msg=error)
        
        if abp.ab_move_finished_bake:
            scene.autobake_queuelist.move(self.item_index, len(scene.autobake_queuelist)-1)
        
        
    def invoke(self, context, event):
        scene = context.scene
        abp = scene.autobake_properties
        
        if (abp.ab_export_list == 'Ask' and len(scene.autobake_imageexport) > 0) or (abp.ab_start_popup_settings and any(setting for setting in [abp.ab_start_popup_final_object, abp.ab_start_popup_object_offset, abp.ab_start_popup_final_material, abp.ab_start_popup_final_shader, abp.ab_start_popup_texture_apply, abp.ab_start_popup_export_textures, abp.ab_start_popup_export_objects, abp.ab_start_popup_selected_to_active, abp.ab_start_popup_keep_textures])):
            return context.window_manager.invoke_props_dialog(self, width=275)
        
        elif abp.ab_export_list == 'Clear':
            scene.autobake_imageexport.clear()
            
        if event.ctrl or not context.scene.autobake_properties.ab_confirm_start_bake:
            return self.execute(context)
        else:
            return context.window_manager.invoke_confirm(self, event)


    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties

    # Settings
        if abp.ab_start_popup_settings and any(setting for setting in [abp.ab_start_popup_final_object, abp.ab_start_popup_object_offset, abp.ab_start_popup_final_material, abp.ab_start_popup_final_shader, abp.ab_start_popup_texture_apply, abp.ab_start_popup_export_textures, abp.ab_start_popup_export_objects, abp.ab_start_popup_selected_to_active, abp.ab_start_popup_keep_textures]):
            flow = layout.grid_flow(align=True, columns=1)

        # Material
            if abp.ab_start_popup_final_material:
                box = flow.box()
                row = box.row()
                row.alignment = 'CENTER'
                row.label(text='• Material •')
                
                grid = box.grid_flow(align=False, columns=1)
                grid.prop(abp, 'ab_final_material', text='Create Final      ', icon='MATERIAL')
                
                if abp.ab_start_popup_final_shader:
                    row = grid.row(align=True)
                    row.active = abp.ab_final_material
                    row.prop(abp, 'ab_final_shader', text='')
                
                if abp.ab_start_popup_texture_apply:
                    row = grid.row(align=True)
                    row.active = abp.ab_final_material
                    row.prop(abp, 'ab_apply_textures', text='')
            
        # Object
            if abp.ab_start_popup_final_object:
                box = flow.box()
                row = box.row()
                row.alignment = 'CENTER'
                row.label(text='• Object •')
                
                grid = box.grid_flow(align=False, columns=1)
                    
                grid.prop(abp, 'ab_final_object', text='Create Final      ', icon='OBJECT_DATAMODE')
                
                if abp.ab_start_popup_export_objects:
                    row = grid.row(align=True)
                    row.active = abp.ab_final_object
                    
                    row_b = row.row(align=True)
                    row_b.alignment = 'LEFT'
                    row_b.prop(abp, 'ab_offset_direction', text='')
                    row.prop(abp, 'ab_object_offset', text='')
            
                if abp.ab_start_popup_export_objects:
                    row = grid.row(align=True)
                    row.active = abp.ab_final_object
                    row.prop(abp, 'ab_export_object', text='Export Objects      ', icon='EXPORT')
                    
                if abp.ab_start_popup_selected_to_active:
                    grid.prop(scene.render.bake, 'use_selected_to_active', text='Selected to Active      ', icon='CHECKBOX_HLT' if scene.render.bake.use_selected_to_active else 'CHECKBOX_DEHLT')
            
        # Textures
            if abp.ab_start_popup_selected_to_active or abp.ab_start_popup_export_textures or abp.ab_start_popup_keep_textures:
                box = flow.box()
                row = box.row()
                row.alignment = 'CENTER'
                row.label(text='• Texture •')
                
                grid = box.grid_flow(align=False, columns=1)
                    
                if abp.ab_start_popup_export_textures:
                    grid.prop(abp, 'ab_texture_export', text='Export Textures      ', icon='EXPORT')
                
                if abp.ab_start_popup_keep_textures:
                    grid.prop(abp, 'ab_remove_imagetextures', text='Remove Textures      ', icon='TRASH')
            
    # Clear List
            if len(scene.autobake_imageexport) > 0:
                grid.prop(self, 'clear_list', text='Clear Export List      ', icon='X')
        else:
            if len(scene.autobake_imageexport) > 0:
                layout.prop(self, 'clear_list', text='Clear Export List      ', icon='X')
            

    def execute(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
    # Resets
        scene.autobake_queuelist.clear()
        scene.autobake_objectqueuelist.clear()
        
        self.node_tiling.clear()
        
        if self.clear_list:
            scene.autobake_imageexport.clear()
        
    # List has Item
        if abp.ab_udim_bake:
            if len(scene.autobake_udimlist) < 1:
                self.report({'ERROR'}, "Auto Bake: 'Bake List' is empty.")
                return {'CANCELLED'}
            if len(scene.autobake_udimtilelist) < 1:
                self.report({'ERROR'}, "Auto Bake: 'Bake List' is empty.")
                return {'CANCELLED'}
        else:
            if len(scene.autobake_bakelist) < 1:
                self.report({'ERROR'}, "Auto Bake: 'Bake List' is empty.")
                return {'CANCELLED'}
            
    # List has Enabled
        if abp.ab_udim_bake:
            if not any(item.Gate for item in scene.autobake_udimlist):
                self.report({'ERROR'}, "Auto Bake: 'Bake List' has no enabled item.")
                return {'CANCELLED'}
            if not any(item.Gate for item in scene.autobake_udimtilelist):
                self.report({'ERROR'}, "Auto Bake: 'UDIM Tiles' list has no enabled item.")
                return {'CANCELLED'}
        else:
            if not any(item.Gate for item in scene.autobake_bakelist):
                self.report({'ERROR'}, "Auto Bake: 'Bake List' has no enabled item.")
                return {'CANCELLED'}

    # Selected to Active
        self.selected_to_active = scene.render.bake.use_selected_to_active
            
    # Object Requests (Non StA)
        if not self.selected_to_active:
            if abp.ab_report_requests:
                self.report({'INFO'}, "Auto Bake: Collecting objects to bake...")
            for obj in bpy.context.selected_objects:
                if obj.type == 'MESH':
                    item = scene.autobake_objectqueuelist.add()
                    item.Object = obj
                    item.name = f"{item.Object.name} {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')
        
        # Object Requests Report (Non StA)
            if len(scene.autobake_objectqueuelist) > 0:
                if abp.ab_report_requests:
                    self.report({'INFO'}, f"Auto Bake: ...{len(scene.autobake_objectqueuelist)} objects are added to the object queue list.")
            else:
                self.report({'ERROR'}, f"Auto Bake: Could not collect any objects...must have objects selected in the viewport")
                return {'CANCELLED'}
                
    # Object Requests (StA)
        elif self.selected_to_active:
            if abp.ab_report_requests:
                self.report({'INFO'}, "Auto Bake: Collecting objects to bake...")
            if bpy.context.active_object in bpy.context.selected_objects and bpy.context.active_object.type == 'MESH':
                item = scene.autobake_objectqueuelist.add()
                item.Object = bpy.context.active_object
                item.name = f"{item.Object.name} {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')
                
            self.source_objects = []
            for obj in bpy.context.selected_objects:
                if obj != bpy.context.active_object and obj.type in ['MESH', 'CURVE', 'SURFACE', 'META', 'FONT']:
                    self.source_objects.append(obj)
            
        # Object Requests Report (StA)
            if len(scene.autobake_objectqueuelist) > 0 and len(self.source_objects) > 0:
                if abp.ab_report_requests:
                    self.report({'INFO'}, f"Auto Bake: 1 target with {len(self.source_objects)} source objects are added to the object queue list...")
            else:
                self.report({'ERROR'}, f"Auto Bake: Not enough selected objects... must have a target and one or more source objects selected")
                return {'CANCELLED'}
                
        # Has Cage
            if scene.render.bake.use_cage and scene.render.bake.cage_object is None:
                self.report({'ERROR'}, "Auto Bake: Must have a set 'Cage Object' if 'Cage' is 'True'.")
                return {'CANCELLED'}
    
    # Bake Requests
        if abp.ab_report_requests:
            self.report({'INFO'}, "Auto Bake: Collecting bake requests...")
        if abp.ab_udim_bake:
            for item in scene.autobake_udimlist:
                if item.Gate and not any(item.Type == queue.Type and item.Size == queue.Multiplier for queue in scene.autobake_queuelist):
                    item_new = scene.autobake_queuelist.add()
                    item_new.Type = item.Type
                    item_new.Multiplier = item.Size
                    item_new.name = f"{item_new.Type} {item_new.Multiplier:.2f} {item_new.Status} Enabled"
            self.udim_tiles = []
            for item in scene.autobake_udimtilelist:
                if item.Gate and not any(item.UDIM == tile[0] for tile in self.udim_tiles):
                    self.udim_tiles.append((item.UDIM, item.Size, item.Label))
        else:
            for item in scene.autobake_bakelist:
                if item.Gate and not any(item.Type == queue.Type and item.Size == queue.Size for queue in scene.autobake_queuelist):
                    item_new = scene.autobake_queuelist.add()
                    item_new.Type = item.Type
                    item_new.Size = item.Size
                    item_new.name = f"{item_new.Type} {item_new.Size} {item_new.Status} Enabled"
               
    # Bake Requests Report
        if len(scene.autobake_queuelist) > 0:
            if abp.ab_report_requests:
                self.report({'INFO'}, f"Auto Bake: ...{len(scene.autobake_queuelist)} bake requests are added to the queue.")
        else:
            self.report({'ERROR'}, f"Auto Bake: Could not collect any bake request...")
            return {'CANCELLED'}
        
    # Variable Set
        self.img_start_loc = {}
        self.gridscalex = math.ceil(math.sqrt(len(scene.autobake_queuelist)))
        self.scale_reset = {}
        
        global bake_status
        bake_status = 'INPROGRESS'
        
        global is_udim_bake
        is_udim_bake = abp.ab_udim_bake
        
        global next_object_bake
        next_object_bake = True
        
        global bake_results
        bake_results = {}
        
        global bake_order
        bake_order = []
        for item in scene.autobake_queuelist:
            bake_order.append(f"{item.Type} " + (f"{item.Multiplier:.2f}" if is_udim_bake else f"{item.Size}"))
        
    # Modal
        self._timer = bpy.context.window_manager.event_timer_add(.5, window=context.window)
        bpy.context.window_manager.modal_handler_add(self)
        
        return {'RUNNING_MODAL'}
    

    def modal(self, context, event):
        if event.type == 'TIMER':
            scene = context.scene
            bake = scene.render.bake
            abp = scene.autobake_properties
            
            global bake_status
            global next_object_bake
            
# PHASE: Pause
            if (bake_status == 'PAUSED' or not next_object_bake) and self.phase_finished_locked and self.phase_export_locked:
                return{'PASS_THROUGH'}
            
            global bake_results
            
# PHASE: Bake
            if not self.phase_bake_locked and not bpy.app.is_job_running('OBJECT_BAKE'):
                
            # No Active Bake Item
                if not any(item.Enabled for item in scene.autobake_queuelist):
                    self.phase_bake_locked = True
                    self.phase_finished_locked = False
                    return {'PASS_THROUGH'}
                
            # Active Target Object
                if any(item.Status == 'Baking' for item in scene.autobake_objectqueuelist):
                    for item in scene.autobake_objectqueuelist:
                        if item.Status == 'Baking':
                            target_obj = item.Object
                            break
                        
                elif any(item.Enabled for item in scene.autobake_objectqueuelist):
                    for index, item in enumerate(scene.autobake_objectqueuelist):
                        if item.Enabled:
                            item.Enabled = False
                            
                            if item.Status == 'Pending':
                                item.Status = 'Baking'
                                item.Icon = 'RENDER_STILL'
                                item.name = f"{item.Object.name} {item.Status} " + ('Enabled' if item.Enabled else 'Disabled')
                                target_obj = item.Object
                                
                            # Start Report
                                if abp.ab_report_object_start:
                                    self.report({'INFO'}, f"Auto Bake: Object '{item.Object.name}' has been started baking.")

                            # Final Material
                                if abp.ab_final_material:
                                    if not abp.ab_shared_textures or all(item.Status in ['Baking', 'Pending'] for item in scene.autobake_objectqueuelist):
                                        final_mat_name = f"{target_obj.name}" if not abp.ab_shared_textures or abp.ab_prefix == '' else abp.ab_prefix
                                        self.final_material = bpy.data.materials.new(name=final_mat_name)
                                        if self.final_material.name != final_mat_name:
                                            name_switch = str(self.final_material.name)
                                            self.final_material.name = ''
                                            bpy.data.materials[final_mat_name].name = name_switch
                                            self.final_material.name = final_mat_name
                                        self.final_material.use_nodes = True
                                        
                                        if abp.ab_final_shader != 'BSDF_PRINCIPLED':
                                            for node in self.final_material.node_tree.nodes:
                                                if node.type == 'BSDF_PRINCIPLED':
                                                    final_shader = self.final_material.node_tree.nodes.new(abp.ab_final_shader)
                                                    final_shader.location = node.location
                                                    self.final_material.node_tree.links.new(final_shader.outputs[0], node.outputs[0].links[0].to_socket)
                                                    self.final_material.node_tree.nodes.remove(node)
                                                    break
                                else:
                                    self.final_material = None
                                    
                            # Move Active Bake
                                if abp.ab_move_active_bake:
                                    scene.autobake_objectqueuelist.move(index, 0)

                                break
                            
                            elif item.Status == 'Canceled':
                                item.name = f"{item.Object.name} {item.Status} " + ('Enabled' if item.Enabled else 'Disabled')
                                
                            # Store Bake Results
                                results = []
                                for item_ in scene.autobake_queuelist:
                                    if item_.Status == 'Pending':
                                        item_.Enabled = False
                                        item_.Status = 'Canceled'
                                        item_.Icon = 'CANCEL'
                                        item_.Error = ''
                                    
                                    results.append({f"{item_.Type} " + (f"{item_.Multiplier:.2f}" if is_udim_bake else f"{item_.Size}"): {'Status': item_.Status, 'Icon': item_.Icon, 'Error': item_.Error, 'Cancel': item_.Cancel}})
                                bake_results[item.Object.name] = results
                                
                            # Move Item
                                if abp.ab_move_finished_bake:
                                    scene.autobake_objectqueuelist.move(index, len(scene.autobake_objectqueuelist)-1)

                                return {'PASS_THROUGH'}
                        
                else:
                    self.phase_bake_locked = True
                    self.phase_finished_locked = False
                    
                    return {'PASS_THROUGH'}

            # Active Source Object
                if not self.selected_to_active:
                    self.source_objects = []
                    self.source_objects.append(target_obj)
                    
            # Is Object Ready
                is_object_ready = obj_bake_ready(self, context, target_obj, self.source_objects)
                
            # Not Ready
                if is_object_ready != True:
                    self.phase_bake_locked = True
                    self.phase_finished_locked = False
                    
                # Pending Items to Failed
                    for item in scene.autobake_queuelist:
                        self.update_queue_item(context, 'Pending', False, 'Failed', 'ERROR', is_object_ready)

                # Deactivating Canceled Items
                    for item in scene.autobake_queuelist:
                        self.update_queue_item(context, 'Canceled', False, 'Canceled', 'CANCEL', '')
                    
                    return {'PASS_THROUGH'}

            # Active Bake Item
                for index, item in enumerate(scene.autobake_queuelist):
                    if item.Enabled:
                        item.Enabled = False
                        if item.Status == 'Pending':
                            item.Status = 'Baking'
                            item.Icon = 'RENDER_STILL'
                            item.name = f"{item.Type} " + (f"{item.Multiplier:.2f}" if abp.ab_udim_bake else f"{item.Size}") + f" {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')
                            
                            global finished_bake_count
                            finished_bake_count = len(scene.autobake_queuelist) - (sum(1 for item in scene.autobake_queuelist if item.Status == 'Pending'))
                            
                            baketype, label = self.bake_type_info[item.Type]
                            self.label = label
                            
                            self.image_scale_name = f"{item.Multiplier:.2f}" if is_udim_bake else f"{item.Size}"
                            
                            self.item_index = index
                            
                            if abp.ab_move_active_bake:
                                scene.autobake_queuelist.move(index, 0)
                                self.item_index = 0
                                item = scene.autobake_queuelist[self.item_index]
                                
                            self.phase_bake_locked = True
                            break
                        
                    # Not a valid bake item
                        elif item.Status == 'Canceled':
                            item.name = f"{item.Type} " + (f"{item.Multiplier:.2f}" if abp.ab_udim_bake else f"{item.Size}") + f" {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')

                            if abp.ab_move_finished_bake:
                                scene.autobake_queuelist.move(index, len(scene.autobake_queuelist)-1)
                                
                            return {'PASS_THROUGH'}
                        
            # Bake Item Check
                if label in ['Float', 'Color', 'Vector']:
                    pass
                        
            # Object specific
                elif label == "Color Attribute":
                    if not self.selected_to_active:
                        if (len(target_obj.data.color_attributes)) < 1:
                            if abp.ab_report_bake_error:
                                self.report({'ERROR'}, f"Auto Bake: Skipping '{item.Type} - {self.image_scale_name}'... {target_obj.name} does not have any color attribute.")
                            self.bake_item_check_fail(context, error=f"Object does not have any color attribute!")
                            return{"PASS_THROUGH"}
                    else: # Selected to Active
                        if any((len(obj.data.color_attributes)) < 1 for obj in self.source_objects):
                            if abp.ab_report_bake_error:
                                self.report({'ERROR'}, f"Auto Bake: Skipping '{item.Type} - {self.image_scale_name}'... All source object must have color attributes.")
                            self.bake_item_check_fail(context, error="All source object must have color attributes!")
                            return{"PASS_THROUGH"}

                elif label == "Multires":
                    if not self.selected_to_active:
                        if not any(modifier.type == 'MULTIRES' for modifier in target_obj.modifiers):
                            if abp.ab_report_bake_error:
                                self.report({'ERROR'}, f"Auto Bake: Skipping '{item.Type} - {self.image_scale_name}'... Multires data bake requires multi-resolution object.")
                            self.bake_item_check_fail(context, error="Object must have 'Multiresolution' modifier!")
                            return{"PASS_THROUGH"}
                    else: # Selected to Active
                        if abp.ab_report_bake_error:
                            self.report({'ERROR'}, f"Auto Bake: Skipping '{item.Type} - {self.image_scale_name}'... Can't use 'Selected to Active' for 'Multires' bakes.")
                        self.bake_item_check_fail(context, error="Can't use 'Selected to Active' for 'Multires' bakes!")
                        return{"PASS_THROUGH"}

            # Non-Object specific
                elif label == 'AO':
                    if self.selected_to_active and abp.ab_ao_local_only == 'COLLECTION':
                        if abp.ab_report_bake_error:
                            self.report({'ERROR'}, "Auto Bake: Skipping '{item.Type} - {self.image_scale_name}'... Can't use 'Collection' for 'Local Only' with 'Selected to Active'")
                        self.bake_item_check_fail(context, error="Can't use 'Collection' for 'Local Only' with 'Selected to Active'!")
                        return{"PASS_THROUGH"}

                elif label == "Channel Packing":
                    if abp.ab_channel_pack_r == 'None' and abp.ab_channel_pack_g == 'None' and abp.ab_channel_pack_b == 'None':
                        if abp.ab_report_bake_error:
                            self.report({'ERROR'}, f"Auto Bake: Skipping '{item.Type} - {self.image_scale_name}'... Can't bake 'Channel Packing' with all channels set to 'None'.")
                        self.bake_item_check_fail(context, error="All channels are set to 'None'!")
                        return{"PASS_THROUGH"}
                    
                elif label in ['Diffuse', 'Transmission', 'Glossy']:
                    if not (bake.use_pass_direct or bake.use_pass_indirect or bake.use_pass_color):
                        if abp.ab_report_bake_error:
                            self.report({'ERROR'}, f"Auto Bake: Skipping '{item.Type} - {self.image_scale_name}'... Bake type requires Direct, Indirect, or Color contributions to be enabled")
                        self.bake_item_check_fail(context, error="Requires Direct, Indirect, or Color contributions!")
                        return{"PASS_THROUGH"}
                    
                elif label == "Combined":
                    if not (bake.use_pass_emit or ((bake.use_pass_indirect or bake.use_pass_direct) and (bake.use_pass_diffuse or bake.use_pass_glossy or bake.use_pass_transmission))):
                        if abp.ab_report_bake_error:
                            self.report({'ERROR'}, "Auto Bake: Skipping '{item.Type} - {self.image_scale_name}'...  Combined bake requires Emit, or a light pass with Direct or Indirect contributions enabled")
                        self.bake_item_check_fail(context, error="Requires Emit, or a light pass with Direct or Indirect contributions!")
                        return{"PASS_THROUGH"}
                                        
            # Anti-aliasing Method
                self.antialiasing_method = abp.ab_scaled_antialiasing
                    
            # Image
                if not abp.ab_shared_textures or item.Image is None:
                    
                # Prefix Name
                    self.prefix = str(target_obj.name) if abp.ab_prefix == '' or (len(scene.autobake_objectqueuelist) > 1 and not abp.ab_shared_textures) else str(abp.ab_prefix) if abp.ab_prefix not in ['""', "''"] else ''
                
                # Type Name
                    type_names = {}
                    for item_ in abp.ab_baketype_name_all.split(', '):
                        type_names[item_.split(':')[0]] = item_.split(':')[1]
                    self.type_name = type_names[item.Type]
                
                # Image Name
                    name_structure = {"prefix": self.prefix, "bridge": abp.ab_bridge, "suffix": abp.ab_suffix, "type": self.type_name, "size": self.image_scale_name, "udim": '<UDIM>', "uvtile": '<UVTILE>'}
                    texture_name = abp.ab_name_structure_udim if is_udim_bake else abp.ab_name_structure
                    texture_name = texture_name.lower().format(**name_structure)
                    texture_name = re.sub('[{}]'.format(re.escape('<>:"/\|?*')), '', texture_name)

                # New Image
                    if not is_udim_bake:
                        size_final = item.Size
                        if self.antialiasing_method == 'UPSCALED': 
                            size_final = min(int(item.Size * (abp.ab_antialiasing_upscaled/100)), 65536)
                        img = bpy.data.images.new(name=texture_name, width=size_final, height=size_final, alpha=False, float_buffer=abp.ab_floatbuffer, is_data=False, tiled=is_udim_bake)
                        self.scale_reset[img] = [item.Size]
                        
                # New UDIM Image
                    else:
                        if any(udim_item[0] == 1001 for udim_item in self.udim_tiles):
                            for tile in self.udim_tiles:
                                if tile[0] == 1001:
                                    size_adjusted = min(int(tile[1] * item.Multiplier), 65536)
                                    size_final = size_adjusted
                                    if self.antialiasing_method == 'UPSCALED':
                                        size_final = min(int(size_adjusted * (abp.ab_antialiasing_upscaled/100)), 65536)
                                    img = bpy.data.images.new(name=texture_name, width=size_final, height=size_final, tiled=True, alpha=False, float_buffer=abp.ab_floatbuffer)
                                    self.scale_reset[img] = [[tile[0], int(size_adjusted)]]
                                    break
                        else:
                            img = bpy.data.images.new(name=texture_name, width=1, height=1, tiled=True, alpha=False, float_buffer=abp.ab_floatbuffer)
                            self.scale_reset[img] = [[1001, 1]]

                    # UDIM Tiles
                        for udim_item in self.udim_tiles:
                            if udim_item[0] != 1001:
                                size_adjusted = min(int(udim_item[1] * item.Multiplier), 65536)
                                size_final = size_adjusted
                                if self.antialiasing_method == 'UPSCALED':
                                    size_final = min(int(size_adjusted * (abp.ab_antialiasing_upscaled/100)), 65536)
                                        
                                with context.temp_override(edit_image=img):
                                    bpy.ops.image.tile_add(number=udim_item[0], width=size_final, height=size_final, alpha=False, float=abp.ab_floatbuffer, fill=True)
                                    self.scale_reset[img].append([udim_item[0], int(size_adjusted)])

                        if not any(udim_item[0] == 1001 for udim_item in self.udim_tiles):
                            img.tiles.remove(img.tiles[0])
                            self.scale_reset[img].remove([1001, 1])

                    # Tile label
                        if abp.ab_udim_label:
                            for tile in self.udim_tiles:
                                if tile[2] != '#Unknown':
                                    for img_tile in img.tiles:
                                        if img_tile.number == tile[0]:
                                            img_tile.label = tile[2]
                                            
                # Color Space
                    if item.Type in ["Metallic", "Roughness", "IOR", "Alpha", "Subsurface Weight", "Subsurface Scale", "Subsurface IOR", "Subsurface Anisotropy", "Specular IOR Level", "Anisotropic", "Anisotropic Rotation", "Transmission Weight", "Coat Weight", "Coat Roughness", "Coat IOR", "Sheen Weight", "Sheen Roughness", "Emission Strength", "Displacement", "Roughness ", "Glossy", "Shadow", "Ambient Occlusion", "Subsurface", "Specular", "Sheen", "Clearcoat", "Clearcoat Roughness", "Transmission Roughness", "Channel Packing"]:
                        img.colorspace_settings.name = abp.ab_color_space_float
                    elif item.Type in ["Base Color", "Specular Tint", "Coat Tint", "Sheen Tint", "Emission Color", "Color Attribute", "Combined", "Diffuse", "Transmission", "Environment", "Emit", "Subsurface Color", "Emission"]:
                        img.colorspace_settings.name = abp.ab_color_space_color
                    elif item.Type in ["Normal", "Subsurface Radius", "Tangent", "Coat Normal", "Normals", "Normal ", "Position", "UV", "Displacement ", "Clearcoat Normal"]:
                        img.colorspace_settings.name = abp.ab_color_space_vector

                # Force Name
                    if img.name != texture_name:
                        name_switch = str(img.name)
                        img.name = ''
                        bpy.data.images[texture_name].name = name_switch
                        img.name = texture_name
                    
                # Texture Image
                    self.img = img

                # Fake User
                    self.img.use_fake_user = abp.ab_textures_fakeuser
                    
                # Shared Texture
                    if abp.ab_shared_textures:
                        item.Image = self.img
                        item.Prefix = self.prefix
                        item.Type_Name = self.type_name
                    
            # Shared Data
                else:
                    self.img = item.Image
                    self.prefix = item.Prefix
                    self.type_name = item.Type_Name

        # Shader Alerts
                frames = {}

        # Node Bake Setup: Source
                global delete_nodes
                global reconnect_nodes

                if label in ["Float", "Color", "Vector", "Channel Packing"]:
                    for obj in self.source_objects:
                        for slot in obj.material_slots:
                            if slot.material is not None and slot.material.use_nodes:
                                shader_nodes = []
                                
                                node_trees = []
                                node_trees.append(slot.material.node_tree)

                            # Searching: Node Trees
                                for tree in node_trees:
                                    branches = []
                                    
                                # Searching: Branch Start
                                    for node in tree.nodes:
                                        if node.type in ['OUTPUT_MATERIAL', 'GROUP_OUTPUT'] and node.is_active_output:
                                        # Material Output
                                            if node.type == 'OUTPUT_MATERIAL':
                                                
                                            # Material Output: Detach Displacement (Material Output)
                                                if node.inputs[2].is_linked:
                                                    reconnect_nodes.append((tree, node.inputs[2].links[0].from_socket, node.inputs[2]))
                                                    tree.links.remove(node.inputs[2].links[0])
                                            
                                            # Material Output: Add Branch
                                                if node.inputs[0].is_linked:
                                                    if node.inputs[0].links[0].from_node not in branches:
                                                        branches.append(node.inputs[0].links[0].from_node)
                                                    
                                        # Group Output
                                            elif node.type == 'GROUP_OUTPUT':
                                                
                                            # Group Output: Add Branch
                                                for input in node.inputs:
                                                    if input.is_linked:
                                                        if input.links[0].from_node not in branches:
                                                            branches.append(input.links[0].from_node)
                                            break
                                        
                                # Branching
                                    for branch in branches:
                                        
                                    # Shader Found
                                        if branch.type in ['BSDF_PRINCIPLED', 'EMISSION', 'BSDF_TRANSLUCENT', 'BSDF_TRANSPARENT', 'BSDF_REFRACTION', 'BSDF_DIFFUSE', 'BSDF_GLASS', 'BSDF_SHEEN', 'BSDF_HAIR', 'BSDF_TOON', 'BSDF_GLOSSY', 'VOLUME_ABSORPTION', 'VOLUME_SCATTER', 'SUBSURFACE_SCATTERING', 'PRINCIPLED_VOLUME', 'BSDF_HAIR_PRINCIPLED', 'BSDF_VELVET', 'BSDF_ANISOTROPIC', 'EEVEE_SPECULAR']:
                                            shader_nodes.append((branch, tree))
                                            
                                    # New Tree & Branch
                                        elif branch.type == 'GROUP':
                                            for input in branch.inputs:
                                                if input.is_linked:
                                                    if input.links[0].from_node not in branches:
                                                        branches.append(input.links[0].from_node)
                                                        
                                            if branch.node_tree not in node_trees:
                                                node_trees.append(branch.node_tree)
                                                    
                                     # New Branch   
                                        else:
                                            for input in branch.inputs:
                                                if input.is_linked:
                                                    if input.links[0].from_node not in branches:
                                                        branches.append(input.links[0].from_node)
                                
                            # Source BSDF
                                for shader_node_ in shader_nodes:
                                    shader_node = shader_node_[0]
                                    nodetree = shader_node_[1]
                         
                                # Save Output Links
                                    for link in shader_node.outputs[0].links:
                                        reconnect_nodes.append((nodetree, shader_node.outputs[0], link.to_socket))
                                        
                                # Channel Packing
                                    if label == "Channel Packing":
                                        channel_packing_node = nodetree.nodes.new('ShaderNodeCombineColor')
                                        channel_packing_node.location = shader_node.location + mathutils.Vector((0, 160))
                                        
                                        for index, channel in enumerate([abp.ab_channel_pack_r, abp.ab_channel_pack_g, abp.ab_channel_pack_b]):
                                            if channel in self.bake_type_info:
                                                
                                                if channel == 'Ambient Occlusion':
                                                    ao_node = nodetree.nodes.new('ShaderNodeAmbientOcclusion')
                                                    delete_nodes.append((nodetree, ao_node))
                                                    ao_node.location = channel_packing_node.location + mathutils.Vector((-200, 150-index*150))
                                                    
                                                    ao_node.samples = abp.ab_ao_sample if abp.ab_ao_sample_use else 0
                                                    ao_node.inside = abp.ab_ao_inside
                                                    ao_node.only_local = abp.ab_ao_only_local
                                                    ao_node.inputs[1].default_value = abp.ab_ao_distance

                                                    nodetree.links.new(ao_node.outputs[1], channel_packing_node.inputs[index])
                                        
                                                elif channel == "Pointiness":
                                                    geometry_node = nodetree.nodes.new('ShaderNodeNewGeometry')
                                                    delete_nodes.append((nodetree, geometry_node))
                                                    geometry_node.location = channel_packing_node.location + mathutils.Vector((-400, 150-index*150))

                                                    contrast_node = nodetree.nodes.new('ShaderNodeBrightContrast')
                                                    delete_nodes.append((nodetree, contrast_node))
                                                    contrast_node.location = channel_packing_node.location + mathutils.Vector((-200, 150-index*150))
                                                    
                                                    contrast_node.inputs[1].default_value = abp.ab_pointiness_brightness
                                                    contrast_node.inputs[2].default_value = abp.ab_pointiness_contrast
                                                    
                                                    nodetree.links.new(geometry_node.outputs[7], contrast_node.inputs[0])
                                                    nodetree.links.new(contrast_node.outputs[0], channel_packing_node.inputs[index])
                                        
                                                else:
                                                    aliases = [channel,]    
                                                    if channel in type_aliases:
                                                        aliases.extend(type_aliases[channel])
                                                         
                                                    socketindex_cp = None
                                                    for socket_index, input_socket in enumerate(shader_node.inputs):
                                                        if input_socket.name in aliases:
                                                            socketindex_cp = socket_index
                                                            break
                                                    
                                                    if socketindex_cp is not None:
                                                        if shader_node.inputs[socketindex_cp].is_linked:
                                                            nodetree.links.new(shader_node.inputs[socketindex_cp].links[0].from_socket, channel_packing_node.inputs[index])
                                                        else:
                                                            channel_packing_node.inputs[0].default_value = shader_node.inputs[socketindex_cp].default_value
                                                
                                    # Linking & Delete     
                                        for link in shader_node.outputs[0].links:
                                            nodetree.links.new(channel_packing_node.outputs[0], link.to_socket)
                                        delete_nodes.append((nodetree, channel_packing_node))
                                        
                                    else:
                                        
                                    # Aliases
                                        aliases = [item.Type,]    
                                        if item.Type in type_aliases:
                                            aliases.extend(type_aliases[item.Type])
                                             
                                    # Socket Index
                                        socketindex = None
                                        for socket_index, input_socket in enumerate(shader_node.inputs):
                                            if input_socket.name in aliases:
                                                socketindex = socket_index
                                                break
                                        
                                        if socketindex == None:
                                            for link in shader_node.outputs[0].links:
                                                nodetree.links.remove(link)
                                        
                                        else:
                                            
                                            if label in ["Float", "Color"]:
                                                if shader_node.inputs[socketindex].is_linked:
                                                    for link in shader_node.outputs[0].links:
                                                        nodetree.links.new(shader_node.inputs[socketindex].links[0].from_socket, link.to_socket)
                                                else:
                                                    if label == "Float":
                                                        value_node = nodetree.nodes.new('ShaderNodeValue')
                                                        value_node.location = shader_node.location + mathutils.Vector((0, 90))
                                                        value_node.outputs[0].default_value = shader_node.inputs[socketindex].default_value
                                                        
                                                        for link in shader_node.outputs[0].links:
                                                            nodetree.links.new(value_node.outputs[0], link.to_socket)
                                                            
                                                        delete_nodes.append((nodetree, value_node))
                                                        
                                                    elif label == "Color":
                                                        rgb_node = nodetree.nodes.new('ShaderNodeRGB')
                                                        rgb_node.location = shader_node.location + mathutils.Vector((0, 190))
                                                        rgb_node.outputs[0].default_value = shader_node.inputs[socketindex].default_value
                                                        
                                                        for link in shader_node.outputs[0].links:
                                                            nodetree.links.new(rgb_node.outputs[0], link.to_socket)
                                                            
                                                        delete_nodes.append((nodetree, rgb_node))

                                            elif label == "Vector":
                                                diffuse_node = nodetree.nodes.new('ShaderNodeBsdfDiffuse')
                                                diffuse_node.location = shader_node.location + mathutils.Vector((0, 130))

                                                if shader_node.inputs[socketindex].is_linked:
                                                    nodetree.links.new(shader_node.inputs[socketindex].links[0].from_socket, diffuse_node.inputs[2])

                                                for link in shader_node.outputs[0].links:
                                                    nodetree.links.new(diffuse_node.outputs[0], link.to_socket)
                                                    
                                                delete_nodes.append((nodetree, diffuse_node))
                                                 
                            # Node Alert
                                if abp.ab_nodealert:
                                    tree = slot.material.node_tree
                                    frame = tree.nodes.new('NodeFrame')
                                    delete_nodes.append((slot.material.node_tree, frame))
                                    frames[slot.material] = frame
                                    frame.label = "Please wait until the bake is over..."
                                    frame.use_custom_color = True
                                    frame.color = (1, .23, .23)
                                    frame.label_size = 30
                                    
                                    for node in tree.nodes:
                                        node.parent = frame

                elif label == "Displacement":
                    for obj in self.source_objects:
                        for slot in obj.material_slots:
                            if slot.material is not None and slot.material.use_nodes:
                                
                            # Displacement
                                if not abp.ab_displacement_source:
                                    tree = slot.material.node_tree
                                    for node in tree.nodes:
                                        if node.type == 'OUTPUT_MATERIAL' and node.is_active_output:
                                            if node.inputs[0].is_linked:
                                                reconnect_nodes.append((tree, node.inputs[0].links[0].from_socket, node.inputs[0]))
                                                tree.links.remove(node.inputs[0].links[0])
                                            if node.inputs[2].is_linked:
                                                tree.links.new(node.inputs[2].links[0].from_socket, node.inputs[0])
                                            break
                                    
                            # Source Only
                                else:
                                    displacements = []
                                    node_trees = []
                                    node_trees.append(slot.material.node_tree)

                                # Searching: Node Trees
                                    for tree in node_trees:
                                        branches = []
                                        
                                    # Searching: Branch Start
                                        for node in tree.nodes:
                                            if node.type in ['OUTPUT_MATERIAL', 'GROUP_OUTPUT'] and node.is_active_output:
                                                
                                            # Material Output
                                                if node.type == 'OUTPUT_MATERIAL':
                                                    
                                                # Material Output: Detach Surface (Material Output)
                                                    if node.inputs[0].is_linked:
                                                        reconnect_nodes.append((tree, node.inputs[0].links[0].from_socket, node.inputs[0]))
                                                        tree.links.remove(node.inputs[0].links[0])
                                                
                                                # Material Output: Add Branch
                                                    if node.inputs[2].is_linked:
                                                        if node.inputs[2].links[0].from_node not in branches:
                                                            branches.append(node.inputs[2].links[0].from_node)
                                                        tree.links.new(node.inputs[2].links[0].from_socket, node.inputs[0])
                                                        
                                            # Group Output
                                                elif node.type == 'GROUP_OUTPUT':
                                                    
                                                # Group Output: Add Branch
                                                    for input in node.inputs:
                                                        if input.is_linked:
                                                            if input.links[0].from_node not in branches:
                                                                branches.append(input.links[0].from_node)
                                                break
                                            
                                    # Branching
                                        for branch in branches:
                                            
                                        # Match
                                            if branch.type in ['DISPLACEMENT', 'VECTOR_DISPLACEMENT']:
                                                displacements.append((branch, tree))
                                                
                                        # New Tree & Branch
                                            elif branch.type == 'GROUP':
                                                for input in branch.inputs:
                                                    if input.is_linked:
                                                        if input.from_node not in branches:
                                                            branches.append(input.from_node)
                                                if branch.node_tree not in node_trees:
                                                    node_trees.append(branch.node_tree)
                                                        
                                         # New Branch   
                                            else:
                                                for input in branch.inputs:
                                                    if input.is_linked:
                                                        if input.links[0].from_node not in branches:
                                                            branches.append(input.links[0].from_node)
                                            
                                # Source Displacement
                                    for displacement in displacements:
                                        displacement_node = displacement[0]
                                        nodetree = displacement[1]
                                        
                                    # Save Linking
                                        for link in displacement_node.outputs[0].links:
                                            reconnect_nodes.append((nodetree, displacement_node.outputs[0], link.to_socket))
                                            
                                        # Source Linking
                                            if displacement_node.inputs[0].is_linked:
                                                nodetree.links.new(displacement_node.inputs[0].links[0].from_socket, link.to_socket)
                                                
                                        # Remove Linking
                                            else:
                                                nodetree.links.remove(link)
                              
                            # Node Alert
                                if abp.ab_nodealert:
                                    tree = slot.material.node_tree
                                    frame = tree.nodes.new('NodeFrame')
                                    delete_nodes.append((slot.material.node_tree, frame))
                                    frames[slot.material] = frame
                                    frame.label = "Please wait until the bake is over..."
                                    frame.use_custom_color = True
                                    frame.color = (1, .23, .23)
                                    frame.label_size = 30
                                    
                                    for node in tree.nodes:
                                        node.parent = frame

                elif label in ["UV", "Color Attribute", "AO", "Ambient Occlusion", "Pointiness"]:
                    for obj in self.source_objects:
                        for slot in obj.material_slots:
                            if slot.material is not None and slot.material.use_nodes:
                                tree = slot.material.node_tree
                                for node in tree.nodes:
                                    if node.type == 'OUTPUT_MATERIAL' and node.is_active_output:
                                        if node.inputs[0].is_linked:
                                            reconnect_nodes.append((tree, node.inputs[0].links[0].from_socket, node.inputs[0]))
                                        
                                        if label == "AO":
                                            if not abp.ab_ao_use_normal:
                                                tree.links.remove(node.inputs[0].links[0])
                                        
                                        elif label == 'Ambient Occlusion':
                                            ao_node = tree.nodes.new('ShaderNodeAmbientOcclusion')
                                            delete_nodes.append((tree, ao_node))
                                            ao_node.location = node.location + mathutils.Vector((-200, 0))
                                            
                                            ao_node.samples = abp.ab_ao_sample if abp.ab_ao_sample_use else 0
                                            ao_node.inside = abp.ab_ao_inside
                                            ao_node.only_local = abp.ab_ao_only_local
                                            ao_node.inputs[1].default_value = abp.ab_ao_distance

                                            tree.links.new(ao_node.outputs[1], node.inputs[0])
                                            
                                        elif label == "Pointiness":
                                            geometry_node = tree.nodes.new('ShaderNodeNewGeometry')
                                            delete_nodes.append((tree, geometry_node))
                                            geometry_node.location = node.location + mathutils.Vector((-400, 0))

                                            contrast_node = tree.nodes.new('ShaderNodeBrightContrast')
                                            delete_nodes.append((tree, contrast_node))
                                            contrast_node.location = node.location + mathutils.Vector((-200, 0))
                                            
                                            contrast_node.inputs[1].default_value = abp.ab_pointiness_brightness
                                            contrast_node.inputs[2].default_value = abp.ab_pointiness_contrast
                                            
                                            tree.links.new(geometry_node.outputs[7], contrast_node.inputs[0])
                                            tree.links.new(contrast_node.outputs[0], node.inputs[0])
                                            
                                        elif label == 'UV':
                                            uv_node = tree.nodes.new('ShaderNodeUVMap')
                                            delete_nodes.append((tree, uv_node))
                                            uv_node.location = node.location + mathutils.Vector((-200, 0))
                                            
                                            if len(obj.data.uv_layers) > 0:
                                                if abp.ab_uv_target in obj.data.uv_layers:
                                                    for index, uv in enumerate(obj.data.uv_layers):
                                                        if abp.ab_uv_target == uv.name:
                                                            obj.data.uv_layers.active_index = index
                                                else:
                                                    for index, uv in enumerate(obj.data.uv_layers):
                                                        if uv.active_render:
                                                            obj.data.uv_layers.active_index = index
                                                            
                                            tree.links.new(uv_node.outputs[0], node.inputs[0])
                                                         
                                        elif label == 'Color Attribute':
                                            attribute_node = tree.nodes.new('ShaderNodeVertexColor')
                                            delete_nodes.append((tree, attribute_node))
                                            attribute_node.location = node.location + mathutils.Vector((-200, 0))
                                            
                                            if len(obj.data.attributes) > 0:
                                                if abp.ab_attribute_target in obj.data.color_attributes:
                                                    attribute_node.layer_name = abp.ab_attribute_target
                                                else:
                                                    attribute_node.layer_name = obj.data.attributes[obj.data.attributes.render_color_index].name
                                                    
                                            tree.links.new(attribute_node.outputs[0], node.inputs[0])
                                            
                                        break

        # Node Bake Setup: Target
                target_materials = []
                for slot in target_obj.material_slots:
                    if slot.material is not None and slot.material.use_nodes:
                        target_materials.append(slot.material)
                
                if self.final_material is not None and (not abp.ab_shared_textures or all(item.Status in ['Baking', 'Pending'] for item in scene.autobake_objectqueuelist)):
                    target_materials.append(self.final_material)

            # Prepare Materials
                for mat in target_materials:

                # Deselect Nodes
                    for node in mat.node_tree.nodes:
                        node.select = False

                # Start Location
                    if mat in self.img_start_loc:
                        startloc = self.img_start_loc[mat]
                    else:
                        startloc = mathutils.Vector((0, 0))
                        for node in mat.node_tree.nodes:
                            if node.type == 'OUTPUT_MATERIAL' and node.is_active_output:
                                self.img_start_loc[mat] = mathutils.Vector((node.location.x, node.location.y - 160))
                                startloc = self.img_start_loc[mat]
                                break
                            
                # Image Texture
                    nodetree = mat.node_tree
                    img_node = nodetree.nodes.new('ShaderNodeTexImage')
                            
                # Node Label
                    img_node.label = item.Type if abp.ab_node_label == 'Type' else self.type_name if abp.ab_node_label == 'Name' else ''
                            
                # Remove Nodes
                    if abp.ab_remove_imagetextures:
                        if mat != self.final_material:
                            delete_nodes.append((nodetree, img_node))
                        
                # Update Nodes
                    img_node.image = self.img
                    nodetree.nodes.active = img_node
                    img_node.select = True
                
                # Image Shader Alert 
                    if abp.ab_nodealert:
                        if mat in frames:
                            img_node.parent = frames[mat]
                        else:
                            tree = mat.node_tree
                            frame = tree.nodes.new('NodeFrame')
                            delete_nodes.append((mat.node_tree, frame))
                            frames[mat] = frame
                            frame.label = "Please wait until the bake is over..."
                            frame.use_custom_color = True
                            frame.color = (1, .23, .23)
                            frame.label_size = 30
                            
                            for node in tree.nodes:
                                node.parent = frame
                            
                # Node location
                    if abp.ab_node_tiling == 'OFF':
                        img_node.location.x = startloc.x
                        img_node.location.y = startloc.y

                    if not abp.ab_remove_imagetextures or mat == self.final_material:
                        if abp.ab_node_tiling == 'GRID':
                            queue_place = sum(item.Status in ['Baked', 'Exported'] for item in scene.autobake_queuelist)

                            grid_x_new = math.ceil(math.sqrt(len(scene.autobake_queuelist) - (sum(1 for item in scene.autobake_queuelist if item.Status == 'Canceled' and not item.Enabled))))
                            self.gridscalex = grid_x_new if queue_place < grid_x_new+1 else self.gridscalex
                            img_node.location.x = startloc.x + 250 * ((queue_place)-(int((queue_place)/self.gridscalex))*self.gridscalex)
                            img_node.location.y = startloc.y + (280 if abp.ab_pack_texture else 255) - ((280 if abp.ab_pack_texture else 255) * ((math.ceil(((queue_place)+1)/self.gridscalex))))
                             
                        elif abp.ab_node_tiling == 'SINGLEROW':
                            img_node.location.x = startloc.x + 250 * (sum(1 for item in scene.autobake_queuelist if item.Status in ['Baked', 'Exported']))
                            img_node.location.y = startloc.y
                            
                        elif abp.ab_node_tiling == 'ROWBYTYPE':
                            baking_bake_type = next((item.Type for item in scene.autobake_queuelist if item.Status == 'Baking'), None)
                            img_node.location.x = startloc.x + 250 * (sum(1 for item in scene.autobake_queuelist if item.Type == baking_bake_type and item.Status in ['Baked', 'Exported']))

                            if baking_bake_type not in self.node_tiling:
                                self.node_tiling.append(baking_bake_type)
                            img_node.location.y = startloc.y - (280 if abp.ab_pack_texture else 255) * self.node_tiling.index(baking_bake_type)

                        elif abp.ab_node_tiling == 'ROWBYTYPECOMBINED':
                            baking_bake_type = next((item.Type for item in scene.autobake_queuelist if item.Status == 'Baking'), None)
                            img_node.location.x = startloc.x + 250 * (sum(1 for item in scene.autobake_queuelist if item.Type == baking_bake_type and item.Status in ['Baked', 'Exported']))
                            
                            if sum(1 for item in scene.autobake_queuelist if item.Type == baking_bake_type) == 1:
                                baking_bake_type = 'Combined'
                                img_node.location.x = startloc.x + 250 * (sum(1 for Item in scene.autobake_queuelist if sum(1 for item in scene.autobake_queuelist if item.Type == Item.Type) == 1 and Item.Status in ['Baked', 'Exported']))
                            
                            if baking_bake_type not in self.node_tiling:
                                self.node_tiling.append(baking_bake_type)                             
                            img_node.location.y = startloc.y - (280 if abp.ab_pack_texture else 255) * self.node_tiling.index(baking_bake_type)
                    
            # Disable Objects Render
                if label == "AO" and abp.ab_ao_local_only != 'FALSE':
                    for object in scene.objects:
                        self.render_disabled[object] = (object.hide_render)
                        
                        if abp.ab_ao_local_only == 'SELECTED':
                            object.hide_render = (object != target_obj) if (not self.selected_to_active) else (object not in self.source_objects and object != target_obj)

                        elif abp.ab_ao_local_only == 'COLLECTION':
                            object.hide_render = object.users_collection[0] != target_obj.users_collection[0]
                        
            # Material Index
                if not self.selected_to_active:
                    target_obj.active_material_index = 0
        
            # Render Engine
                scene.render.engine = 'CYCLES'
                
            # Margin
                self.margin_original = scene.render.bake.margin
                if abp.ab_adaptive_margin:
                    scene.render.bake.margin = int(scene.render.bake.margin * ((self.img.tiles[0].size[0] if is_udim_bake else self.img.size[0]) / 1024))
                
            # Sampling
                cycles = scene.cycles
                self.sampling = [cycles.use_adaptive_sampling, cycles.adaptive_threshold, cycles.samples, cycles.adaptive_min_samples, cycles.time_limit, cycles.use_denoising, cycles.denoiser, cycles.denoising_input_passes, cycles.denoising_prefilter]
     
                if not abp.ab_sampling_use_render:
                    pick_low_sample = True
                    
                    if abp.ab_auto_pick_sampling and (item.Type in ['Combined', 'Ambient Occlusion ', 'Glossy', 'Diffuse', 'Transmission', 'Shadow', 'Environment'] or (item.Type == 'Ambient Occlusion' and not abp.ab_ao_sample_use)):
                        pick_low_sample = False
                                                       
               # Set Sample
                    cycles.use_adaptive_sampling = abp.ab_sampling_low_adaptive if pick_low_sample else abp.ab_sampling_high_adaptive
                    cycles.adaptive_threshold = abp.ab_sampling_low_noise_threshold if pick_low_sample else abp.ab_sampling_high_noise_threshold
                    cycles.samples = abp.ab_sampling_low_max if pick_low_sample else abp.ab_sampling_high_max
                    cycles.adaptive_min_samples = abp.ab_sampling_low_min if pick_low_sample else abp.ab_sampling_high_min
                    cycles.time_limit = abp.ab_sampling_low_time_limit if pick_low_sample else abp.ab_sampling_high_time_limit
                    cycles.use_denoising = abp.ab_sampling_low_denoise if pick_low_sample else abp.ab_sampling_high_denoise
                    cycles.denoiser = abp.ab_sampling_low_denoiser if pick_low_sample else abp.ab_sampling_high_denoiser
                    cycles.denoising_input_passes = abp.ab_sampling_low_passes if pick_low_sample else abp.ab_sampling_high_passes
                    cycles.denoising_prefilter = abp.ab_sampling_low_prefilter if pick_low_sample else abp.ab_sampling_high_prefilter

            # Multires Bake 
                if label == "Multires":
                    
                    if abp.ab_multires_level != -1:
                        for modifier in target_obj.modifiers:
                            if modifier.type == 'MULTIRES':
                                multires_modifier = modifier
                                multires_level = int(modifier.levels)
                                modifier.levels = abp.ab_multires_level
                                break
                    
                    scene.render.bake_type = baketype
                    scene.render.use_bake_clear = False
                    scene.render.bake_margin = bake.margin
                    scene.render.bake_margin_type = bake.margin_type
                    
                # Bake Handler
                    bpy.app.timers.register(self.ab_bake_complete_multires)

                # Bake Start
                    with context.temp_override(selected_editable_objects=self.source_objects):
                        bpy.ops.object.bake_image('INVOKE_DEFAULT')
                    
                # Reset Viewport Level
                    if abp.ab_multires_level != -1:
                        modifier.levels = multires_level
                    
            # Diffuse Bake
                else:
                    
                # Bake Handlers
                    bpy.app.handlers.object_bake_complete.append(self.ab_bake_complete)
                    bpy.app.handlers.object_bake_cancel.append(self.ab_bake_cancel)
                    
                # Bake Start
                    with context.temp_override(active_object=target_obj, selected_objects=self.source_objects):
                        bpy.ops.object.bake('INVOKE_DEFAULT',
                            type = baketype,
                            pass_filter = bake.pass_filter,
                            margin = bake.margin,
                            margin_type = bake.margin_type,
                            use_selected_to_active = self.selected_to_active,
                            max_ray_distance = bake.max_ray_distance,
                            cage_extrusion = bake.cage_extrusion,
                            cage_object = bake.cage_object.name if self.selected_to_active and bake.use_cage else '',
                            normal_space = bake.normal_space,
                            normal_r = bake.normal_r,
                            normal_g = bake.normal_g,
                            normal_b = bake.normal_b,
                            target = "IMAGE_TEXTURES",
                            use_clear = False,
                            use_cage = bake.use_cage,
                            use_split_materials = False,
                            use_automatic_name = False,
                            uv_layer = '')
                            
                if abp.ab_report_bake_start:
                    self.report({'INFO'}, f"Auto Bake: Texture '{self.img.name}' is started baking.")
                
                return {'PASS_THROUGH'}
    
# PHASE: Export
            if not self.phase_export_locked and not bpy.app.is_job_running('OBJECT_BAKE'):
                self.phase_export_locked = True
                
            # Update Item
                item = scene.autobake_queuelist[self.item_index]
                item.Status = 'Baked'
                item.name = f"{item.Type} {item.Size} {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')
                item.Icon = 'CHECKMARK'
                
            # Restore Disabled Renders 
                for obj in scene.objects:
                    if obj in self.render_disabled:
                        obj.hide_render = self.render_disabled[obj]
                self.render_disabled = {}
                        
            # Remove Timers
                if bpy.app.timers.is_registered(restore_nodes):
                    bpy.app.timers.unregister(restore_nodes)
                
            # Successful Bake
                if not self.bake_canceled:
                    
                # Anti Aliasing
                    if not abp.ab_shared_textures or not any(item.Status == 'Pending' for item in scene.autobake_objectqueuelist):
                        if self.antialiasing_method == 'UPSCALED':
                            if not is_udim_bake:
                                self.img.scale(self.scale_reset[self.img][0], self.scale_reset[self.img][0])
                            else:    
                                for index, tile in enumerate(self.img.tiles):
                                    self.img.tiles.active_index = index
                                    for tile_scale in self.scale_reset[self.img]:
                                        if tile_scale[0] == tile.number:
                                            with context.temp_override(edit_image=self.img):
                                                bpy.ops.image.resize(size=(tile_scale[1], tile_scale[1]))
                                            break
                        elif self.antialiasing_method == 'DOWNSCALED':
                            if not is_udim_bake:
                                for scale in range(abp.ab_antialiasing_repeat):
                                    self.img.scale(width=int(self.img.size[0] * (abp.ab_antialiasing_downscaled/100)), height=int(self.img.size[0] * (abp.ab_antialiasing_downscaled/100)))
                                    self.img.scale(self.scale_reset[self.img][0], self.scale_reset[self.img][0])
                            else:
                                for scale in range(abp.ab_antialiasing_repeat):
                                    for index, tile in enumerate(self.img.tiles):
                                        self.img.tiles.active_index = index
                                        for tile_scale in self.scale_reset[self.img]:
                                            if tile_scale[0] == tile.number:
                                                with context.temp_override(edit_image=self.img):
                                                    bpy.ops.image.resize(size=(int(tile.size[0] * (abp.ab_antialiasing_downscaled/100)), int(tile.size[0] * (abp.ab_antialiasing_downscaled/100))))
                                                    bpy.ops.image.resize(size=(tile_scale[1], tile_scale[1]))
                                                break

                # Connect Texture
                    if abp.ab_apply_textures != "False" and self.final_material is not None and (not abp.ab_shared_textures or all(item.Status in ['Baking', 'Pending'] for item in scene.autobake_objectqueuelist)):
                        for node in self.final_material.node_tree.nodes:
                            if node.type == 'OUTPUT_MATERIAL' and node.is_active_output and node.inputs[0].is_linked:
                                target_shader = node.inputs[0].links[0].from_node
                                
                                for index, input in enumerate(target_shader.inputs):
                                    item_aliases = [item.Type,]
                                    if item.Type in type_aliases:
                                        item_aliases.extend(type_aliases[item.Type])
                                
                                    if input.name in item_aliases:
                                        if input.is_linked:
                                            
                                            if abp.ab_apply_textures == 'Highest':
                                                if input.links[0].from_node.type == 'TEX_IMAGE' and (input.links[0].from_node.image.size[0] * input.links[0].from_node.image.size[1] < self.img.size[0] * self.img.size[1]):
                                                    input.links[0].from_node.image = self.img
                                                    
                                            elif abp.ab_apply_textures == 'Lowest':
                                                if input.links[0].from_node.type == 'TEX_IMAGE' and (input.links[0].from_node.image.size[0] * input.links[0].from_node.image.size[1] > self.img.size[0] * self.img.size[1]):
                                                    input.links[0].from_node.image = self.img
                                                
                                            elif abp.ab_apply_textures == 'First':
                                                pass
                                                
                                            elif abp.ab_apply_textures == 'Last':
                                                input.links[0].from_node.image = self.img
                                        
                                        else:
                                            img_node = self.final_material.node_tree.nodes.new('ShaderNodeTexImage')
                                            img_node.label = item.Type if abp.ab_node_label == 'Type' else self.type_name if abp.ab_node_label == 'Name' else ''
                                            img_node.image = self.img
                                            img_node.location = mathutils.Vector((target_shader.location.x - 400, target_shader.location.y))
                                            self.final_material.node_tree.links.new(img_node.outputs[0], target_shader.inputs[index])
                                        
                                        break
                                break
                            
                    if not abp.ab_shared_textures or not any(item.Status == 'Pending' for item in scene.autobake_objectqueuelist):
                        
                    # Pack Image
                        if abp.ab_pack_texture:
                            self.img.pack()

                    # Export Name
                        export_name_structure = {"prefix": self.prefix, "bridge": abp.ab_bridge, "suffix": abp.ab_suffix, "type": self.type_name, "size": self.image_scale_name, "udim": '<UDIM>', "uvtile": '<UVTILE>'}
                        export_name_structure_to_use = abp.ab_name_structure_udim_export if is_udim_bake else abp.ab_name_structure
                        export_name = export_name_structure_to_use.lower().format(**export_name_structure)
                        
                    # Export
                        if abp.ab_texture_export:
                            export_texture(self, context, self.img, export_name, self.type_name, self.label, self.prefix)
                                
                        # Update Item
                            item = scene.autobake_queuelist[self.item_index]
                            item.Status = 'Exported'
                            item.name = f"{item.Type} {item.Size} {item.Status} " + ('Enabled' if item.Enabled else 'Disabled')
                            item.Icon = 'EXPORT'
                            
                    # Export Later
                        else:
                            export_img = scene.autobake_imageexport.add()
                            export_img.name = export_name
                            export_img.Name = export_name
                            export_img.Image = self.img
                            export_img.Type = self.type_name
                            export_img.Label = self.label
                            export_img.Prefix = self.prefix

                # Move Item   
                    if abp.ab_move_finished_bake:
                        scene.autobake_queuelist.move(self.item_index, len(scene.autobake_queuelist)-1)

                self.phase_finished_locked = False

# PHASE: Finish
            elif not self.phase_finished_locked and not bpy.app.is_job_running('OBJECT_BAKE'):
                self.phase_finished_locked = True
        
            # Sampling Restore
                if self.sampling != []:
                    cycles = scene.cycles
                    cycles.use_adaptive_sampling = self.sampling[0]
                    cycles.adaptive_threshold = self.sampling[1]
                    cycles.samples = self.sampling[2]
                    cycles.adaptive_min_samples = self.sampling[3]
                    cycles.time_limit = self.sampling[4]
                    cycles.use_denoising = self.sampling[5]
                    cycles.denoiser = self.sampling[6]
                    cycles.denoising_input_passes = self.sampling[7]
                    cycles.denoising_prefilter = self.sampling[8]
                    self.sampling = []
                
            # Margin Restore
                scene.render.bake.margin = self.margin_original
                
            # Bake Check
                if any(item.Enabled for item in scene.autobake_queuelist):
                    self.phase_bake_locked = False
                else:
                        
                # Object Finished
                    for index, item in enumerate(scene.autobake_objectqueuelist):
                        if item.Status == 'Baking':
                            if any(item.Status in ['Baked', 'Exported'] for item in scene.autobake_queuelist):
                                if any(item.Status == 'Failed' for item in scene.autobake_queuelist):
                                    item.Status = 'Mixed'
                                    item.Icon = 'QUESTION'
                                    item.Error = 'Some bake items had failed for this object!'
                                else:
                                    item.Status = 'Baked'
                                    item.Icon = 'CHECKMARK'
                            else:
                                item.Status = 'Failed'
                                item.Icon = 'ERROR'
                                item.Error = 'All bake items had failed for this object!'
                            
                            item.name = f"{item.Object.name} {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')
                            
                        # Finish Report
                            if abp.ab_report_object_end:
                                self.report({'INFO'},
                                    f"Auto Bake: Object '{item.Object.name}' is finished. " +
                                    (f"Baked & Exported: {sum(1 for item in scene.autobake_queuelist if item.Status in ['Baked', 'Exported'])}; " if sum(1 for item in scene.autobake_queuelist if item.Status in ['Baked', 'Exported']) == sum(1 for item in scene.autobake_queuelist if item.Status == 'Exported') else f"Baked: {sum(1 for item in scene.autobake_queuelist if item.Status in ['Baked', 'Exported'])}; Exported: {sum(1 for item in scene.autobake_queuelist if item.Status == 'Exported')}; ") +
                                    f"Canceled: {sum(1 for item in scene.autobake_queuelist if item.Status == 'Canceled')}; Failed: {sum(1 for item in scene.autobake_queuelist if item.Status == 'Failed')}")
                                
                        # Final Material
                            if abp.ab_apply_textures != 'False' and self.final_material is not None:
                                rearrange_nodes = []
                                shader_node = None
                                for node in self.final_material.node_tree.nodes:
                                    if node.type == 'OUTPUT_MATERIAL' and node.is_active_output and node.inputs[0].is_linked:
                                        shader_node = node.inputs[0].links[0].from_node
                                        for input in shader_node.inputs:
                                            if input.is_linked and input.links[0].from_node.type == 'TEX_IMAGE':
                                                rearrange_nodes.append(input.links[0].from_node)
                                        break

                            # Y Location
                                if len(rearrange_nodes) > 1:
                                    for index, tex_node in enumerate(rearrange_nodes):
                                        tex_node.location.y = (shader_node.location.y + index * (-275)) + ((len(rearrange_nodes)/2) * 275) - 130
                                        
                            # Normal Maps
                                if shader_node is not None:
                                    nodetree = self.final_material.node_tree
                                    for input in shader_node.inputs:
                                        if input.name in ['Normal', 'Clearcoat Normal', 'Coat Normal', 'Clear Coat Normal'] and input.is_linked:
                                            normal_node = nodetree.nodes.new('ShaderNodeNormalMap')
                                            tex_node = input.links[0].from_node
                                            
                                            normal_node.location = tex_node.location + mathutils.Vector((90, -40))
                                            tex_node.location.x -= 210
                                            
                                            nodetree.links.new(normal_node.outputs[0], tex_node.outputs[0].links[0].to_socket)
                                            nodetree.links.new(tex_node.outputs[0], normal_node.inputs[1])
                                            
                                            break

                        # Final Object
                            final_object = None
                            if self.selected_to_active and abp.ab_active_as_final:
                                final_object = item.Object
                            elif abp.ab_final_object:
                                
                            #Object Data
                                final_object = item.Object.copy()
                                final_object.data = item.Object.data.copy()
                                
                            # Clear Material
                                final_object.data.materials.clear()
            
                            # Object Rename
                                object_name = str(item.Object.name) if not item.Object.name.endswith(abp.ab_object_differentiator) else str(item.Object.name)[:-len(abp.ab_object_differentiator)]
                                if abp.ab_export_object:
                                    item.Object.name = ''
                                    item.Object.data.name = ''
                                    final_object.name = object_name
                                    final_object.data.name = object_name
                                    
                            # Location
                                if abp.ab_object_location == 'Copy' and abp.ab_object_offset != 0:
                                    if abp.ab_offset_direction == 'X':
                                        final_object.location.x += abp.ab_object_offset
                                    elif abp.ab_offset_direction == 'Y':
                                        final_object.location.y += abp.ab_object_offset
                                    elif abp.ab_offset_direction == 'Z':
                                        final_object.location.z += abp.ab_object_offset
                                elif abp.ab_object_location == 'Clear':
                                    final_object.location = (0, 0, 0)
                                
                            # Collection
                                if not abp.ab_export_object or not abp.ab_export_object_remove:
                                    if abp.ab_final_collection != '':
                                        if abp.ab_final_collection not in bpy.data.collections:
                                            final_collection = bpy.data.collections.new(abp.ab_final_collection)
                                            final_collection.color_tag = abp.ab_final_collection_color
                                            context.scene.collection.children.link(final_collection)
                                            final_collection.objects.link(final_object)
                                        else:
                                            bpy.data.collections[abp.ab_final_collection].objects.link(final_object)
                                    else:
                                        for collection in item.Object.users_collection:
                                            collection.objects.link(final_object)
                                else:
                                    context.scene.collection.objects.link(final_object)
                                    
                        # Final Object
                            if final_object is not None:
                                
                            # Final Material
                                if self.final_material is not None:
                                    if self.selected_to_active:
                                        final_object.data.materials.clear()
                                    final_object.data.materials.append(self.final_material)
                                    
                            # Export Object
                                if abp.ab_export_object and (abp.ab_export_always == 'Always' or not any(item.Status == "Failed" for item in scene.autobake_queuelist)):
                                    selected_ = [final_object,]
                        
                                # Store Transform
                                    stored_loc = mathutils.Vector(final_object.location)
                                    if abp.ab_export_clear_location:
                                        final_object.location = (0,0,0)
                                    stored_rot = mathutils.Vector(final_object.rotation_euler)
                                    if abp.ab_export_clear_rotation:
                                        final_object.rotation_euler = (0,0,0)
                                    stored_scale = mathutils.Vector(final_object.scale)
                                    if abp.ab_export_clear_scale:
                                        final_object.scale = (1,1,1)
                                    
                                # Folder
                                    foldername = re.sub('[{}]'.format(re.escape('<>:"/\|?*')), '', final_object.name)
                                    foldername = foldername.strip()
                                    mainpath = os.path.join(abp.ab_filepath, foldername)
                                    
                                    if not os.path.isdir(mainpath):
                                        os.makedirs(mainpath)

                                # Export FBX
                                    if abp.ab_export_object_as == 'FBX':
                                        path_ = os.path.join(mainpath, f"{final_object.name}.fbx")
                                        
                                        with context.temp_override(active_object=final_object, selected_objects=selected_):
                                            bpy.ops.export_scene.fbx(
                                                filepath = path_,
                                                path_mode = abp.ab_export_pathmode,
                                                embed_textures = abp.ab_export_embedtextures,
                                                
                                                global_scale = abp.ab_export_scale,
                                                apply_scale_options = abp.ab_fbx_apply_scaling,
                                                axis_forward = abp.ab_fbx_forward,
                                                axis_up = abp.ab_fbx_up,
                                                apply_unit_scale = abp.ab_export_applyunit,
                                                use_space_transform = abp.ab_export_usespacetransform,
                                                mesh_smooth_type = abp.ab_export_smoothing,
                                                use_subsurf = abp.ab_export_subdivisionsurface,
                                                use_mesh_modifiers = abp.ab_export_applymodifiers,
                                                use_mesh_modifiers_render = abp.ab_export_evalmode == 'DAG_EVAL_RENDER',
                                                use_mesh_edges = abp.ab_export_looseedges,
                                                use_triangles = abp.ab_export_triangulatefaces,
                                                use_tspace = abp.ab_export_tangentspace,
                                                colors_type = abp.ab_export_vertexcolors,
                                                prioritize_active_color = abp.ab_export_prioritizeactivecolor,
                                                use_custom_props = abp.ab_export_customprops,
                                                
                                                use_selection = True,
                                                object_types = {'MESH'},
                                                bake_anim = False,
                                                use_visible = False,
                                                use_active_collection = False,
                                                batch_mode = 'OFF',
                                                use_metadata = True,
                                                bake_space_transform = False)
                                    
                                # Export OBJ
                                    elif abp.ab_export_object_as == 'OBJ':
                                        stored_selection = list(context.selected_objects)
                                        for obj in reversed(context.selected_objects):
                                            obj.select_set(False)
                                        final_object.select_set(True)
                                        
                                        path_ = os.path.join(mainpath, f"{final_object.name}.obj")
                                            
                                        bpy.ops.wm.obj_export(
                                            filepath = path_,
                                            path_mode = abp.ab_export_pathmode,
                                            
                                            global_scale = abp.ab_export_scale,
                                            forward_axis = abp.ab_obj_forward,
                                            up_axis = abp.ab_obj_up,
                                            apply_modifiers = abp.ab_export_applymodifiers,
                                            export_eval_mode = abp.ab_export_evalmode,
                                            export_colors = abp.ab_export_exportcolors,
                                            export_triangulated_mesh = abp.ab_export_triangulatedmesh,
                                            export_pbr_extensions = abp.ab_export_pbrextension,
                                            export_vertex_groups = abp.ab_export_vertexgroups,
                                            export_smooth_groups = abp.ab_export_smoothgroups,
                                            smooth_group_bitflags = abp.ab_export_groupbitflag,
                                            
                                            export_selected_objects = True,
                                            export_materials = True,
                                            export_uv = True,
                                            export_animation = False,
                                            export_normals = True,
                                            export_object_groups = False,
                                            export_material_groups = False)
                                            
                                   # Restore Selection
                                        final_object.select_set(False)
                                        for obj in stored_selection:
                                            obj.select_set(True)
                                                
                                # Export GLTF
                                    elif abp.ab_export_object_as == 'GLTF':
                                        stored_selection = list(context.selected_objects)
                                        for obj in reversed(context.selected_objects):
                                            obj.select_set(False)
                                        final_object.select_set(True)
                                        
                                        if abp.ab_gltf_format == 'GLB':
                                            path_ = os.path.join(mainpath, f"{final_object.name}.glb")
                                        else:
                                            path_ = os.path.join(mainpath, f"{final_object.name}")
                                    
                                        bpy.ops.export_scene.gltf(
                                            filepath = path_,
                                        
                                            export_format = abp.ab_gltf_format,
                                            export_copyright=abp.ab_gltf_copyright,
                                            export_import_convert_lighting_mode = abp.ab_gltf_lighting,
                                            export_original_specular = abp.ab_gltf_original_spelucar,
                                            export_yup = abp.ab_gltf_yup,
                                            
                                            export_apply = abp.ab_export_applymodifiers,
                                            export_tangents = abp.ab_gltf_tangents,
                                            export_colors = abp.ab_gltf_vertex_colors,
                                            export_attributes = abp.ab_gltf_attributes,
                                            use_mesh_edges = abp.ab_gltf_loose_edges,
                                            use_mesh_vertices = abp.ab_gltf_loose_points,

                                            export_image_format = abp.ab_gltf_images,
                                            export_texture_dir = abp.ab_gltf_texture_folder,
                                            export_keep_originals = abp.ab_gltf_keep_original,
                                            export_image_quality = abp.ab_gltf_image_quality,
                                            export_jpeg_quality = abp.ab_gltf_image_quality,
                                            export_image_add_webp = abp.ab_gltf_create_webp,
                                            export_image_webp_fallback = abp.ab_gltf_webp_fallback,
                 
                                            export_morph = abp.ab_gltf_shape_keys,
                                            export_morph_normal = abp.ab_gltf_shape_keys_normals,
                                            export_morph_tangent = abp.ab_gltf_shape_keys_tangents,
                                            export_try_sparse_sk = abp.ab_gltf_use_sparse,
                                            export_try_omit_sparse_sk = abp.ab_gltf_omitting_sparse,
                                            
                                            export_draco_mesh_compression_enable = abp.ab_gltf_compression,
                                            export_draco_mesh_compression_level = abp.ab_gltf_compression_level,
                                            export_draco_position_quantization = abp.ab_gltf_compression_position,
                                            export_draco_normal_quantization = abp.ab_gltf_compression_normal,
                                            export_draco_texcoord_quantization = abp.ab_gltf_compression_texcoord,
                                            export_draco_color_quantization = abp.ab_gltf_compression_color,
                                            export_draco_generic_quantization = abp.ab_gltf_compression_generic,
                                            
                                            use_selection = True,
                                            use_active_collection_with_nested = False,
                                            export_animations = False,
                                            export_nla_strips = False,
                                            export_skins = False,
                                            
                                            ui_tab = 'GENERAL', 
                                            gltf_export_id = "",
                                            will_save_settings = False,
                                            export_materials = 'EXPORT',
                                            use_visible = False,
                                            use_renderable = False,
                                            use_active_collection = False,
                                            use_active_scene = False,
                                            export_texcoords = True,
                                            export_normals = True,
                                            export_extras = False)
                                       
                                   # Restore Selection
                                        final_object.select_set(False)
                                        for obj in stored_selection:
                                            obj.select_set(True)
                                       
                                # Report Export
                                    if abp.ab_report_object_export:
                                        self.report({'INFO'}, f"Auto Bake: Object '{final_object.name}' has been exported.")
                                        
                                # Export Status
                                    item.Status = 'Exported'
                                    item.Icon = 'EXPORT'
                                    item.Error = ''
                                    item.name = f"{item.Object.name} {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')
                                        
                                # Remove Final
                                    if abp.ab_export_object_remove:
                                        bpy.data.objects.remove(final_object, do_unlink=True)

                                # Restore Transform
                                    elif abp.ab_export_restore_transform:
                                        final_object.location = stored_loc
                                        final_object.rotation_euler = stored_rot
                                        final_object.scale = stored_scale
                                    
                        # Restore Names
                            if abp.ab_final_object and not self.selected_to_active:
                                if item.Status == 'Exported' and abp.ab_export_object_remove:
                                    item.Object.name = object_name
                                    item.Object.data.name = object_name
                                    
                                elif abp.ab_object_differentiator != '':
                                    diff_name = object_name + abp.ab_object_differentiator
                                    
                                    if abp.ab_object_keep_name == 'Original':
                                        final_object.name = diff_name
                                        final_object.data.name = diff_name
                                        item.Object.name = object_name
                                        item.Object.data.name = object_name
                                        
                                        if final_object.name != diff_name:
                                            bpy.data.objects[diff_name].data.name = final_object.name
                                            bpy.data.objects[diff_name].name = final_object.name
                                            final_object.name = diff_name
                                            final_object.data.name = diff_name
                                            
                                    elif abp.ab_object_keep_name == 'Final':
                                        item.Object.name = diff_name
                                        item.Object.data.name = diff_name
                                        final_object.name = object_name 
                                        final_object.data.name = object_name
                                        
                                        if item.Object.name != diff_name:
                                            bpy.data.objects[diff_name].data.name = item.Object.name
                                            bpy.data.objects[diff_name].name = item.Object.name
                                            item.Object.name = diff_name
                                            item.Object.data.name = diff_name
                                            
                        # Move Finished
                            if abp.ab_move_finished_bake:
                                scene.autobake_objectqueuelist.move(index, len(scene.autobake_objectqueuelist)-1)
                                last_item = scene.autobake_objectqueuelist[len(scene.autobake_objectqueuelist)-1]
                            else:
                                last_item = item
                                
                        # Store Bake Results
                            results = []
                            for item in scene.autobake_queuelist:
                                results.append({f"{item.Type} " + (f"{item.Multiplier:.2f}" if is_udim_bake else f"{item.Size}"): {'Status': item.Status, 'Icon': item.Icon, 'Error': item.Error,'Cancel': item.Cancel}})
                            bake_results[last_item.Object.name] = results

                            break
                 
                # Next Object
                    if any(item.Enabled for item in scene.autobake_objectqueuelist):
                        self.phase_bake_locked = False
     
                    # Prepare Next Object
                        if abp.ab_auto_next == 'Always' or (abp.ab_auto_next == 'On-Success' and not any(item.Status == 'Failed' for item in scene.autobake_queuelist)):

                        # Restore for Next Object
                            for item in scene.autobake_queuelist:
                                item.Enabled = True
                                if item.Status in ['Baked', 'Exported', 'Failed']:
                                    item.Status = 'Pending'
                                    item.Icon = 'PREVIEW_RANGE'
                                    item.name = f"{item.Type} " + (f"{item.Multiplier:.2f}" if scene.autobake_properties.ab_udim_bake else f"{item.Size}") + f" {item.Status}" + (' Enabled' if item.Enabled else ' Disabled')
                            
                        # Restore Bake Queue Order
                            global bake_order
                            for index, item in enumerate(bake_order):
                                for index_, item_ in enumerate(scene.autobake_queuelist):
                                    if (f"{item_.Type} " + (f"{item_.Multiplier:.2f}" if is_udim_bake else f"{item_.Size}")) == item:
                                        scene.autobake_queuelist.move(index_, index)
                                        break

                        # Next Object Now
                            next_object_bake = True
                            
                    # Next Object on Click
                        else:
                            next_object_bake = False

                # Bake Finish
                    else:
                        
                    # Reports
                        if len(scene.autobake_objectqueuelist):
                            
                        # Object Summary Report
                            if abp.ab_report_object_summary:
                                self.report({'INFO'},
                                    f"Auto Bake: Objects are finished baking. " +
                                    (f"Baked & Exported: {sum(1 for item in scene.autobake_objectqueuelist if item.Status in ['Baked', 'Exported'])}; " if (sum(1 for item in scene.autobake_objectqueuelist if item.Status in ["Baked", 'Exported']) == sum(1 for item in scene.autobake_objectqueuelist if item.Status == "Exported")) else f"Baked: {sum(1 for item in scene.autobake_objectqueuelist if item.Status == 'Baked')}; Exported: {sum(1 for item in scene.autobake_objectqueuelist if item.Status == 'Exported')}; ") +
                                    f"Canceled: {sum(1 for item in scene.autobake_objectqueuelist if item.Status == 'Canceled')}; Mixed: {sum(1 for item in scene.autobake_objectqueuelist if item.Status == 'Mixed')}; Failed: {sum(1 for item in scene.autobake_objectqueuelist if item.Status == 'Failed')}")

                        # Bake Summary
                            obj_baked_report = 0
                            obj_exported_report = 0
                            obj_canceled_report = 0
                            obj_failed_report = 0
                            
                            for obj in scene.autobake_objectqueuelist:
                                if obj.Object.name in bake_results:
                                    for obj_results in bake_results[obj.Object.name]:
                                        for bake in obj_results:
                                            if obj_results[bake]['Status'] == 'Baked':
                                                obj_baked_report += 1
                                            elif obj_results[bake]['Status'] == 'Exported':
                                                obj_baked_report += 1
                                                obj_exported_report += 1
                                            elif obj_results[bake]['Status'] == 'Canceled':
                                                obj_canceled_report += 1
                                            elif obj_results[bake]['Status'] == 'Failed':
                                                obj_failed_report += 1
          
                        # Bake Summary Report
                            if abp.ab_report_bake_summary:
                                self.report({'INFO'},
                                    f"Auto Bake: Bakes are finished for all ({len(scene.autobake_objectqueuelist)}) objects: " + 
                                    (f"Baked & Exported: {obj_baked_report}; " if (obj_baked_report == obj_exported_report) else f"Baked: {obj_baked_report}; Exported: {obj_exported_report}; ") +
                                    f"Canceled: {obj_canceled_report}; Failed: {obj_failed_report}")

                    # Confirm Results
                        if abp.ab_auto_confirm == 'Always' or (abp.ab_auto_confirm == 'On-Success' and not any(item.Status in ['Mixed', 'Failed'] for item in scene.autobake_objectqueuelist)):
                            scene.autobake_queuelist.clear()
                            scene.autobake_objectqueuelist.clear()
                            
                    # Close Session
                        context.window_manager.event_timer_remove(self._timer)
                        bake_status = 'IDLE'
                        
                        return {'FINISHED'}
                    
        return {'PASS_THROUGH'}


#------------------------------------------------------------------------------------
#   Panels

class AUTOBAKE_PT_Main:
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "render"

class AUTOBAKE_PT_Bake(AUTOBAKE_PT_Main, Panel):
    bl_idname = "AUTOBAKE_PT_Bake"
    bl_label = "Auto Bake"

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
     
        col = layout.column_flow(columns=1)
     
        if bake_status == 'IDLE':
            col.alert = (not bool(os.path.exists(abp.ab_filepath)) and abp.ab_texture_export) or (abp.ab_bake_list_item_count < 1 if not abp.ab_udim_bake else (abp.ab_udim_list_item_count < 1 or abp.ab_udimtype_list_item_count < 1))
            col.operator("autobake.start_bake", icon="RENDER_STILL", text="Auto Bake       ")
            
            row = col.row(align=True)
            col1 = row.column()
            col1.enabled = len(scene.autobake_imageexport) > 0 and bool(os.path.exists(abp.ab_filepath))
            col1.alert = not bool(os.path.exists(abp.ab_filepath))
            col1.operator("autobake.export_textures", icon="EXPORT", text="   Export")
            
            row.separator()
            col2 = row.column()
            col2.alert = not bool(os.path.exists(abp.ab_filepath)) and abp.ab_texture_export 
            col2.prop(scene.autobake_properties, "ab_texture_export", text="", toggle=False, emboss=True)
            
        elif bake_status in ['INPROGRESS', 'PAUSED']:
            col.enabled = any(item.Enabled for item in scene.autobake_queuelist) or any(item.Enabled for item in scene.autobake_objectqueuelist)
            
            if bake_status == 'PAUSED':
                if next_object_bake:
                    col.operator("autobake.resume_bake", icon="PLAY", text="Resume        ")
                else:
                    col.operator("autobake.next_object", icon="PLAY", text="Resume & Next Object        ")
            else:
                row_pause = col.row()
                row_pause.enabled = next_object_bake
                row_pause.operator("autobake.pause_bake", icon="PAUSE", text="Pause        ")

            col.operator("autobake.cancel_bake", icon="CANCEL", text="Cancel        ")
            
            row = layout.row(align=True)
            sub = row.row()
            sub.label(text="Bake Queue", icon="TEMP")
            sub = row.row(align=True)
            sub.alignment = "RIGHT"
            sub.label(text="Textures: "+str(finished_bake_count)+'/'+str(len(scene.autobake_queuelist)), icon="TEXTURE")
            
        if len(scene.autobake_queuelist) > 0:
            col = layout.column_flow(columns=1)
            col.template_list("AUTOBAKE_UL_BakeQueue", "Bake_Queue_List", scene, "autobake_queuelist", scene, "autobake_queuelist_index", rows=1)

            if len(scene.autobake_objectqueuelist) > 1:
                col.template_list("AUTOBAKE_UL_ObjectQueue", "Object_Queue_List", scene, "autobake_objectqueuelist", scene, "autobake_objectqueuelist_index", rows=1)
    
            if bake_status == 'IDLE' or abp.ab_auto_next == 'Always' or (abp.ab_auto_next == 'On-Success' and next_object_bake):
                confirm_row = col.row()
                confirm_row.enabled = bake_status == 'IDLE'
                confirm_row.operator('autobake.confirm_bakes', text='Confirm        ', icon='CHECKMARK')
            else:
                confirm_row = col.row()
                confirm_row.enabled = not next_object_bake
                confirm_row.operator('autobake.next_object', text=' Next Object        ', icon='FRAME_NEXT')
                

class AUTOBAKE_PT_Lists(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Bake"
    bl_idname = "AUTOBAKE_PT_Lists"
    bl_label = ""

    def draw_header(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
        
        if abp.ab_udim_bake:
            layout.alert = abp.ab_udimtype_list_item_count < 1
            layout.label(text="Bake List" +(" (" + str(abp.ab_udimtype_list_item_count) +")" if abp.ab_udimtype_list_item_count > 0 else ""), icon="OPTIONS")
        else:
            layout.alert = abp.ab_bake_list_item_count < 1
            layout.label(text="Bake List" +(" (" + str(abp.ab_bake_list_item_count) +")" if abp.ab_bake_list_item_count > 0 else ""), icon="OPTIONS")

    def draw(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        list =  scene.autobake_udimlist if abp.ab_udim_bake else scene.autobake_bakelist
    
    # Layout
        layout = self.layout
        layout = layout.column_flow(columns=1)
        row = layout.row(align=True)

    # Lists
        if abp.ab_udim_bake:
            if abp.ab_list_padding in ['Compact', 'Moderate']:
                row.template_list("AUTOBAKE_UL_UDIMType", "UDIM_Type_List", scene, "autobake_udimlist", scene, "autobake_udimlist_index", rows=4)
            else:
                row.template_list("AUTOBAKE_UL_UDIMType", "UDIM_Type_List", scene, "autobake_udimlist", scene, "autobake_udimlist_index", rows=5)
        else:
            if abp.ab_list_padding in ['Compact', 'Moderate']:
                row.template_list("AUTOBAKE_UL_Bake", "Bake_List", scene, "autobake_bakelist", scene, "autobake_bakelist_index", rows=4)
            else:
                row.template_list("AUTOBAKE_UL_Bake", "Bake_List", scene, "autobake_bakelist", scene, "autobake_bakelist_index", rows=5)   
        
    # Buttons
        if abp.ab_list_padding == 'Compact':
            row = row.row()
            col = row.column(align=True)
            col.operator("autobake.bakelist_add", icon='ADD', text="")
            if list:
                col.operator("autobake.bakelist_remove", icon='REMOVE', text="")
                if len(list) > 1:
                    col_a = col.column(align=True)
                    col_a.operator("autobake.bakelist_move_up", icon='TRIA_UP', text="")
                    col_a.operator("autobake.bakelist_move_down", icon='TRIA_DOWN', text="")
            col.menu(AUTOBAKE_MT_BakeList.bl_idname, text="", icon="DOWNARROW_HLT")
            
        elif abp.ab_list_padding == 'Moderate':
            row.separator()
            col = row.column(align=False)
            col_c = col.column(align=True)
            col_c.operator("autobake.bakelist_add", icon='ADD', text="")
            if list:
                col_c.operator("autobake.bakelist_remove", icon='REMOVE', text="")
                if len(list) > 1:
                    col_a = col.column(align=True)
                    col_a.operator("autobake.bakelist_move_up", icon='TRIA_UP', text="")
                    col_a.operator("autobake.bakelist_move_down", icon='TRIA_DOWN', text="")
            col.menu(AUTOBAKE_MT_BakeList.bl_idname, text="", icon="DOWNARROW_HLT")
            
        elif abp.ab_list_padding == 'Spacious':
            row.separator()
            col = row.column(align=True)
            col.operator("autobake.bakelist_add", icon='ADD', text="")
            if list:
                col.operator("autobake.bakelist_remove", icon='REMOVE', text="")
                if len(list) > 1:
                    col.separator()
                    col.operator("autobake.bakelist_move_up", icon='TRIA_UP', text="")
                    col.operator("autobake.bakelist_move_down", icon='TRIA_DOWN', text="")
            col.separator()
            col.menu(AUTOBAKE_MT_BakeList.bl_idname, text="", icon="DOWNARROW_HLT")
            col.separator()
            col.separator()

    # Type & Size
        if abp.ab_udim_bake and list and abp.ab_udim_item_default:
            item = list[scene.autobake_udimlist_index]
            
            if abp.ab_list_padding in ['Moderate', 'Compact']:
                row = layout.row(align=True)
                
            elif abp.ab_list_padding == 'Spacious':
                row = layout.row(align=False)
                row = row.row(align=False)

            row.scale_x = 1.3
            row.alert = item.IsDuplicate
            row.prop(item, "Type")
            row = row.row()
            row.prop(item, "Size")
    
        elif not abp.ab_udim_bake and list and abp.ab_item_default:
            item = scene.autobake_bakelist[scene.autobake_bakelist_index]
            
            if abp.ab_list_padding in ['Moderate', 'Compact']:
                row = layout.row(align=True)
                
            elif abp.ab_list_padding == 'Spacious':
                row = layout.row(align=False)

            row.alert = item.IsDuplicate
            col1 = row.column()
            col1.scale_x = 1.15
            col1.prop(item, "Type")
            row = row.row(align=True)
            col2 = row.column(align=True)
            col2.scale_x = 1.05
            col2.operator("autobake.scale_down", text='', icon="TRIA_LEFT")
            row.prop(item, "Size")
            col4 = row.column(align=True)
            col4.scale_x = 1.05
            col4.operator("autobake.scale_up", text="", icon="TRIA_RIGHT")
        
    # Alerts
        if abp.ab_alert_texts:
            if not len(list):
                row = layout.row()
                row.alert = True
                row.label(text="List can't be empty!")
            
            elif (abp.ab_udimtype_list_item_count if abp.ab_udim_bake else abp.ab_bake_list_item_count) < 1:
                row = layout.row()
                row.alert = True
                row.label(text="List must have an enabled item!")
            
    # Item Specific
        if abp.ab_item_details and list:
            rd = context.scene.render
            image_settings = rd.image_settings
            Type = list[scene.autobake_udimlist_index].Type if abp.ab_udim_bake else list[scene.autobake_bakelist_index].Type

            col = layout.column_flow(columns=1, align=True)
            box = col.box()
            
            row = box.grid_flow(columns = 1)
            split = row.split(factor=.4)
            split.label(text="Name")
            split.alert = len(abp.ab_baketype_name) < 1
            split.prop(abp, "ab_baketype_name", text='')

            cbk = context.scene.render.bake
            
            if Type == 'Combined':
                box = col.box()
                box.alert = not (cbk.use_pass_emit or ((cbk.use_pass_indirect or cbk.use_pass_direct) and (cbk.use_pass_diffuse or cbk.use_pass_glossy or cbk.use_pass_transmission)))
            
                split = box.split(factor=.4)
                
                col = split.column(align=True, heading="")
                col.label(text='Influence')
                col.prop(cbk, "use_pass_direct")
                col.prop(cbk, "use_pass_indirect")
                col.prop(cbk, "use_pass_emit")
                
                col = split.column(align=True, heading="")
                col.active = cbk.use_pass_direct or cbk.use_pass_indirect
                col.label(text='')
                col.prop(cbk, "use_pass_diffuse")
                col.prop(cbk, "use_pass_glossy")
                col.prop(cbk, "use_pass_transmission")
            
                if box.alert and abp.ab_alert_texts:
                    box.label(text="Must use Emit, or a light pass with Direct or Indirect influence!")

            if Type in ['Diffuse', 'Glossy', 'Transmission']:
                box = col.box()
                box.alert = not (cbk.use_pass_direct or cbk.use_pass_indirect or cbk.use_pass_color)
            
                split = box.split(factor=.4)
                
                split.label(text='Influence')
                col = split.column(align=True)
                col.prop(cbk, "use_pass_direct")
                col.prop(cbk, "use_pass_indirect")
                col.prop(cbk, "use_pass_color")
                
                if box.alert and abp.ab_alert_texts:
                    box.label(text="Must enable at least one influence setting!")
                
            if Type in ['Normal', 'Normal ', 'Tangent', 'Coat Normal', 'Clearcoat Normal', 'Subsurface Radius']:
                box = col.box()
                col = box.column()
                
                split = col.split(factor=.4)
                split.label(text="Space")
                split.prop(cbk, "normal_space", text='')
                
                split = col.split(factor=.4)
                split.label(text="Red")
                split.prop(cbk, "normal_r", text='')
                
                split = col.split(factor=.4)
                split.label(text="Green")
                split.prop(cbk, "normal_g", text='')
                
                split = col.split(factor=.4)
                split.label(text="Blue")
                split.prop(cbk, "normal_b", text='')

            if Type == 'Channel Packing': 
                box = col.box()
                box.alert = abp.ab_channel_pack_r == "None" and abp.ab_channel_pack_g == "None" and abp.ab_channel_pack_b == "None"
                
                col = box.column()
                
                split = col.split(factor=.4)
                split.label(text="Red")
                split.prop(abp, "ab_channel_pack_r", text='')
                
                split = col.split(factor=.4)
                split.label(text="Green")
                split.prop(abp, "ab_channel_pack_g", text='')
                
                split = col.split(factor=.4)
                split.label(text="Blue")
                split.prop(abp, "ab_channel_pack_b", text='')
                
                if box.alert and abp.ab_alert_texts:
                    box.label(text="Can't all be set to 'None'!")
                
            if Type == 'Displacement':
                box = col.box()
                col = box.column()

                split = col.split(factor=.4)
                split.label(text="Viewport Level")
                split.prop(abp, "ab_multires_level", text='')
                
                split = col.split(factor=.4)
                split.label(text="Mesh Resolution")
                rd = scene.render
                split.prop(rd, "use_bake_lores_mesh", invert_checkbox=True, text='High      ', toggle=True, icon='CHECKBOX_DEHLT' if rd.use_bake_lores_mesh else 'CHECKBOX_HLT')

            if Type == 'Normals':
                box = col.box()
                col = box.column()

                split = col.split(factor=.4)
                split.label(text="Viewport Level")
                split.prop(abp, "ab_multires_level", text='')
                
            if Type == 'Ambient Occlusion ':
                box = col.box()
                col = box.column()
                
                split = col.split(factor=.4)
                split.label(text="Local Only")
                split.alert = scene.render.bake.use_selected_to_active and abp.ab_ao_local_only == 'COLLECTION'
                split.prop(abp, "ab_ao_local_only", icon='RESTRICT_RENDER_OFF', text='')

                split = col.split(factor=.4)
                split.label(text="Normal")
                split.prop(abp, "ab_ao_use_normal", toggle=True, text="Use")

            if Type == 'Ambient Occlusion':
                box = col.box()
                col = box.column()

                split = col.split(factor=.4)
                split.label(text="Sample")
                row = split.row(align=True)
                row.prop(abp, "ab_ao_sample_use", text='')
                row = row.row()
                row.active = abp.ab_ao_sample_use
                row.prop(abp, "ab_ao_sample", text='')
                
                split = col.split(factor=.4)
                split.label(text="Inside")
                split.prop(abp, "ab_ao_inside", text='Enabled      ', toggle=True, icon='CHECKBOX_HLT' if abp.ab_ao_inside else 'CHECKBOX_DEHLT')
                
                split = col.split(factor=.4)
                split.label(text="Only Local")
                split.prop(abp, "ab_ao_only_local", text='Enabled      ', toggle=True, icon='CHECKBOX_HLT' if abp.ab_ao_only_local else 'CHECKBOX_DEHLT')
                
                split = col.split(factor=.4)
                split.label(text="Distance")
                split.prop(abp, "ab_ao_distance", text='')
                
            if Type == 'Pointiness':
                box = col.box()
                col = box.column()

                split = col.split(factor=.4)
                split.label(text="Contrast")
                split.prop(abp, "ab_pointiness_contrast", text='')
                
                split = col.split(factor=.4)
                split.label(text="Brightness")
                split.prop(abp, "ab_pointiness_brightness", text='')
                
            if Type == 'Displacement ':
                box = col.box()
                col = box.column(align=True)
                
                split = col.split(factor=.4)
                split.label(text="Source Only")
                split.prop(abp, "ab_displacement_source", text='Enabled      ', toggle=True, icon='CHECKBOX_HLT' if abp.ab_displacement_source else 'CHECKBOX_DEHLT')
                
            if Type == 'UV':
                box = col.box()
                col = box.column()
                
                split = col.split(factor=.4)
                split.label(text="UV Map")
                split.prop(abp, "ab_uv_target", text='')
                
            if Type == 'Color Attribute':
                box = col.box()
                col = box.column()
                
                split = col.split(factor=.4)
                split.label(text="Color Attribute")
                split.prop(abp, "ab_attribute_target", text='')
                

class AUTOBAKE_PT_List_UDIM(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Lists"
    bl_label = ""

    @classmethod
    def poll(cls, context):
        return context.scene.autobake_properties.ab_udim_bake == True
    
    def draw_header(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
        layout.alert = abp.ab_udim_list_item_count < 1
        layout.label(text="UDIM Tiles" +(" (" + str(abp.ab_udim_list_item_count) +")" if abp.ab_udim_list_item_count > 0 else ""), icon="UV_FACESEL")
        
    def draw(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        list = scene.autobake_udimtilelist
        
        layout = self.layout
        layout = layout.column_flow(columns=1)

        row = layout.row(align=True)
        if abp.ab_list_padding in ['Compact', 'Moderate']:
            row.template_list("AUTOBAKE_UL_UDIMTile", "UDIM_List", scene, "autobake_udimtilelist", scene, "autobake_udimtilelist_index", rows=4)
        else:
            row.template_list("AUTOBAKE_UL_UDIMTile", "UDIM_List", scene, "autobake_udimtilelist", scene, "autobake_udimtilelist_index", rows=5)

    # Buttons
        if abp.ab_list_padding == 'Compact':
            row = row.row()
            col = row.column(align=True)
            col.operator("autobake.udimlist_add", icon='ADD', text="")
            if list:
                col.operator("autobake.udimlist_remove", icon='REMOVE', text="")
                if len(list) > 1:
                    col_a = col.column(align=True)
                    col_a.operator("autobake.udimlist_up", icon='TRIA_UP', text="")
                    col_a.operator("autobake.udimlist_down", icon='TRIA_DOWN', text="")
            col.menu(AUTOBAKE_MT_UDIMList.bl_idname, text="", icon="DOWNARROW_HLT")
            
        elif abp.ab_list_padding == 'Moderate':
            row.separator()
            col = row.column(align=False)
            col_c = col.column(align=True)
            col_c.operator("autobake.udimlist_add", icon='ADD', text="")
            if list:
                col_c.operator("autobake.udimlist_remove", icon='REMOVE', text="")
                if len(list) > 1:
                    col_a = col.column(align=True)
                    col_a.operator("autobake.udimlist_up", icon='TRIA_UP', text="")
                    col_a.operator("autobake.udimlist_down", icon='TRIA_DOWN', text="")
            col.menu(AUTOBAKE_MT_UDIMList.bl_idname, text="", icon="DOWNARROW_HLT")
            
        elif abp.ab_list_padding == 'Spacious':
            row.separator()
            col = row.column(align=True)
            col.operator("autobake.udimlist_add", icon='ADD', text="")
            if list:
                col.operator("autobake.udimlist_remove", icon='REMOVE', text="")
                if len(list) > 1:
                    col.separator()
                    col.operator("autobake.udimlist_up", icon='TRIA_UP', text="")
                    col.operator("autobake.udimlist_down", icon='TRIA_DOWN', text="")
            col.separator()
            col.menu(AUTOBAKE_MT_UDIMList.bl_idname, text="", icon="DOWNARROW_HLT")
            col.separator()
            col.separator()
        
        
        if len(list) > 0 and abp.ab_udim_item_tilescale_default:
            item = scene.autobake_udimtilelist[scene.autobake_udimtilelist_index]
            
            if abp.ab_list_padding in ['Moderate', 'Compact']:
                row = layout.row(align=True)
                
            elif abp.ab_list_padding == 'Spacious':
                row = layout.row(align=False)
            
            col1 = row.column()
            col1.alert = item.IsDuplicate
            col1.scale_x = 1.15
            col1.prop(item, "UDIM")
            row = row.row(align=True)
            col2 = row.column(align=True)
            col2.scale_x = 1.05
            col2.operator("autobake.scale_down", text="", icon="TRIA_LEFT")
            row.prop(item, "Size")
            col4 = row.column(align=True)
            col4.scale_x = 1.05
            col4.operator("autobake.scale_up", text="", icon="TRIA_RIGHT")
        
        if abp.ab_alert_texts:
            if not len(list):
                row = layout.row()
                row.alert = True
                row.label(text="List can't be empty!")
            
            elif abp.ab_udim_list_item_count < 1:
                row = layout.row()
                row.alert = True
                row.label(text="List must have an enabled item!")


class AUTOBAKE_PT_Image(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Bake"
    bl_idname = "AUTOBAKE_PT_Image"
    bl_label = ""

    def draw_header(self, context):
        self.layout.label(text="Image", icon="IMAGE")

    def draw(self, context):
        pass


class AUTOBAKE_PT_Name(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Image"
    bl_label = ""
    
    def draw_header(self, context):
        layout = self.layout
        layout.label(text='Name')
        row = layout.row()
        row.active=False
        row.operator('autobake.edit_name_structure', text='', icon='GREASEPENCIL', emboss=False)
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
      
        col = layout.column()
      
        split = col.split(factor=.4)
        split.label(text="Prefix")
        split.prop(abp, "ab_prefix", text="")
        
        split = col.split(factor=.4)
        split.label(text="Bridge")
        split.prop(abp, "ab_bridge", text="")
        
        split = col.split(factor=.4)
        split.label(text="Suffix")
        split.prop(abp, "ab_suffix", text="")


class AUTOBAKE_PT_Format(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Image"
    bl_label = "Format"
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
        image_settings = scene.render.image_settings
        file_format = image_settings.file_format

        col = layout.column()

        split = col.split(factor=.4)
        split.label(text="File Format")
        split.prop(abp, "ab_fileformat", text="", icon="IMAGE_DATA")
        
        split = col.split(factor=.4)
        split.label(text="Use Float")
        split.prop(abp, "ab_floatbuffer", text=str(abp.ab_floatbuffer), toggle=True, expand=True)           
            
        if abp.ab_fileformat in ["PNG", "OPEN_EXR_MULTILAYER", "OPEN_EXR", "TIFF", "JPEG2000", "DPX"]:
            split = col.split(factor=.4)
            split.label(text="Color Depth")
            split.row().prop(image_settings, "color_depth", text=' ', expand=True)
        
        if abp.ab_fileformat == "PNG":
            split = col.split(factor=.4)
            split.label(text="Compression")
            split.prop(image_settings, "compression", text="")

        if abp.ab_fileformat in ["JPEG", "JPEG2000", "WEBP"]:
            split = col.split(factor=.4)
            split.label(text="Quality")
            split.prop(image_settings, "quality", text="")
            
        if abp.ab_fileformat == "JPEG2000":
            split = col.split(factor=.4)
            split.label(text="Codec")
            split.prop(image_settings, "jpeg2k_codec", text="")

            split = col.split(factor=.4)
            split.label(text="Cinema")
            split.prop(image_settings, "use_jpeg2k_cinema_preset", text=str(image_settings.use_jpeg2k_cinema_preset), toggle=True, expand=True)
            
            split = col.split(factor=.4)
            split.label(text="Cinema (48)")
            split.prop(image_settings, "use_jpeg2k_cinema_48", text=str(image_settings.use_jpeg2k_cinema_48), toggle=True, expand=True)
            
            split = col.split(factor=.4)
            split.label(text="YCC")
            split.prop(image_settings, "use_jpeg2k_ycc", text=str(image_settings.use_jpeg2k_ycc), toggle=True, expand=True)
            
        if abp.ab_fileformat == "DPX":
            split = col.split(factor=.4)
            split.label(text="Log")
            split.prop(image_settings, "use_cineon_log", text=str(image_settings.use_cineon_log), toggle=True, expand=True)
        
        if abp.ab_fileformat in ["OPEN_EXR_MULTILAYER", "OPEN_EXR"]:
            split = col.split(factor=.4)
            split.label(text="Codec")
            split.prop(image_settings, "exr_codec", text="")

        if abp.ab_fileformat == "TIFF":
            split = col.split(factor=.4)
            split.label(text="Compression")
            split.prop(image_settings, "tiff_codec", text="")
            
            
class AUTOBAKE_PT_Export(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Image"
    bl_idname = "AUTOBAKE_PT_Export"
    bl_label = "Export"

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
        
        row = layout.row()
        row.label(text="File Path", icon="EXPORT")
        sub = row.row(align=True)
        sub.alignment = "RIGHT"
        sub.prop(abp, "ab_subfolders", text="Subfolders", toggle=False)

        column = layout.column(align=True)
        row = column.row(align=True)
        col = row.column(align=True)
        col.alert = not bool(os.path.exists(abp.ab_filepath))
        col.prop(abp, "ab_filepath", text="")
        row.operator("autobake.file_explorer", text="", icon='FILEBROWSER')
        
        if abp.ab_alert_texts and col.alert:
            row = column.row()
            row.alert = True
            row.label(text="Filepath is Invalid!")
        
        
class AUTOBAKE_PT_ColorOverride(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Export"
    bl_idname = 'AUTOBAKE_PT_ColorOverride'
    bl_label = ""
    bl_options = {'DEFAULT_CLOSED'}
    
    def draw_header(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        self.layout.prop(abp, "ab_custom_color_management", text="Color Override")
        
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        image_settings = scene.render.image_settings
        abp = scene.autobake_properties
    
        layout.enabled = abp.ab_custom_color_management
 
        layout.use_property_split = True
        layout.use_property_decorate = False
        
        layout.row().prop(image_settings, "color_management", text=" ", expand=True)
        
        flow = layout.grid_flow(row_major=True, columns=1, even_columns=False, even_rows=False, align=True)

        if image_settings.color_management == 'OVERRIDE':
            owner = image_settings
        else:
            owner = scene
            flow.enabled = False

        col = flow.column()

        if image_settings.has_linear_colorspace:
            if hasattr(owner, "linear_colorspace_settings"):
                col.prop(owner.linear_colorspace_settings, "name", text="Color Space")
        else:
            col.prop(owner.display_settings, "display_device")
            col.separator()         
            col.template_colormanaged_view_settings(owner, "view_settings")


class AUTOBAKE_PT_Settings(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = 'AUTOBAKE_PT_Bake'
    bl_idname = "AUTOBAKE_PT_Settings"
    bl_label = ""

    def draw_header(self, context):
        self.layout.label(text="Settings", icon="TOOL_SETTINGS")
        
    def draw(self, context):
        pass

        
class AUTOBAKE_PT_Margin(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Settings"
    bl_label = "Margin"
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
        cbk = scene.render.bake

        col = layout.column()
        
        split = col.split(factor=.4)
        split.label(text="Type")
        split.prop(cbk, "margin_type", text="")

        split = col.split(factor=.4)
        split.label(text="Size")
        split.prop(cbk, "margin", text="")
        
        split = col.split(factor=.4)
        split.label(text="")
        split.prop(abp, "ab_adaptive_margin", text="Auto Adjust")
        
        
class AUTOBAKE_PT_SelectedToActive(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Settings"
    bl_idname = 'AUTOBAKE_PT_SelectedToActive'
    bl_label = ""
    
    def draw_header(self, context):
        cbk = context.scene.render.bake
        self.layout.prop(cbk, "use_selected_to_active", text="Selected to Active")
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        cbk = scene.render.bake
        abp = scene.autobake_properties

        layout.active = cbk.use_selected_to_active

        col = layout.column()

        split = col.split(factor=.4)
        split.label(text="Cage")
        split.prop(cbk, "use_cage", text='Use       ', toggle=True,  icon='CHECKBOX_HLT' if cbk.use_cage else 'CHECKBOX_DEHLT')

        if cbk.use_cage:
            split = col.split(factor=.4)
            split.label(text="Cage Object")
            split.prop(cbk, "cage_object", text="")

        split = col.split(factor=.4)
        split.label(text="Cage Extrusion" if cbk.use_cage else "Extrusion")
        split.prop(cbk, "cage_extrusion", text="")
        
        split = col.split(factor=.4)
        split.label(text="Max Ray Distance")
        split.prop(cbk, "max_ray_distance", text="")
        
        split = col.split(factor=.4)
        split.label(text="")
        split.prop(abp, "ab_active_as_final")
        

class AUTOBAKE_PT_SelectionHelp(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_SelectedToActive"
    bl_label = "Selection Helper"
    bl_options = {'DEFAULT_CLOSED'}
    
    target_error = StringProperty(options = set(), default='None')
    source_error = StringProperty(options = set(), default='None')
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties

        self.target_error = "None"
        self.source_error = "None"
        if layout.enabled:
            if abp.ab_target_object == None:
                self.target_error = "Select an object!"
            elif abp.ab_target_object.name not in bpy.context.scene.objects:
                self.target_error = "Object is not from this scene!"
            elif abp.ab_target_object.type != 'MESH':
                self.target_error = "Object is not a 'Mesh'"

            if abp.ab_source_object_method == 'COLLECTION':
                if abp.ab_source_collection == None:
                    self.source_error = "Select a collection!"
                elif abp.ab_source_collection not in scene.collection.children_recursive:
                    self.source_error = "Collection is not from this scene!"
                elif not len(abp.ab_source_collection.objects):
                    self.source_error = "Collection is empty!"
                    
            elif len(scene.autobake_sourceobject) < 2:
                self.source_error = "List can't be empty!"
                
        layout.active = scene.render.bake.use_selected_to_active

        row = layout.row()
        row.enabled = self.target_error == "None" and self.source_error == "None" and layout.active
        row.operator('autobake.select_from_list', text='Select Objects      ', icon='SELECT_SET')
        
        split = layout.split(factor=.4)
        split.label(text='Target Object')
        col = split.column(align=True)
        col.alert = layout.active and self.target_error != "None"
        col.prop(abp, 'ab_target_object', text='')

        if layout.active and abp.ab_alert_texts and self.target_error != "None":
            col.label(text=self.target_error)
                
        split = layout.split(factor=.4)
        split.label(text='Source Objects')
        
        if abp.ab_source_object_method == 'COLLECTION':
            col = split.column()
            col.alert = layout.active and self.source_error != "None"
            col.row().prop(abp, "ab_source_object_method", text=' ', expand=True)
            col.prop(abp, "ab_source_collection", text='')

            if layout.active and abp.ab_alert_texts and self.source_error != "None":
                col.label(text=self.source_error)
        else:
            split.row().prop(abp, "ab_source_object_method", text=' ', expand=True)

            col = layout.column(align=True)
            col.alert = layout.active and self.source_error != "None"
            col.template_list("AUTOBAKE_UL_SourceObjects", "Source_Objects", scene, "autobake_sourceobject", scene, "autobake_sourceobject_index", rows=2)
            col.operator('autobake.load_selected_objects', icon='IMPORT', text='Load from Selected      ')
            
            if layout.active and abp.ab_alert_texts and self.source_error != "None":
                row = layout.row()
                row.alert = True
                row.label(text=self.source_error)


class AUTOBAKE_PT_Sampling(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Settings"
    bl_idname = "AUTOBAKE_PT_Sampling"
    bl_label = "Sampling"
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties        

        col = layout.column()
        
        split = col.split(factor=.4)
        split.alignment = 'RIGHT'
        split.label(text="Render Sampling")
        split.prop(abp, "ab_sampling_use_render", toggle=True, text='Use       ', icon='CHECKBOX_HLT' if abp.ab_sampling_use_render else 'CHECKBOX_DEHLT')
        
        if not abp.ab_sampling_use_render:
            split = col.split(factor=.4)
            split.active = not abp.ab_sampling_use_render
            split.label(text="")
            split.prop(abp, "ab_auto_pick_sampling")
        
        else:
            cscene = scene.cycles
            
            split = col.split(factor=.4)
            split.label(text="")
            split.prop(cscene, "use_adaptive_sampling", text='Adaptive Sampling')
            
            if cscene.use_adaptive_sampling:
                split = col.split(factor=.4)
                split.alignment = 'RIGHT'
                split.label(text="Noise Threshold")
                split.prop(cscene, "adaptive_threshold", text="")
            
            split = col.split(factor=.4)
            split.alignment = 'RIGHT'
            split.label(text="Max Samples" if cscene.use_adaptive_sampling else "Samples")
            split.prop(cscene, "samples", text="")
            
            if cscene.use_adaptive_sampling:
                split = col.split(factor=.4)
                split.alignment = 'RIGHT'
                split.label(text="Min Samples")
                split.prop(cscene, "adaptive_min_samples", text="")
                
            split = col.split(factor=.4)
            split.alignment = 'RIGHT'
            split.label(text="Time Limit")
            split.prop(cscene, "time_limit", text="")
        

class AUTOBAKE_PT_Denoise(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Sampling"
    bl_label = ""
    bl_options = {'DEFAULT_CLOSED'}
        
    @classmethod
    def poll(cls, context):
        return context.scene.autobake_properties.ab_sampling_use_render
    
    def draw_header(self, context):
        scene = context.scene
        self.layout.prop(scene.cycles, "use_denoising", text="Denoise")                     
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        cscene = scene.cycles
        
        layout.active = cscene.use_denoising
        
        split = layout.split(factor=.4)
        split.label(text="Denoiser")
        split.prop(cscene, "denoiser", text="")

        split = layout.split(factor=.4)
        split.label(text="Passes")
        split.prop(cscene, "denoising_input_passes", text="") 
        
        if cscene.denoiser == 'OPENIMAGEDENOISE':
            row = layout.row()
            split = row.split(factor=.4)
            split.label(text="Prefilter")
            split.prop(cscene, "denoising_prefilter", text="")


class AUTOBAKE_PT_Sampling_Low(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Sampling"
    bl_idname = "AUTOBAKE_PT_Sampling_Low"
    bl_label = "Low"
        
    @classmethod
    def poll(cls, context):
        return context.scene.autobake_properties.ab_sampling_use_render == False
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
    
        layout.use_property_split = True
        layout.use_property_decorate = False
    
        col = layout.column()
        col.prop(abp, 'ab_sampling_low_adaptive')
        
        if abp.ab_sampling_low_adaptive:
            col.prop(abp, 'ab_sampling_low_noise_threshold')
            col.prop(abp, 'ab_sampling_low_max')
            col.prop(abp, 'ab_sampling_low_min')
        else:
            col.prop(abp, 'ab_sampling_low_max', text='Samples')
        
        col.prop(abp, 'ab_sampling_low_time_limit')
        

class AUTOBAKE_PT_Denoise_Low(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Sampling_Low"
    bl_label = ""
    bl_options = {'DEFAULT_CLOSED'}
    
    def draw_header(self, context):
        scene = context.scene
        self.layout.prop(context.scene.autobake_properties, "ab_sampling_low_denoise")                     
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
        
        layout.use_property_split = True
        layout.use_property_decorate = False
        layout.active = abp.ab_sampling_low_denoise
        
        col = layout.column()
        col.prop(abp, 'ab_sampling_low_denoiser')
        col.prop(abp, 'ab_sampling_low_passes')
        col = col.column()
        col.active = abp.ab_sampling_low_denoiser == 'OPENIMAGEDENOISE'
        col.prop(abp, 'ab_sampling_low_prefilter')


class AUTOBAKE_PT_Sampling_High(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Sampling"
    bl_idname = "AUTOBAKE_PT_Sampling_High"
    bl_label = "High" 
        
    @classmethod
    def poll(cls, context):
        return context.scene.autobake_properties.ab_sampling_use_render == False
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
    
        layout.use_property_split = True
        layout.use_property_decorate = False
    
        col = layout.column()
        col.prop(abp, 'ab_sampling_high_adaptive')
        
        if abp.ab_sampling_high_adaptive:
            col.prop(abp, 'ab_sampling_high_noise_threshold')
            col.prop(abp, 'ab_sampling_high_max')
            col.prop(abp, 'ab_sampling_high_min')
        else:
            col.prop(abp, 'ab_sampling_high_max', text='Samples')
        
        col.prop(abp, 'ab_sampling_high_time_limit')


class AUTOBAKE_PT_Denoise_High(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Sampling_High"
    bl_label = ""
    bl_options = {'DEFAULT_CLOSED'}
    
    def draw_header(self, context):
        scene = context.scene
        self.layout.prop(context.scene.autobake_properties, "ab_sampling_high_denoise")                     
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
        
        layout.use_property_split = True
        layout.use_property_decorate = False
        layout.active = abp.ab_sampling_high_denoise
        
        col = layout.column()
        col.prop(abp, 'ab_sampling_high_denoiser')
        col.prop(abp, 'ab_sampling_high_passes')
        col = col.column()
        col.active = abp.ab_sampling_high_denoiser == 'OPENIMAGEDENOISE'
        col.prop(abp, 'ab_sampling_high_prefilter')
            
                
class AUTOBAKE_PT_Addon(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_Bake"
    bl_idname = "AUTOBAKE_PT_addon_settings"
    bl_label = ""

    def draw_header(self, context):
        layout = self.layout
        scene = context.scene
        layout.label(text='Addon', icon='PLUGIN')
        
    def draw(self, context):
        pass   


class AUTOBAKE_PT_ImageSettings(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_addon_settings"
    bl_label = "Textures"
        
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
        
        col = layout.column()

        split = col.split(factor=.4)
        split.label(text="Pack Images")
        split.prop(abp, "ab_pack_texture", text='Pack      ', toggle=True, icon='PACKAGE' if abp.ab_pack_texture else 'UGLYPACKAGE')
        
        split = col.split(factor=.4)
        split.label(text="Fake User")
        split.prop(abp, "ab_textures_fakeuser", text='Use      ', toggle=True, icon='FAKE_USER_ON' if abp.ab_textures_fakeuser else 'FAKE_USER_OFF')

        split = col.split(factor=.4)
        split.label(text='Anti-aliasing')
        
        col2 = split.column(align=True, heading="")
        col2.prop(abp, "ab_scaled_antialiasing", text='')
        
        if abp.ab_scaled_antialiasing == 'UPSCALED':
            col2.prop(abp, "ab_antialiasing_upscaled", text='Upscale')
            
        elif abp.ab_scaled_antialiasing == 'DOWNSCALED':
            col2.prop(abp, "ab_antialiasing_downscaled", text='Downscale', slider=True)
            col2.prop(abp, "ab_antialiasing_repeat", text='Iteration ')
        
        split = col.split(factor=.4)
        split.label(text='Color Space')
        split.menu(AUTOBAKE_MT_ColorSpace.bl_idname, text='Texture Type', icon='COLOR')
        
        split = col.split(factor=.4)
        split.label(text="Shared Textures")
        split.prop(abp, "ab_shared_textures", text='Enabled      ', toggle=True,  icon='CHECKBOX_HLT' if abp.ab_shared_textures else 'CHECKBOX_DEHLT')
    
        
class AUTOBAKE_PT_Materials(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_addon_settings"
    bl_label = "Materials"
        
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties

        col = layout.column()
        
        split = col.split(factor=.4)
        split.label(text="Image Textures")
        split.prop(abp, "ab_remove_imagetextures", text=str('Remove      '), toggle=True, icon='TRASH')
        
        col = layout.column()
        
        split = col.split(factor=.4)
        split.label(text='Final Material')
        split.prop(abp, "ab_final_material", text='Create      ', icon='MATERIAL')
        
        split = col.split(factor=.4)
        split.active = abp.ab_final_material
        split.label(text='Material Shader')
        split.prop(abp, "ab_final_shader", text='', icon='MATSHADERBALL')
        
        split = col.split(factor=.4)
        split.active = abp.ab_final_material
        split.label(text='Apply Textures')
        split.prop(abp, "ab_apply_textures", text='', icon='NODE_TEXTURE')
        
        col = layout.column()
        col.active = not abp.ab_remove_imagetextures or abp.ab_final_material
        
        split = col.split(factor=.4)
        split.label(text="Node Tiling")
        split.prop(abp, "ab_node_tiling", text='', toggle=True, icon='NODE_CORNER')
        
        split = col.split(factor=.4)
        split.label(text="Node Labels")
        split.prop(abp, "ab_node_label", text='', toggle=True, icon='NODE')
        
        
class AUTOBAKE_PT_Objects(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_addon_settings"
    bl_idname = "AUTOBAKE_PT_objects"
    bl_label = "Objects"
        
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties

        layout.active = not scene.render.bake.use_selected_to_active or not abp.ab_active_as_final
        
        col = layout.column()

        split = col.split(factor=.4)
        split.label(text="Final Object")
        split.prop(abp, "ab_final_object", text='Create      ', icon='OBJECT_DATAMODE')
        
        col = col.column()
        col.active = abp.ab_final_object and (not abp.ab_export_object or (abp.ab_export_object and not abp.ab_export_object_remove))

        split = col.split(factor=.4)
        split.label(text="Collection")
        row = split.row(align=True)
        row_b = row.row(align=True)
        row_b.alignment = 'LEFT'
        row_b.prop(abp, 'ab_final_collection_color', text='', icon_only=True)
        row.prop(abp, 'ab_final_collection', text='')

        split = col.split(factor=.4)
        split.label(text="Keep Name")
        split.row().prop(abp, "ab_object_keep_name", text='', icon='OBJECT_DATAMODE')
        
        split = col.split(factor=.4)
        split.alert = abp.ab_object_differentiator == ''
        split.label(text="Differentiator")
        
        
        col_b = split.column(align=True)
        col_b.prop(abp, "ab_object_differentiator", text='')
        
        if split.alert:
            col_b.label(text="This shouldn't be none!")

        split = col.split(factor=.4)
        split.label(text="Location")
        split.row().prop(abp, "ab_object_location", text=' ', expand=True)
        
        split = col.split(factor=.4)
        split.active = abp.ab_object_location == 'Copy'
        split.label(text="Location Offset")
        row = split.row(align=True)
        row_b = row.row(align=True)
        row_b.alignment = 'LEFT'
        row_b.prop(abp, 'ab_offset_direction', text='')
        row.prop(abp, 'ab_object_offset', text='')
        
        
class AUTOBAKE_PT_ObjectExport(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_objects"
    bl_idname = "AUTOBAKE_PT_objectexport"
    bl_label = ""
    bl_options = {'DEFAULT_CLOSED'}
        
    def draw_header(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        self.layout.active = abp.ab_final_object or scene.render.bake.use_selected_to_active
        self.layout.prop(abp, "ab_export_object", text="Export Object")
        
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
        
        layout.active = abp.ab_export_object and (abp.ab_final_object or scene.render.bake.use_selected_to_active)

        col = layout.column()
        
        split = col.split(factor=.4)
        split.label(text="Export When")
        split.prop(abp, "ab_export_always", text='')
        
        split = col.split(factor=.4)
        split.label(text="Final Object")
        split.prop(abp, "ab_export_object_remove", text='Delete       ', toggle=True, icon='TRASH')
        
        split = col.split(factor=.4)
        split.label(text="Clear Transform")
        row = split.row(align=True)
        row.prop(abp, 'ab_export_clear_location', toggle=True, text='Loc')
        row.prop(abp, 'ab_export_clear_rotation', toggle=True, text='Rot')
        row.prop(abp, 'ab_export_clear_scale', toggle=True, text='Scale')
        
        split = col.split(factor=.4)
        split.label(text="")
        split.active = not abp.ab_export_object_remove and (abp.ab_export_clear_location or abp.ab_export_clear_rotation or abp.ab_export_clear_scale)
        split.prop(abp, "ab_export_restore_transform")
        
        col = layout.column()
        
        split = col.split(factor=.4)
        split.label(text="Export As")
        split.prop(abp, "ab_export_object_as", text='')
        
        if abp.ab_export_object_as == 'GLTF':
            split = col.split(factor=.4)
            split.label(text="Format")
            split.prop(abp, "ab_gltf_format", text='')
            
            split = col.split(factor=.4)
            split.label(text="Copyright")
            split.prop(abp, "ab_gltf_copyright", text='')
            
            split = col.split(factor=.4)
            split.label(text="Lighting Mode")
            split.prop(abp, "ab_gltf_lighting", text='')
            
            split = col.split(factor=.4)
            split.label(text="")
            split.prop(abp, "ab_gltf_original_spelucar", text='Original Specular')

            split = col.split(factor=.4)
            split.label(text="")
            split.prop(abp, "ab_gltf_yup", text='+Y Up')
        
        else:
            split = col.split(factor=.4)
            split.label(text="Path Mode")
            row = split.row(align=True)
            row.prop(abp, "ab_export_pathmode", text='')
            
            if abp.ab_export_object_as == 'FBX':
                sub = row.row(align=True)
                sub.enabled = abp.ab_export_pathmode == 'COPY'
                sub.prop(abp, "ab_export_embedtextures", text='', icon='PACKAGE' if abp.ab_export_embedtextures else 'UGLYPACKAGE', icon_only=True)
        
        
class AUTOBAKE_PT_Transform(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_objectexport"
    bl_label = ""
        
    @classmethod
    def poll(cls, context):
        return context.scene.autobake_properties.ab_export_object_as in ['FBX', 'OBJ']
    
    def draw_header(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        self.layout.active = abp.ab_final_object and abp.ab_export_object
        self.layout.label(text="Transform")
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties

        layout.use_property_split = True
        layout.use_property_decorate = False
        layout.active = abp.ab_final_object and abp.ab_export_object

        col = layout.column()
        
        if abp.ab_export_object_as == 'FBX':
            col.prop(abp, "ab_export_scale", text='Scale')
            col.prop(abp, "ab_fbx_apply_scaling", text='Apply Scalings')
            col.prop(abp, "ab_fbx_forward")
            col.prop(abp, "ab_fbx_up")
            col.prop(abp, "ab_export_applyunit")
            col.prop(abp, "ab_export_usespacetransform")
        
        elif abp.ab_export_object_as == 'OBJ':
            col.prop(abp, "ab_export_scale", text='Scale')
            col.prop(abp, "ab_obj_forward")
            col.prop(abp, "ab_obj_up")
           

class AUTOBAKE_PT_Geometry(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_objectexport"
    bl_label = ""
        
    def draw_header(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        self.layout.active = abp.ab_final_object and abp.ab_export_object
        self.layout.label(text="Mesh" if abp.ab_export_object_as != 'FBX' else 'Geometry')
        
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties

        layout.use_property_split = True
        layout.use_property_decorate = False
        layout.active = abp.ab_final_object and abp.ab_export_object

        col = layout.column(align=True)
        
        if abp.ab_export_object_as == 'FBX':
            col.prop(abp, "ab_export_smoothing")
            col.prop(abp, "ab_export_subdivisionsurface")
            col.prop(abp, "ab_export_applymodifiers")
            col.prop(abp, "ab_export_looseedges")
            col.prop(abp, "ab_export_triangulatefaces")
            col.prop(abp, "ab_export_tangentspace")
            col.prop(abp, "ab_export_vertexcolors")
            col.prop(abp, "ab_export_prioritizeactivecolor")
            col.prop(abp, "ab_export_evalmode")
            col.prop(abp, "ab_export_customprops")

        elif abp.ab_export_object_as == 'OBJ':
            col.prop(abp, "ab_export_applymodifiers")
            col.prop(abp, "ab_export_evalmode")
            col.prop(abp, "ab_export_exportcolors")
            col.prop(abp, "ab_export_vertexgroups")
            col.prop(abp, "ab_export_triangulatedmesh")
            col.prop(abp, "ab_export_pbrextension")
            col.prop(abp, "ab_export_smoothgroups")
            col.prop(abp, "ab_export_groupbitflag")

        elif abp.ab_export_object_as == 'GLTF':
            col.prop(abp, "ab_export_applymodifiers")
            col.prop(abp, "ab_gltf_tangents")
            col.prop(abp, "ab_gltf_vertex_colors")
            col.prop(abp, "ab_gltf_attributes")
            col.prop(abp, "ab_gltf_loose_edges")
            col.prop(abp, "ab_gltf_loose_points")


class AUTOBAKE_PT_Images_GLTF(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_objectexport"
    bl_label = ""
        
    @classmethod
    def poll(cls, context):
        return context.scene.autobake_properties.ab_export_object_as == 'GLTF'
        
    def draw_header(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        
        self.layout.active = abp.ab_final_object and abp.ab_export_object
        self.layout.label(text='Images')
        
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
        
        layout.active = abp.ab_final_object and abp.ab_export_object

        layout.use_property_split = True
        layout.use_property_decorate = False

        col = layout.column(align=False)
        col.prop(abp, "ab_gltf_images")
        
        col = layout.column(align=True)
        col.active = abp.ab_gltf_images != 'NONE'
        row = col.row()
        row.active = not abp.ab_gltf_keep_original
        row.prop(abp, "ab_gltf_texture_folder", icon='FILE_FOLDER')
        
        col.prop(abp, "ab_gltf_keep_original")
        col.prop(abp, "ab_gltf_image_quality")
        col.prop(abp, "ab_gltf_create_webp")
        col.prop(abp, "ab_gltf_webp_fallback")


class AUTOBAKE_PT_ShapeKeys_GTLF(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_objectexport"
    bl_label = ""
        
    @classmethod
    def poll(cls, context):
        return context.scene.autobake_properties.ab_export_object_as == 'GLTF'
        
    def draw_header(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        self.layout.active = abp.ab_final_object and abp.ab_export_object
        self.layout.prop(abp, "ab_gltf_shape_keys")
        
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties

        layout.use_property_split = True
        layout.use_property_decorate = False
        layout.active = abp.ab_final_object and abp.ab_export_object and abp.ab_gltf_shape_keys

        col = layout.column(align=True)
        col.prop(abp, "ab_gltf_shape_keys_normals")
        col.prop(abp, "ab_gltf_shape_keys_tangents")
        col.prop(abp, "ab_gltf_use_sparse")
        col.prop(abp, "ab_gltf_omitting_sparse")


class AUTOBAKE_PT_Compression_GLTF(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_objectexport"
    bl_label = ""
        
    @classmethod
    def poll(cls, context):
        return context.scene.autobake_properties.ab_export_object_as == 'GLTF'
        
    def draw_header(self, context):
        scene = context.scene
        abp = scene.autobake_properties
        self.layout.active = abp.ab_final_object and abp.ab_export_object
        self.layout.prop(abp, "ab_gltf_compression")
        
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties

        layout.use_property_split = True
        layout.use_property_decorate = False
        layout.active = abp.ab_final_object and abp.ab_export_object and abp.ab_gltf_compression

        col = layout.column(align=True)
        col.prop(abp, "ab_gltf_compression_level")
        col = layout.column(align=True)
        col.prop(abp, "ab_gltf_compression_position")
        col.prop(abp, "ab_gltf_compression_normal")
        col.prop(abp, "ab_gltf_compression_texcoord")
        col.prop(abp, "ab_gltf_compression_color")
        col.prop(abp, "ab_gltf_compression_generic")


class AUTOBAKE_PT_BakeList(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_addon_settings"
    bl_label = "Bake Items"
        
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
    
        col = layout.column()
        
        split = col.split(factor=.4)
        split.label(text="Add New")
        split.row().prop(abp, "ab_new_item_method", text=' ', expand=True)

        split = col.split(factor=.4)
        split.label(text="Dynamic Scale")
        split.prop(abp, "ab_dynamic_scale", text='')
        
        col = layout.column()

        split = col.split(factor=.4)
        split.label(text="Bake List")
        split.menu(AUTOBAKE_MT_ItemEdit.bl_idname, icon='PRESET')
        
        split = col.split(factor=.4)
        split.label(text="UDIM List")
        split.menu(AUTOBAKE_MT_ItemEdit_UDIM.bl_idname, icon="PRESET")
        

class AUTOBAKE_PT_BakeQueue(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_addon_settings"
    bl_label = "Bake Queue"
        
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties
    
        col = layout.column()
        
        split = col.split(factor=.4)
        split.label(text="Finished Bake")
        split.prop(abp, "ab_move_finished_bake", text='Move Down'+'      ', toggle=True, icon='SORT_ASC')
        
        split = col.split(factor=.4)
        split.active = not abp.ab_move_finished_bake
        split.label(text="Active Bake")
        split.prop(abp, "ab_move_active_bake", text='Move Up'+'      ', toggle=True, icon='SORT_DESC')
        
        col = layout.column()
        
        split = col.split(factor=.4)
        split.label(text='Next Object')
        split.prop(abp, 'ab_auto_next', text='')
        
        split = col.split(factor=.4)
        split.label(text='Confirm Results')
        split.prop(abp, 'ab_auto_confirm', text='')
        
        
class AUTOBAKE_PT_Miscellaneous(AUTOBAKE_PT_Main, Panel):
    bl_parent_id = "AUTOBAKE_PT_addon_settings"
    bl_label = "Miscellaneous"
        
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        abp = scene.autobake_properties

        col = layout.column()
        
        split = col.split(factor=.4)
        split.label(text='Confirm Popups')
        split.menu(AUTOBAKE_MT_Confirms.bl_idname, icon="WINDOW", text='Bake')
        
        split = col.split(factor=.4)
        split.label(text='Start Popup')
        split.menu(AUTOBAKE_MT_StartPopupSettings.bl_idname, text='Settings', icon='MOD_HUE_SATURATION')
        
        split = col.split(factor=.4)
        split.label(text="Export List")
        split.prop(abp, "ab_export_list", text='', icon='EXPORT')
        
        col = layout.column()

        split = col.split(factor=.4)
        split.label(text='List Layout')
        split.prop(abp, 'ab_list_padding', icon="SEQ_STRIP_DUPLICATE", text='')
        
        col = layout.column()
        
        split = col.split(factor=.4)
        split.label(text="Text Alerts")
        split.menu(AUTOBAKE_MT_Alerts.bl_idname, icon="ERROR", text='Alert')
        
        split = col.split(factor=.4)
        split.label(text='Reports')
        split.menu(AUTOBAKE_MT_Reports.bl_idname, text='Actions', icon='INFO')
        
        split = col.split(factor=.4)
        split.label(text='Subfolder Name')
        split.prop(abp, 'ab_subfolder_use_prefix', icon="NEWFOLDER", text='Add Prefix      ')


#------------------------------------------------------------------------------------
#   Menu
        
class AUTOBAKE_MT_BakeList(Menu):
    bl_label = "List Settings"
    bl_idname = "AUTOBAKE_MT_bakelist"
    
    def draw(self, context):
        abp = context.scene.autobake_properties
        layout = self.layout
        layout.operator("autobake.bakelist_load_linked", text="Load by Linked", icon='IMPORT')
        layout.separator()
        layout.operator("autobake.move_to_top", text="Selected to Top", icon='TRIA_UP_BAR')
        layout.operator("autobake.move_to_bottom", text="Selected to Bottom", icon='TRIA_DOWN_BAR')
        layout.separator()
        layout.operator("autobake.enable_all", text="Enable All", icon='CHECKBOX_HLT')
        layout.operator("autobake.disable_all", text="Disable All", icon='CHECKBOX_DEHLT')
        layout.operator("autobake.invert_all", text="Invert All", icon='BLANK1')
        layout.separator()
        layout.operator("autobake.remove_all", text="Clear List", icon='REMOVE')
        layout.operator("autobake.remove_duplicates", text="Remove Duplicated", icon='BLANK1')
        layout.operator("autobake.remove_disabled", text="Remove Disabled", icon='BLANK1')
        layout.separator()
        layout.prop(abp, "ab_item_details", text="Item Details")
        layout.separator()
        layout.operator("autobake.reorder", text="Sort Items", icon='SORTALPHA')
        layout.separator()
        layout.prop(abp, "ab_udim_bake", text="UDIM Bake")
        
        
class AUTOBAKE_MT_UDIMList(Menu):
    bl_label = "List Settings"
    bl_idname = "AUTOBAKE_MT_udimlist"
    
    def draw(self, context):
        layout = self.layout
        layout.operator('autobake.udimlist_import', text='Load Tiles', icon='IMPORT')
        layout.separator()
        layout.operator('autobake.move_to_top_udim', text="Selected to Top", icon='TRIA_UP_BAR')
        layout.operator('autobake.move_to_bottom_udim', text="Selected to Bottom", icon='TRIA_DOWN_BAR')
        layout.separator()
        layout.operator('autobake.udimlist_enable_all', icon='CHECKBOX_HLT')
        layout.operator('autobake.udimlist_disable_all', icon='CHECKBOX_DEHLT')
        layout.operator('autobake.udimlist_invert_all')
        layout.separator()
        layout.operator('autobake.udimlist_remove_all', icon='REMOVE', text='Clear List')
        layout.operator('autobake.udimlist_remove_duplicates')
        layout.operator('autobake.udimlist_remove_disabled')
        layout.separator()
        layout.prop(context.scene.autobake_properties, 'ab_udim_label', text='Tile Label')
        layout.separator()
        layout.operator('autobake.udimlist_sort', icon='SORTALPHA')
        

class AUTOBAKE_MT_ItemEdit(Menu):
    bl_label = "Item Edit"
    bl_idname = "AUTOBAKE_MT_itemedit"
    bl_description ='Set how the items in the bake list (non-udim) can be edited'
    
    def draw(self, context):
        abp = context.scene.autobake_properties
        layout = self.layout
        layout.label(text='In-line')
        layout.prop(abp, "ab_item_type", text='Type (In Line)')
        layout.prop(abp, "ab_item_scale", text='Scale (In Line)')
        layout.separator()
        layout.label(text='Default')
        layout.prop(abp, "ab_item_default", text='Type & Scale')
        

class AUTOBAKE_MT_ItemEdit_UDIM(Menu):
    bl_label = "Item Edit"
    bl_idname = "AUTOBAKE_MT_udimitem"
    bl_description ='Set how the items in both of the UDIM Lists can be edited'
    
    def draw(self, context):
        abp = context.scene.autobake_properties
        layout = self.layout
        layout.label(text='In-line')
        layout.prop(abp, "ab_udim_item_type", text='Type')
        layout.prop(abp, "ab_udim_item_multiplier", text='Multiplier')
        layout.prop(abp, "ab_udim_item_tile", text='Tile')
        layout.prop(abp, "ab_udim_item_scale", text='Scale')
        layout.separator()
        layout.label(text='Default')
        layout.prop(abp, "ab_udim_item_default", text='Type & Multiplier')
        layout.prop(abp, "ab_udim_item_tilescale_default", text='Tile & Scale')


class AUTOBAKE_MT_ColorSpace(Menu):
    bl_label = "Color Space"
    bl_idname = "AUTOBAKE_MT_colorspace"
    bl_description ="Set what color space different texture types should use when creating them"
    
    def draw(self, context):
        layout = self.layout
        abp = context.scene.autobake_properties
        layout.label(text='Color Textures', icon='SEQUENCE_COLOR_03')
        layout.prop(abp, "ab_color_space_color", text="")
        layout.separator()
        layout.label(text='Vector Textures', icon='SEQUENCE_COLOR_06')
        layout.prop(abp, "ab_color_space_vector", text="")
        layout.separator()
        layout.label(text='Float Textures', icon='SEQUENCE_COLOR_09')
        layout.prop(abp, "ab_color_space_float", text="")


class AUTOBAKE_MT_StartPopupSettings(Menu):
    bl_label = "Start Settings"
    bl_idname = "AUTOBAKE_MT_startpopupsettings"
    bl_description ="Control what settings are shows in the bake start popup window"
    
    def draw(self, context):
        layout = self.layout
        abp = context.scene.autobake_properties
        
        layout.prop(abp, "ab_start_popup_settings", text='Settings (All)')
        
        col = layout.column()
        col.active = abp.ab_start_popup_settings

        col.separator()
        col.prop(abp, "ab_start_popup_final_material")
        
        col_m = col.column()
        col_m.active = abp.ab_start_popup_final_material
        col_m.prop(abp, "ab_start_popup_final_shader", text='Material Shader')
        col_m.prop(abp, "ab_start_popup_texture_apply", text='Apply Textures')
        
        col.separator()
        col.prop(abp, "ab_start_popup_final_object")
        col_m = col.column()
        col_m.active = abp.ab_start_popup_final_object
        col_m.prop(abp, "ab_start_popup_object_offset", text='Object Offset')
        col_m.prop(abp, "ab_start_popup_export_objects", text='Export Object')
        col.prop(abp, "ab_start_popup_selected_to_active")
        
        col.separator()
        col.prop(abp, "ab_start_popup_export_textures")
        col.prop(abp, "ab_start_popup_keep_textures")


class AUTOBAKE_MT_Confirms(Menu):
    bl_label = "Confirm Windows"
    bl_idname = "AUTOBAKE_MT_confirms"
    bl_description ="Disable confirm windows for buttons. Confirm windows can also be skipped with using 'CTRL + Click'"
    
    def draw(self, context):
        layout = self.layout
        abp = context.scene.autobake_properties
        layout.prop(abp, "ab_confirm_start_bake", text='Start')
        layout.prop(abp, "ab_confirm_cancel_bake", text='Cancel')
        layout.prop(abp, "ab_confirm_bake_results", text='Results')
        layout.prop(abp, "ab_confirm_next_object", text='Next Object')
        layout.prop(abp, "ab_confirm_queue_item_gate", text='Queue Item')
        
       
class AUTOBAKE_MT_Alerts(Menu):
    bl_label = "Alerts"
    bl_idname = "AUTOBAKE_MT_alerts"
    bl_description ='Set the visibility of the addon tooltips'
    
    def draw(self, context):
        abp = context.scene.autobake_properties
        layout = self.layout
        layout.prop(abp, "ab_alert_texts", text='UI Elements')
        layout.prop(abp, "ab_nodealert", text='Shader Editor')
        layout.prop(abp, "ab_bake_error_msg", text='Bake Errors')
        
        
class AUTOBAKE_MT_Reports(Menu):
    bl_label = "Reports"
    bl_idname = "AUTOBAKE_MT_Reports"
    bl_description ="Set which actions Auto Bake should report during the baking process. Some error reports can't be disabled, since they provide valuable feedback when an issue occurs"
    
    def draw(self, context):
        abp = context.scene.autobake_properties
        layout = self.layout
        layout.prop(abp, "ab_report_requests", text="Requests")
        layout.separator()
        layout.prop(abp, "ab_report_bake_start", text="Bake Start")
        layout.prop(abp, "ab_report_bake_end", text="Bake End")
        layout.separator()
        layout.prop(abp, "ab_report_object_start", text="Object Start")
        layout.prop(abp, "ab_report_object_end", text="Object End")
        layout.separator()
        layout.prop(abp, "ab_report_texture_export", text="Texture Export")
        layout.prop(abp, "ab_report_object_export", text="Object Export")
        layout.separator()
        layout.prop(abp, "ab_report_bake_summary", text="Bake Summary")
        layout.prop(abp, "ab_report_object_summary", text="Object Summary")
        layout.separator()
        layout.prop(abp, "ab_report_bake_error", text="Bake Errors")


#------------------------------------------------------------------------------------
#   Register

classes = (AUTOBAKE_PG_Collection, AUTOBAKE_PG_Bake, AUTOBAKE_PG_BakeQueue, AUTOBAKE_PG_UDIMType, AUTOBAKE_PG_SourceObjects, AUTOBAKE_PG_ImageExport, AUTOBAKE_PG_ObjectQueue,
        # Operator
            AUTOBAKE_OT_BakeStart, AUTOBAKE_OT_CancelBake, AUTOBAKE_OT_PauseBake, AUTOBAKE_OT_ResumeBake,
            AUTOBAKE_OT_Add, AUTOBAKE_OT_Remove, AUTOBAKE_OT_Up, AUTOBAKE_OT_Down, AUTOBAKE_OT_ScaleDown, AUTOBAKE_OT_ScaleUp, AUTOBAKE_OT_LoadLinked,
            AUTOBAKE_OT_Add_UDIM, AUTOBAKE_OT_Remove_UDIM, AUTOBAKE_OT_Down_UDIM, AUTOBAKE_OT_Up_UDIM, AUTOBAKE_OT_MoveTop_UDIM, AUTOBAKE_OT_MoveBottom_UDIM,
            AUTOBAKE_OT_RemoveDisabled, AUTOBAKE_OT_RemoveDuplicates, AUTOBAKE_OT_RemoveAll, AUTOBAKE_OT_EnableAll, AUTOBAKE_OT_DisableAll, AUTOBAKE_OT_InvertAll, AUTOBAKE_OT_MoveTop, AUTOBAKE_OT_MoveBottom, AUTOBAKE_PG_UDIMTile, 
            AUTOBAKE_OT_RemoveDisabled_UDIM, AUTOBAKE_OT_RemoveDuplicates_UDIM, AUTOBAKE_OT_RemoveAll_UDIM, AUTOBAKE_OT_EnableAll_UDIM, AUTOBAKE_OT_DisableAll_UDIM, AUTOBAKE_OT_InvertAll_UDIM, AUTOBAKE_OT_Sort_UDIM, AUTOBAKE_OT_ImportTiles_UDIM,
            AUTOBAKE_OT_ToggleQueueItem, AUTOBAKE_OT_ToggleObjectQueueItem, AUTOBAKE_OT_BakeConfirm, AUTOBAKE_OT_NextObject, AUTOBAKE_OT_ExportTextures, AUTOBAKE_OT_Sort, AUTOBAKE_OT_NameStructure, AUTOBAKE_OT_FolderExplorer,
            AUTOBAKE_OT_SelectFromList, AUTOBAKE_OT_LoadFromSelected,
        # UIList
            AUTOBAKE_UL_Bake, AUTOBAKE_UL_BakeQueue, AUTOBAKE_UL_UDIMTile, AUTOBAKE_UL_UDIMType, AUTOBAKE_UL_SourceObjects, AUTOBAKE_UL_ImageExport, AUTOBAKE_UL_ObjectQueue,
        # Menu
            AUTOBAKE_MT_BakeList, AUTOBAKE_MT_UDIMList, AUTOBAKE_MT_ItemEdit, AUTOBAKE_MT_ItemEdit_UDIM, AUTOBAKE_MT_Confirms, AUTOBAKE_MT_Alerts, AUTOBAKE_MT_ColorSpace, AUTOBAKE_MT_StartPopupSettings,AUTOBAKE_MT_Reports,
        # Panel
            AUTOBAKE_PT_Bake, AUTOBAKE_PT_Lists, AUTOBAKE_PT_List_UDIM,
            AUTOBAKE_PT_Image, AUTOBAKE_PT_Name, AUTOBAKE_PT_Format, AUTOBAKE_PT_Export, AUTOBAKE_PT_ColorOverride,
            AUTOBAKE_PT_Settings, AUTOBAKE_PT_Margin, AUTOBAKE_PT_SelectedToActive, AUTOBAKE_PT_SelectionHelp, AUTOBAKE_PT_Sampling, AUTOBAKE_PT_Denoise, AUTOBAKE_PT_Sampling_Low, AUTOBAKE_PT_Denoise_Low, AUTOBAKE_PT_Sampling_High, AUTOBAKE_PT_Denoise_High,
            AUTOBAKE_PT_Addon, AUTOBAKE_PT_ImageSettings, AUTOBAKE_PT_Materials, AUTOBAKE_PT_Objects, AUTOBAKE_PT_ObjectExport, AUTOBAKE_PT_BakeList, AUTOBAKE_PT_BakeQueue, AUTOBAKE_PT_Miscellaneous,
            AUTOBAKE_PT_Transform, AUTOBAKE_PT_Geometry, AUTOBAKE_PT_Compression_GLTF, AUTOBAKE_PT_ShapeKeys_GTLF, AUTOBAKE_PT_Images_GLTF)
            

def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)
     
    bpy.types.Scene.autobake_properties = PointerProperty(type=AUTOBAKE_PG_Collection)

    bpy.types.Scene.autobake_bakelist = CollectionProperty(type=AUTOBAKE_PG_Bake)
    bpy.types.Scene.autobake_bakelist_index = IntProperty(name="Bake Item", default=0, update=name_by_index)

    bpy.types.Scene.autobake_udimlist = CollectionProperty(type=AUTOBAKE_PG_UDIMType)
    bpy.types.Scene.autobake_udimlist_index = IntProperty(name="Bake Item", default=0, update=name_by_index)

    bpy.types.Scene.autobake_udimtilelist = CollectionProperty(type=AUTOBAKE_PG_UDIMTile)
    bpy.types.Scene.autobake_udimtilelist_index = IntProperty(name="UDIM Tile", default=0)

    bpy.types.Scene.autobake_imageexport = CollectionProperty(type=AUTOBAKE_PG_ImageExport)
    bpy.types.Scene.autobake_imageexport_index = IntProperty(name="Image", default=0)

    bpy.types.Scene.autobake_sourceobject = CollectionProperty(type=AUTOBAKE_PG_SourceObjects)
    bpy.types.Scene.autobake_sourceobject_index = IntProperty(name="Source Object", default=0)

    bpy.types.Scene.autobake_queuelist = CollectionProperty(type=AUTOBAKE_PG_BakeQueue)
    bpy.types.Scene.autobake_queuelist_index = IntProperty(name="Queue Item", default=0)
    
    bpy.types.Scene.autobake_objectqueuelist = CollectionProperty(type=AUTOBAKE_PG_ObjectQueue)
    bpy.types.Scene.autobake_objectqueuelist_index = IntProperty(name="Object Bake Item", default=0, update=bake_result)


def unregister():
    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)
        
    del bpy.types.Scene.autobake_properties
    
    del bpy.types.Scene.autobake_bakelist
    del bpy.types.Scene.autobake_bakelist_index
    
    del bpy.types.Scene.autobake_udimlist
    del bpy.types.Scene.autobake_udimlist_index
    
    del bpy.types.Scene.autobake_udimtilelist
    del bpy.types.Scene.autobake_udimtilelist_index
    
    del bpy.types.Scene.autobake_imageexport
    del bpy.types.Scene.autobake_imageexport_index
    
    del bpy.types.Scene.autobake_sourceobject
    del bpy.types.Scene.autobake_sourceobject_index
    
    del bpy.types.Scene.autobake_queuelist
    del bpy.types.Scene.autobake_queuelist_index
    
    del bpy.types.Scene.autobake_objectqueuelist
    del bpy.types.Scene.autobake_objectqueuelist_index
    
if __name__ == "__main__":
    register()
